/*	$Id: cpu_60x.c,v 1.13 2010/05/19 06:51:56 tsubai Exp $	*/

/*-
 * Copyright (c) 2002-2004 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>
#include <machine/spr.h>

/* Hardware Implementation Dependent registers for the 603 and 604 */

#define HID0_EMCP   0x80000000  /* Enable MCP */
#define HID0_ECPE   0x40000000  /* Enable cache parity checking (604) */
#define HID0_EBA    0x20000000  /* Enable 60x bus address parity checking */
#define HID0_EBD    0x10000000  /* Enable 60x bus data parity checking */
#define HID0_SBCLK  0x08000000  /* Select bus clock for test clock pin (603) */
#define HID0_EICE   0x04000000  /* Enable ICE output (603) */
#define HID0_ECLK   0x02000000  /* CLK_OUT clock type selection (603) */
#define HID0_PAR    0x01000000  /* Disable precharge of ARTRY */
#define HID0_DOZE   0x00800000  /* Enable doze mode (603) */
#define HID0_NAP    0x00400000  /* Enable nap mode (603) */
#define HID0_SLEEP  0x00200000  /* Enable sleep mode (603) */
#define HID0_DPM    0x00100000  /* Enable dynamic power management (603) */
#define HID0_RISEG  0x00080000  /* Read I-SEG (603) */
#define HID0_NHR    0x00010000  /* Not hard reset (604) */
#define HID0_ICE    0x00008000  /* Enable i-cache */
#define HID0_DCE    0x00004000  /* Enable d-cache */
#define HID0_ILOCK  0x00002000  /* i-cache lock */
#define HID0_DLOCK  0x00001000  /* d-cache lock */
#define HID0_ICFI   0x00000800  /* i-cache flush invalidate */
#define HID0_DCFI   0x00000400  /* d-cache flush invalidate */
#define HID0_IFEM   0x00000100  /* Enable M-bit for I-fetch (604e) */
#define HID0_SED    0x00000080  /* Serial instruction execution disable (604) */
#define HID0_FBIOB  0x00000010  /* Force branch indirect on bus (603) */
#define HID0_BHT    0x00000004  /* Enable branch history table (604) */
#define HID0_BTACD  0x00000002  /* BTAC disable (604e) */
#define HID0_NOOPTI 0x00000001  /* No-op the dcbt(st) (603) */

#define HID0_BITMASK "\020" \
    "\040EMCP\037ECPE\036EBA\035EBD\034SBCLK\033EICE\032ECLK\031PAR" \
    "\030DOZE\027NAP\026SLEEP\025DPM\024RISEG\021NHR" \
    "\020ICE\017DCE\016ILOCK\015DLOCK\014ICFI\013DCFI\011IFEM" \
    "\010SED\005FBIOB\003BHT\002BTACD\001NOOPTI"

void
mpc603_config_hid()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	/* XXX for now */
	hid0 &= ~(HID0_DOZE | HID0_NAP | HID0_SLEEP);
	hid0 |= HID0_DOZE | HID0_DPM;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc604_config_hid()
{
	u_int hid0;

	asm ("mfspr %0,HID0" : "=r"(hid0));

	hid0 &= ~(HID0_ILOCK | HID0_DLOCK);
	hid0 |= HID0_SED;

	asm volatile ("mtspr HID0,%0" :: "r"(hid0));
}

void
mpc60x_print_hid(name)
	const char *name;
{
	u_int hid0;
	char buf[160];

	asm ("mfspr %0,HID0" : "=r"(hid0));
	snprintb(buf, sizeof buf, HID0_BITMASK, hid0);
	printf("%s: HID0 %s\n", name, buf);
}

void
mpc60x_enable_L1cache()
{
	u_int hid0;

	/* Enable L1 caches. */
	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	if ((hid0 & HID0_ICE) == 0) {
		hid0 |= HID0_ICFI|HID0_ICE;
		asm volatile ("isync");
		asm volatile ("mtspr HID0,%0" :: "r"(hid0));
		asm volatile ("isync");
	}

	asm volatile ("mfspr %0,HID0" : "=r"(hid0));
	if ((hid0 & HID0_DCE) == 0) {
		hid0 |= HID0_DCFI|HID0_DCE;
		asm volatile ("sync");
		asm volatile ("mtspr HID0,%0" :: "r"(hid0));
	}
}

u_int
mpc603_corefreq(busfreq)
	u_int busfreq;
{
	u_int hid1, pll;

	/* Bus to Core Multiplier * 2 */
	static const char pll_mul2[] = {
		2,  2, 14,  2, 4,  4, 5, 9,
		6, 11,  8, 10, 3, 12, 7, 0
	};

	switch (mfpvr() >> 16) {
	case MPC603ev:
		asm ("mfspr %0,HID1" : "=r"(hid1));
		pll = hid1 >> 28;
		return busfreq * pll_mul2[pll] / 2;
	default:	/* XXX not confirmed yet */
		return 0;
	}
}

u_int
mpc604_corefreq(busfreq)
	u_int busfreq;
{
	u_int hid1, pll;

	/* Bus to Core Multiplier * 2 */
	static const char pll_mul2[] = {
		2,  2, 14,  2, 4, 13, 5, 9,
		6, 11,  8, 10, 3, 12, 7, 0
	};

	switch (mfpvr() >> 16) {
	case MPC604e:
		asm ("mfspr %0,HID1" : "=r"(hid1));
		pll = hid1 >> 28;
		return busfreq * pll_mul2[pll] / 2;
	default:	/* XXX not confirmed yet */
		return 0;
	}
}

/*
 * PLL_CFG configuration for 603e (PID6 and PID7)
 * [0:3]
 * 0000  0  1x
 * 0001  1  1x
 * 0011  3  PLL Bypass
 * 0100  4  2x
 * 0101  5  2x
 * 0110  6  2.5x
 * 0111  7  4.5x
 * 1000  8  3x
 * 1001  9  5.5x
 * 1010 10  4x
 * 1011 11  5x
 * 1100 12  1.5x
 * 1101 13  6x
 * 1110 14  3.5x
 * 1111 15  Off

 * for 604e
 * 0000    0    1:1
 * 0001    1    1:1
 * 0010    2    7:1
 * 0011    3   bypass
 * 0100    4    2:1
 * 0101    5    6.5:1
 * 0110    6    2.5:1
 * 0111    7    4.5:1
 * 1000    8    3:1
 * 1001    9    5.5:1
 * 1010    10   4:1
 * 1011    11   5:1
 * 1100    12   1.5:1
 * 1101    13   6:1
 * 1110    14   3.5:1
 * 1111    15  off
 */
