/*	$Id: db_disasm.c,v 1.3 2007/03/01 21:40:27 tsubai Exp $	*/

/*-
 * Copyright (c) 2000-2002 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>

#include <machine/db_machdep.h>

#include <ddb/db_access.h>
#include <ddb/db_sym.h>
#include <ddb/db_variables.h>
#include <ddb/db_interface.h>
#include <ddb/db_output.h>

/*
 * TODO:
 *
 * Use simplified mnemonic.  (b)
 */

static char *sprname(int);
static inline u_int extract_field(u_int, int, int);

#define D	0x00000001		/* rD */
#define S	0x00000002		/* rS */
#define A	0x00000004		/* rA */
#define B	0x00000008		/* rB */
#define C	0x00000010		/* rC */
#define Rc	0x00000020		/* . */
#define OE	0x00000040		/* o */
#define d	0x00000100		/* nnn(rA) */
#define UIMM	0x00000200		/* unsigned immediate */
#define SIMM	0x00000400		/* signed immediate */
#define crfD	0x00001000
#define crfS	0x00002000

/* XXX XXX XXX */
#define NB	B
#define TO	D
#define SH	B
#define crbA	A
#define crbB	B
#define crbD	D

#define AA	0
#define LK	0
#define BO	0x00000001
#define BI	0x00000002
#define BD	0
#define LI	0
#define MB	0
#define ME	0
#define SR	0
#define FM	0
#define CRM	0
#define SPR	0
#define IMM	0
/* #define L	0 */
/* #define tbr	0 */

struct opcode {
	char name[8];
	u_int mask;
	u_int code;
	u_int flags;
};

static struct opcode opcodes[] = {
 { "nop", 	0xffffffff, 0x60000000 },		/* simplified */
 { "li", 	0xfc1f0000, 0x38000000, D | SIMM },	/* simplified */
 { "lis", 	0xfc1f0000, 0x3c000000, D | SIMM },	/* simplified */

 { "add",	0xfc0003fe, 0x7c000214, D | A | B | OE | Rc },
 { "addc",	0xfc0003fe, 0x7c000014, D | A | B | OE | Rc },
 { "adde",	0xfc0003fe, 0x7c000114, D | A | B | OE | Rc },
 { "addi",	0xfc000000, 0x38000000, D | A | SIMM },
 { "addic",	0xfc000000, 0x30000000, D | A | SIMM },
 { "addic.",	0xfc000000, 0x34000000, D | A | SIMM },
 { "addis",	0xfc000000, 0x3c000000, D | A | SIMM },
 { "addme",	0xfc00fbfe, 0x7c0001d4, D | A | OE | Rc },
 { "addze",	0xfc00fbfe, 0x7c000194, D | A | OE | Rc },
 { "and",	0xfc0007fe, 0x7c000038, S | A | B | Rc },
 { "andc",	0xfc0007fe, 0x7c000078, S | A | B | Rc },
 { "andi.",	0xfc000000, 0x70000000, S | A | UIMM },
 { "andis.",	0xfc000000, 0x74000000, S | A | UIMM },

 { "bctr",	0xfffffffe, 0x4e800420, LK },		/* simplified */
 { "blr",	0xfffffffe, 0x4e800020, LK },		/* simplified */
 { "beq",	0xffdf0000, 0x41820000, BD | AA | LK },	/* simplified */
 { "bge",	0xffdf0000, 0x40800000, BD | AA | LK },	/* simplified */
 { "bgt",	0xffdf0000, 0x41810000, BD | AA | LK },	/* simplified */
 { "ble",	0xffdf0000, 0x40810000, BD | AA | LK },	/* simplified */
 { "blt",	0xffdf0000, 0x41800000, BD | AA | LK },	/* simplified */
 { "bne",	0xffdf0000, 0x40820000, BD | AA | LK },	/* simplified */
 { "bns",	0xffdf0000, 0x40830000, BD | AA | LK },	/* simplified */
 { "bso",	0xffdf0000, 0x41830000, BD | AA | LK },	/* simplified */

 { "bdnzf",	0xffc00000, 0x40000000, BI | BD | AA | LK }, /* simplified */
 { "bdzf",	0xffc00000, 0x40400000, BI | BD | AA | LK }, /* simplified */
 { "bdnzt",	0xffc00000, 0x41000000, BI | BD | AA | LK }, /* simplified */
 { "bdzt",	0xffc00000, 0x41400000, BI | BD | AA | LK }, /* simplified */
 { "bdnz",	0xffdf0000, 0x42000000, BD | AA | LK },	/* simplified */
 { "bdz",	0xffdf0000, 0x42400000, BD | AA | LK },	/* simplified */
 { "bt",	0xffc00000, 0x41800000, BI | BD | AA | LK }, /* simplified */
 { "bf",	0xffc00000, 0x40800000, BI | BD | AA | LK }, /* simplified */

 { "b",		0xfc000000, 0x48000000, LI | AA | LK },
 { "bc",	0xfc000000, 0x40000000, BO | BI | BD | AA | LK },
 { "bcctr",	0xfc00fffe, 0x4c000420, BO | BI | LK },
 { "bclr",	0xfc00fffe, 0x4c000020, BO | BI | LK },
 { "cmpw",	0xfc6007ff, 0x7c000000, crfD | A | B },    /* simplified */
 { "cmpwi",	0xfc600000, 0x2c000000, crfD | A | SIMM }, /* simplified */
 { "cmplw",	0xfc6007ff, 0x7c000040, crfD | A | B },    /* simplified */
 { "cmplwi",	0xfc600000, 0x28000000, crfD | A | UIMM }, /* simplified */
 { "cmpd",	0xfc6007ff, 0x7c200000, crfD | A | B },    /* simplified */
 { "cmpdi",	0xfc600000, 0x2c200000, crfD | A | SIMM }, /* simplified */
 { "cmpld",	0xfc6007ff, 0x7c200040, crfD | A | B },    /* simplified */
 { "cmpldi",	0xfc600000, 0x28200000, crfD | A | UIMM }, /* simplified */
/* { "cmp",	0xfc4007ff, 0x7c000000, crfD | L | A | B },	*/
/* { "cmpi",	0xfc400000, 0x2c000000, crfD | L | A | SIMM },	*/
/* { "cmpl",	0xfc4007ff, 0x7c000040, crfD | L | A | B },	*/
/* { "cmpli",	0xfc400000, 0x28000000, crfD | L | A | UIMM },	*/
 { "cntlzw",	0xfc00fffe, 0x7c000034, S | A | Rc },
 { "crand",	0xfc0007ff, 0x4c000202, crbD | crbA | crbB },
 { "crandc",	0xfc0007ff, 0x4c000102, crbD | crbA | crbB },
 { "creqv",	0xfc0007ff, 0x4c000242, crbD | crbA | crbB },
 { "crnand",	0xfc0007ff, 0x4c0001c2, crbD | crbA | crbB },
 { "crnor",	0xfc0007ff, 0x4c000042, crbD | crbA | crbB },
 { "cror",	0xfc0007ff, 0x4c000382, crbD | crbA | crbB },
 { "crorc",	0xfc0007ff, 0x4c000342, crbD | crbA | crbB },
 { "crxor",	0xfc0007ff, 0x4c000182, crbD | crbA | crbB },
 { "dcba",	0xffe007ff, 0x7c0005ec, A | B },
 { "dcbf",	0xffe007ff, 0x7c0000ac, A | B },
 { "dcbi",	0xffe007ff, 0x7c0003ac, A | B },
 { "dcbst",	0xffe007ff, 0x7c00006c, A | B },
 { "dcbt",	0xffe007ff, 0x7c00022c, A | B },
 { "dcbtst",	0xffe007ff, 0x7c0001ec, A | B },
 { "dcbz",	0xffe007ff, 0x7c0007ec, A | B },
 { "divw",	0xfc0003fe, 0x7c0003d6, D | A | B | OE | Rc },
 { "divwu",	0xfc0003fe, 0x7c000396, D | A | B | OE | Rc },
 { "eciwx",	0xfc0007ff, 0x7c00026c, D | A | B },
 { "ecowx",	0xfc0007ff, 0x7c00036c, S | A | B },
 { "eieio",	0xffffffff, 0x7c0006ac },
 { "eqv",	0xfc0007fe, 0x7c000238, S | A | B | Rc },
 { "extsb",	0xfc00fffe, 0x7c000774, S | A | Rc },
 { "extsh",	0xfc00fffe, 0x7c000734, S | A | Rc },
 { "fabs",	0xfc1f07fe, 0xfc000210, D | B | Rc },
 { "fadd",	0xfc0007fe, 0xfc00002a, D | A | B | Rc },
 { "fadds",	0xfc0007fe, 0xec00002a, D | A | B | Rc },
 { "fcmpo",	0xfc6007ff, 0xfc000040, crfD | A | B },
 { "fcmpu",	0xfc6007ff, 0xfc000000, crfD | A | B },
 { "fctiw",	0xfc1f07fe, 0xfc00001c, D | B | Rc },
 { "fctiwz",	0xfc1f07fe, 0xfc00001e, D | B | Rc },
 { "fdiv",	0xfc0007fe, 0xfc000024, D | A | B | Rc },
 { "fdivs",	0xfc0007fe, 0xec000024, D | A | B | Rc },
 { "fmadd",	0xfc00003e, 0xfc00003a, D | A | B | C | Rc },
 { "fmadds",	0xfc00003e, 0xec00003a, D | A | B | C | Rc },
 { "fmr",	0xfc1f07fe, 0xfc000090, D | B | Rc },
 { "fmsub",	0xfc00003e, 0xfc000038, D | A | B | C | Rc },
 { "fmsubs",	0xfc00003e, 0xec000038, D | A | B | C | Rc },
 { "fmul",	0xfc00f83e, 0xfc000032, D | A | C | Rc },
 { "fmuls",	0xfc00f83e, 0xec000032, D | A | C | Rc },
 { "fnabs",	0xfc1f07fe, 0xfc000110, D | B | Rc },
 { "fneg",	0xfc1f07fe, 0xfc000050, D | B | Rc },
 { "fnmadd",	0xfc00003e, 0xfc00003e, D | A | B | C | Rc },
 { "fnmadds",	0xfc00003e, 0xec00003e, D | A | B | C | Rc },
 { "fnmsub",	0xfc00003e, 0xfc00003c, D | A | B | C | Rc },
 { "fnmsubs",	0xfc00003e, 0xec00003c, D | A | B | C | Rc },
 { "fres",	0xfc1f07fe, 0xec000030, D | B | Rc },
 { "frsp",	0xfc1f07fe, 0xfc000018, D | B | Rc },
 { "frsqrte",	0xfc1f07fe, 0xfc000034, D | B | Rc },
 { "fsel",	0xfc00003e, 0xfc00002e, D | A | B | C | Rc },
 { "fsqrt",	0xfc1f07fe, 0xfc00002c, D | B | Rc },
 { "fsqrts",	0xfc1f07fe, 0xec00002c, D | B | Rc },
 { "fsub",	0xfc0007fe, 0xfc000028, D | A | B | Rc },
 { "fsubs",	0xfc0007fe, 0xec000028, D | A | B | Rc },
 { "icbi",	0xffe007ff, 0x7c0007ac, A | B },
 { "isync",	0xffffffff, 0x4c00012c },
 { "lbz",	0xfc000000, 0x88000000, D | A | d },
 { "lbzu",	0xfc000000, 0x8c000000, D | A | d },
 { "lbzux",	0xfc0007ff, 0x7c0000ee, D | A | B },
 { "lbzx",	0xfc0007ff, 0x7c0000ae, D | A | B },
 { "ld",	0xfc000000, 0xe8000000, D | A | d },
 { "lfd",	0xfc000000, 0xc8000000, D | A | d },
 { "lfdu",	0xfc000000, 0xcc000000, D | A | d },
 { "lfdux",	0xfc0007ff, 0x7c0004ee, D | A | B },
 { "lfdx",	0xfc0007ff, 0x7c0004ae, D | A | B },
 { "lfs",	0xfc000000, 0xc0000000, D | A | d },
 { "lfsu",	0xfc000000, 0xc4000000, D | A | d },
 { "lfsux",	0xfc0007ff, 0x7c00046e, D | A | B },
 { "lfsx",	0xfc0007ff, 0x7c00042e, D | A | B },
 { "lha",	0xfc000000, 0xa8000000, D | A | d },
 { "lhau",	0xfc000000, 0xac000000, D | A | d },
 { "lhaux",	0xfc0007ff, 0x7c0002ee, D | A | B },
 { "lhax",	0xfc0007ff, 0x7c0002ae, D | A | B },
 { "lhbrx",	0xfc0007ff, 0x7c00062c, D | A | B },
 { "lhz",	0xfc000000, 0xa0000000, D | A | d },
 { "lhzu",	0xfc000000, 0xa4000000, D | A | d },
 { "lhzux",	0xfc0007ff, 0x7c00026e, D | A | B },
 { "lhzx",	0xfc0007ff, 0x7c00022e, D | A | B },
 { "lmw",	0xfc000000, 0xb8000000, D | A | d },
 { "lswi",	0xfc0007ff, 0x7c0004aa, D | A | NB },
 { "lswx",	0xfc0007ff, 0x7c00042a, D | A | B },
 { "lwarx",	0xfc0007ff, 0x7c000028, D | A | B },
 { "lwbrx",	0xfc0007ff, 0x7c00042c, D | A | B },
 { "lwz",	0xfc000000, 0x80000000, D | A | d },
 { "lwzu",	0xfc000000, 0x84000000, D | A | d },
 { "lwzux",	0xfc0007ff, 0x7c00006e, D | A | B },
 { "lwzx",	0xfc0007ff, 0x7c00002e, D | A | B },
 { "mcrf",	0xfc63ffff, 0x4c000000, crfD | crfS },
 { "mcrfs",	0xfc63ffff, 0xfc000080, crfD | crfS },
 { "mcrxr",	0xfc7fffff, 0x7c000400, crfD },
 { "mfcr",	0xfc1fffff, 0x7c000026, D },
 { "mffs",	0xfc1ffffe, 0xfc00048e, D | Rc },
 { "mfmsr",	0xfc1fffff, 0x7c0000a6, D },
 { "mfspr",	0xfc0007ff, 0x7c0002a6, D | SPR },
 { "mfsr",	0xfc10ffff, 0x7c0004a6, D | SR },
 { "mfsrin",	0xfc1f07ff, 0x7c000526, D | B },
 { "mftb",	0xfc1fffff, 0x7c0c42e6, D },	/* simplified mnemonic */
 { "mftbu",	0xfc1fffff, 0x7c0d42e6, D },	/* simplified mnemonic */
/* { "mftb",	0xfc0007ff, 0x7c0002e6, D | tbr }, */
 { "mtcrf",	0xfc100fff, 0x7c000120, S | CRM },
 { "mtfsb0",	0xfc1ffffe, 0xfc00008c, crbD | Rc },
 { "mtfsb1",	0xfc1ffffe, 0xfc00004c, crbD | Rc },
 { "mtfsf",	0xfe0107fe, 0xfc00058e, FM | B | Rc },
 { "mtfsfi",	0xfc7f0ffe, 0xfc00010c, crfD | IMM | Rc },
 { "mtmsr",	0xfc1fffff, 0x7c000124, S },
 { "mtspr",	0xfc0007ff, 0x7c0003a6, S | SPR },
 { "mtsr",	0xfc10ffff, 0x7c0001a4, S | SR },
 { "mtsrin",	0xfc1f07ff, 0x7c0001e4, S | B },
 { "mulhw",	0xfc0007fe, 0x7c000096, D | A | B | Rc },
 { "mulhwu",	0xfc0007fe, 0x7c000016, D | A | B | Rc },
 { "mulli",	0xfc000000, 0x1c000000, D | A | SIMM },
 { "mullw",	0xfc0003fe, 0x7c0001d6, D | A | B | OE | Rc },
 { "nand",	0xfc0007fe, 0x7c0003b8, S | A | B | Rc },
 { "neg",	0xfc00fbfe, 0x7c0000d0, D | A | OE | Rc },
 { "nor",	0xfc0007fe, 0x7c0000f8, S | A | B | Rc },
 { "or",	0xfc0007fe, 0x7c000378, S | A | B | Rc },
 { "orc",	0xfc0007fe, 0x7c000338, S | A | B | Rc },
 { "ori",	0xfc000000, 0x60000000, S | A | UIMM },
 { "oris",	0xfc000000, 0x64000000, S | A | UIMM },
 { "rfi",	0xffffffff, 0x4c000064 },
 { "rlwimi",	0xfc000000, 0x50000000, S | SH | MB | ME | Rc },
 { "rlwinm",	0xfc000000, 0x54000000, S | SH | MB | ME | Rc },
 { "rlwnm",	0xfc000000, 0x5c000000, S | B | MB | ME | Rc },
 { "sc",	0xffffffff, 0x44000002 },
 { "slw",	0xfc0007fe, 0x7c000030, S | A | B | Rc },
 { "sraw",	0xfc0007fe, 0x7c000630, S | A | B | Rc },
 { "srawi",	0xfc0007fe, 0x7c000670, S | A | SH | Rc },
 { "srw",	0xfc0007fe, 0x7c000430, S | A | B | Rc },
 { "stb",	0xfc000000, 0x98000000, S | A | d },
 { "stbu",	0xfc000000, 0x9c000000, S | A | d },
 { "stbux",	0xfc0007ff, 0x7c0001ee, S | A | B },
 { "stbx",	0xfc0007ff, 0x7c0001ae, S | A | B },
 { "std",	0xfc000000, 0xf8000000, D | A | d },
 { "stfd",	0xfc000000, 0xd8000000, S | A | d },
 { "stfdu",	0xfc000000, 0xdc000000, S | A | d },
 { "stfdux",	0xfc0007ff, 0x7c0005ee, S | A | B },
 { "stfdx",	0xfc0007ff, 0x7c0005ae, S | A | B },
 { "stfiwx",	0xfc0007ff, 0x7c0007ae, S | A | B },
 { "stfs",	0xfc000000, 0xd0000000, S | A | d },
 { "stfsu",	0xfc000000, 0xd4000000, S | A | d },
 { "stfsux",	0xfc0007ff, 0x7c00056e, S | A | B },
 { "stfsx",	0xfc0007ff, 0x7c00052e, S | A | B },
 { "sth",	0xfc000000, 0xb0000000, S | A | d },
 { "sthbrx",	0xfc0007ff, 0x7c00072c, S | A | B },
 { "sthu",	0xfc000000, 0xb4000000, S | A | d },
 { "sthux",	0xfc0007ff, 0x7c00036e, S | A | B },
 { "sthx",	0xfc0007ff, 0x7c00032e, S | A | B },
 { "stmw",	0xfc000000, 0xbc000000, S | A | d },
 { "stswi",	0xfc0007ff, 0x7c0005aa, S | A | NB },
 { "stswx",	0xfc0007ff, 0x7c00052a, S | A | B },
 { "stw",	0xfc000000, 0x90000000, S | A | d },
 { "stwbrx",	0xfc0007ff, 0x7c00052c, S | A | B },
 { "stwcx.",	0xfc0007ff, 0x7c00012d, S | A | B },
 { "stwu",	0xfc000000, 0x94000000, S | A | d },
 { "stwux",	0xfc0007ff, 0x7c00016e, S | A | B },
 { "stwx",	0xfc0007ff, 0x7c00012e, S | A | B },
 { "subf",	0xfc0003fe, 0x7c000050, D | A | B | OE | Rc },
 { "subfc",	0xfc0003fe, 0x7c000010, D | A | B | OE | Rc },
 { "subfe",	0xfc0003fe, 0x7c000110, D | A | B | OE | Rc },
 { "subfic",	0xfc000000, 0x20000000, D | A | SIMM },
 { "subfme",	0xfc00fbfe, 0x7c0001d0, D | A | OE | Rc },
 { "subfze",	0xfc00fbfe, 0x7c000190, D | A | OE | Rc },
 { "sync",	0xffffffff, 0x7c0004ac },
 { "tlbia",	0xffffffff, 0x7c0002e4 },
 { "tlbie",	0xffff07ff, 0x7c000264, B },
 { "tlbld",	0xffff07ff, 0x7c0007a4, B },	/* 603 only */
 { "tlbli",	0xffff07ff, 0x7c0007e4, B },	/* 603 only */
 { "tlbsync",	0xffffffff, 0x7c00046c },
 { "tw",	0xfc0007ff, 0x7c000008, TO | A | B },
 { "twi",	0xfc000000, 0x0c000000, TO | A | SIMM },
 { "xor",	0xfc0007fe, 0x7c000278, S | A | B | Rc },
 { "xori",	0xfc000000, 0x68000000, S | A | UIMM },
 { "xoris",	0xfc000000, 0x6c000000, S | A | UIMM },
 { "" }
};

struct sprcode {
	u_int code;
	char name[8];
};

struct sprcode sprcodes[] = {
 {   1, "xer" },
 {   8, "lr" },
 {   9, "ctr" },
 {  18, "dsisr" },
 {  19, "dar" },
 {  22, "dec" },
 {  25, "sdr1" },
 {  26, "srr0" },
 {  27, "srr1" },
 { 272, "sprg0" },
 { 273, "sprg1" },
 { 274, "sprg2" },
 { 275, "sprg3" },
 { 280, "asr" },
 { 282, "ear" },
 { 284, "tbl" },
 { 285, "tbu" },
 { 287, "pvr" },
 { 528, "ibat0u" },
 { 529, "ibat0l" },
 { 530, "ibat1u" },
 { 531, "ibat1l" },
 { 532, "ibat2u" },
 { 533, "ibat2l" },
 { 534, "ibat3u" },
 { 535, "ibat3l" },
 { 536, "dbat0u" },
 { 537, "dbat0l" },
 { 538, "dbat1u" },
 { 539, "dbat1l" },
 { 540, "dbat2u" },
 { 541, "dbat2l" },
 { 542, "dbat3u" },
 { 543, "dbat3l" },
#if 0
 { 976, "dmiss" },
 { 977, "dcmp" },
 { 978, "hash1" },
 { 979, "hash2" },
 { 980, "imiss" },
 { 981, "icmp" },
 { 982, "rpa" },
 { 1008, "hid0" },
 { 1010, "iabr" },
#endif
 { 0, "" }
};

char *
sprname(spr)
	int spr;
{
	struct sprcode *s = sprcodes;

	while (s->name[0]) {
		if (s->code == spr)
			return s->name;
		s++;
	}
	return NULL;
}

inline u_int
extract_field(value, base, width)
	u_int value;
	int base, width;
{
	u_int mask = (1U << width) - 1;

	return (value >> (32 - base - width)) & mask;
}

db_addr_t
db_disasm(loc, extended)
	db_addr_t loc;
	bool extended;
{
	const struct opcode *op = opcodes;
	u_int code = *(u_int *)loc;
	int flags;

	while (op->name[0]) {
		if ((code & op->mask) == op->code)
			break;
		op++;
	}

	if (op->name[0] == 0) {
		db_printf(".long	0x%08x", code);
		goto out;
	}

	/* special cases */
	if ((code & 0xfc000000) == 0x48000000) {		/* b */
		int aa = extract_field(code, 30, 1);
		int lk = extract_field(code, 31, 1);
		int li = extract_field(code, 6, 24);

		db_printf("%s", op->name);
		if (lk) db_printf("l");
		if (aa) db_printf("a");
		db_printf("\t");
		if (li & 0x800000)
			li -= 0x1000000;
		if (aa)
			db_printf("0x%x", li * 4);
		else
			db_printf("0x%lx", loc + li * 4);
		goto out;

	} else if ((code & 0xfc00fbfe) == 0x4c000020) {		/* bclr/ctr */
		int lk = extract_field(code, 31, 1);
		int bo = extract_field(code, 6, 5);
		int bi = extract_field(code, 11, 5);

		db_printf("%s", op->name);
		if (lk) db_printf("l");
		if (!(bo == 20 && bi == 0))
			db_printf("	%d,%d", bo, bi);
		goto out;

	} else if ((code & 0xfc000000) == 0x40000000) {		/* bc */
		int aa = extract_field(code, 30, 1);
		int lk = extract_field(code, 31, 1);
		int bo = extract_field(code, 6, 5);
		int bi = extract_field(code, 11, 5);
		int bd = extract_field(code, 16, 14);

		if (bd & 0x2000)
			bd -= 0x4000;

		db_printf("%s", op->name);
		if (lk) db_printf("l");
		if (aa) db_printf("a");
		if (bo & 1) {		/* XXX bo == 21 is illegal. */
			if (bd < 0)
				db_printf("-");
			else
				db_printf("+");
		}
		db_printf("\t");

		if (op->flags & BO)
			db_printf("%d,", bo);
		if (op->flags & BI)
			db_printf("%d,", bi);

		if (aa)
			db_printf("0x%x", bd * 4);
		else
			db_printf("0x%lx", loc + bd * 4);
		goto out;

	} else if ((code & 0xfc000000) == 0x50000000 ||		/* rlwimi */
		   (code & 0xfc000000) == 0x54000000 ||		/* rlwinm */
		   (code & 0xfc000000) == 0x5c000000) {		/* rlwnm */
		int rc = extract_field(code, 31, 1);

		db_printf("%s", op->name);
		if (rc) db_printf(".");
		db_printf("\t");
		db_printf("%d,%d,%d,%d,%d",
			extract_field(code, 11, 5),	/* A */
			extract_field(code, 6, 5),	/* S */
			extract_field(code, 16, 5),	/* SH or B */
			extract_field(code, 21, 5),	/* MB */
			extract_field(code, 26, 5));	/* ME */
		goto out;

	} else if ((code & 0xfc0007ff) == 0x7c0002a6) {		/* mfspr */
		int rD = extract_field(code, 6, 5);
		int h = extract_field(code, 16, 5);
		int l = extract_field(code, 11, 5);
		int spr = h << 5 | l;
		char *name = sprname(spr);

		if (name == NULL) {
			db_printf("%s	", op->name);
			db_printf("%d,%d", rD, spr);
			goto out;
		}

		/* [id]bat[0-3][ul] */
		if (strncmp(name + 1, "bat", 3) == 0) {
			db_printf("mf%cbat%c	%d,%c",
			    name[0], name[5], rD, name[4]);
			goto out;
		}

		/* sprg[0-3] */
		if (strncmp(name, "sprg", 4) == 0) {
			db_printf("mfsprg	%d,%c", rD, name[4]);
			goto out;
		}

		db_printf("mf%s	%d", name, rD);
		goto out;

	} else if ((code & 0xfc0007ff) == 0x7c0003a6) {		/* mtspr */
		int rS = extract_field(code, 6, 5);
		int h = extract_field(code, 16, 5);
		int l = extract_field(code, 11, 5);
		int spr = h << 5 | l;
		char *name = sprname(spr);

		if (name == NULL) {
			db_printf("%s	", op->name);
			db_printf("%d,%d", spr, rS);
			goto out;
		}

		/* [id]bat[0-3][ul] */
		if (strncmp(name + 1, "bat", 3) == 0) {
			db_printf("mt%cbat%c	%c,%d",
			    name[0], name[5], name[4], rS);
			goto out;
		}

		/* sprg[0-3] */
		if (strncmp(name, "sprg", 4) == 0) {
			db_printf("mtsprg	%c,%d", name[4], rS);
			goto out;
		}

		db_printf("mt%s	%d", name, rS);
		goto out;

	} else if ((code & 0xfc10ffff) == 0x7c0004a6) {		/* mfsr */
		int rD = extract_field(code, 6, 5);
		int sr = extract_field(code, 12, 4);

		db_printf("%s	", op->name);
		db_printf("%d,%d", rD, sr);
		goto out;

	} else if ((code & 0xfc10ffff) == 0x7c0001a4) {		/* mtsr */
		int rS = extract_field(code, 6, 5);
		int sr = extract_field(code, 12, 4);

		db_printf("%s	", op->name);
		db_printf("%d,%d", sr, rS);
		goto out;

	} else if ((code & 0xfe0107fe) == 0xfc00058e) {		/* mtfsf */
		int fm = extract_field(code, 7, 8);
		int frB = extract_field(code, 16, 5);
		int rc = extract_field(code, 31, 1);

		db_printf("%s", op->name);
		if (rc) db_printf(".");
		db_printf("\t");
		db_printf("%d,%d", fm, frB);
		goto out;

	} else if ((code & 0xfc7f0ffe) == 0xfc00010c) {		/* mtfsfi */
		int crfd = extract_field(code, 6, 3);
		int imm = extract_field(code, 16, 4);
		int rc = extract_field(code, 31, 1);

		db_printf("%s", op->name);
		if (rc) db_printf(".");
		db_printf("\t");
		db_printf("%d,%d", crfd, imm);
		goto out;

	} else if ((code & 0xfc100fff) == 0x7c000120) {		/* mtcrf */
		int crm = extract_field(code, 12, 8);
		int rS = extract_field(code, 6, 5);

		if (crm == 0xff)
			db_printf("mtcr	%d", rS);
		else {
			db_printf("%s	", op->name);
			db_printf("%d,%d", crm, rS);
		}
		goto out;

	} else if ((code & 0xfc0007fe) == 0x7c000378) {		/* mr */
		int rS = extract_field(code, 6, 5);
		int rA = extract_field(code, 11, 5);
		int rB = extract_field(code, 16, 5);

		if (rS == rB) {
			db_printf("mr	%d,%d", rA, rS);
			goto out;
		}
	}

	db_printf(op->name);

	if (op->flags & OE)
		if (extract_field(code, 21, 1))
			db_printf("o");
	if (op->flags & Rc)
		if (extract_field(code, 31, 1))
			db_printf(".");

	db_printf("\t");

	flags = op->flags;

	if (flags & d) {				/* disp(rA) */
		int rA = extract_field(code, 11, 5);
		int disp = extract_field(code, 16, 16);

		if (disp & 0x8000)
			disp -= 0x10000;
		db_printf("%d,%d(%d)", extract_field(code, 6, 5), disp, rA);
		goto out;
	}

	if (flags & crfD)
		db_printf("%d", extract_field(code, 6, 3));
	if (flags & crfS)
		db_printf(",%d", extract_field(code, 11, 3));
	if ((flags & (S|A|B)) == (S|A|B)) {
		int rS = extract_field(code, 6, 5);
		int rA = extract_field(code, 11, 5);
		int rB = extract_field(code, 16, 5);

		db_printf("%d,%d,%d", rS, rA, rB);
		goto out;
	}
	if ((flags & (A|S)) == (A|S))
		db_printf("%d,", extract_field(code, 11, 5));	/* rA */
	if (flags & S)
		db_printf("%d", extract_field(code, 6, 5));
	if (flags & D)
		db_printf("%d", extract_field(code, 6, 5));
	if ((flags & (A|S)) == A) {
		if (flags & (crfD|crfS|S|D)) db_printf(",");
		db_printf("%d", extract_field(code, 11, 5));	/* rA */
	}
	if (flags & C)
		db_printf(",%d", extract_field(code, 21, 5));
	if (flags & B) {
		if (flags & (crfD|crfS|S|D|A)) db_printf(",");
		db_printf("%d", extract_field(code, 16, 5));
	}
	if (flags & SIMM) {
		int simm = extract_field(code, 16, 16);

		if (simm & 0x8000)
			simm -= 0x10000;
		db_printf(",%d", simm);
	}
	if (flags & UIMM)
		db_printf(",%d", extract_field(code, 16, 16));

out:
	db_printf("\n");
	return loc + 4;
}
