/*	$Id: netbsd32_machdep.c,v 1.14 2011/10/04 07:28:17 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_altivec.h"
#include "opt_compat_netbsd.h"
#include "opt_multiprocessor.h"

#include <sys/param.h>
#include <sys/exec.h>
#include <sys/ioctl.h>
#include <sys/kmem.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/ucontext.h>

#include <uvm/uvm_extern.h>

#include <compat/netbsd32/netbsd32.h>
#include <compat/netbsd32/netbsd32_syscallargs.h>
#include <compat/netbsd32/netbsd32_exec.h>

#include <machine/altivec.h>
#include <machine/fpu.h>
#include <machine/frame.h>
#include <machine/pcb.h>
#include <machine/vmparam.h>

/*
 * Set up registers on exec.
 */
void
netbsd32_setregs(l, pack, stack)
	struct lwp *l;
	struct exec_package *pack;
	u_long stack;
{
	struct proc *p = l->l_proc;
	struct trapframe *tf = trapframe(l);
	struct ps_strings arginfo;

	/* Mark this as a 32-bit emulation */
	p->p_flag |= PK_32;

	netbsd32_adjust_limits(p);

	bzero(tf, sizeof *tf);
	tf->fixreg[1] = -roundup(-stack + 8, 16);

	/*
	 * XXX Machine-independent code has already copied arguments and
	 * XXX environment to userland.  Get them back here.
	 */
	copyin_psstrings(p, &arginfo);

	/*
	 * Set up arguments for _start():
	 *	_start(argc, argv, envp, obj, cleanup, ps_strings);
	 *
	 * Notes:
	 *	- obj and cleanup are the auxiliary and termination
	 *	  vectors.  They are fixed up by ld.elf_so.
	 *	- ps_strings is a NetBSD extension, and will be
	 *	  ignored by executables which are strictly
	 *	  compliant with the SVR4 ABI.
	 *
	 * XXX We have to set both regs and retval here due to different
	 * XXX calling convention in trap.c and init_main.c.
	 */
	tf->fixreg[3] = arginfo.ps_nargvstr;
	tf->fixreg[4] = (register_t)arginfo.ps_argvstr;
	tf->fixreg[5] = (register_t)arginfo.ps_envstr;
	tf->fixreg[6] = 0;			/* auxiliary vector */
	tf->fixreg[7] = 0;			/* termination vector */
	tf->fixreg[8] = p->p_psstrp;		/* NetBSD extension */

	tf->srr0 = pack->ep_entry;
	tf->srr1 = PSL_MBO | PSL_USERSET | PSL_FE_DFLT;
	/* tf->srr1 &= ~PSL_SF; */

	((struct pcb *)l->l_addr)->pcb_flags = 0;
}

struct netbsd32_sigframe {
	u_int sf_sp;
	u_int sf_lr;
	siginfo32_t sf_si;
	ucontext32_t sf_uc;
};

void
netbsd32_sendsig(ksi, mask)
	const ksiginfo_t *ksi;
	const sigset_t *mask;
{
	int sig = ksi->ksi_signo;
	struct lwp *l = curlwp;
	struct proc *p = l->l_proc;
	struct sigacts *ps = p->p_sigacts;
	struct trapframe *tf;
	struct netbsd32_sigframe *fp, frame;
	int onstack, error;
	sig_t catcher = SIGACTION(p, sig).sa_handler;

#ifdef COMPAT_16
#error COMPAT_16
	if (ps->sa_sigdesc[ksi->ksi_signo].sd_vers < 2) {
		sendsig_sigcontext(sig, mask, KSI_TRAPCODE(ksi));
		return;
	}
#endif /* COMPAT_16 */

	tf = trapframe(l);

	/* Do we need to jump onto the signal stack? */
	onstack =
	    (l->l_sigstk.ss_flags & (SS_DISABLE | SS_ONSTACK)) == 0 &&
	    (SIGACTION(p, sig).sa_flags & SA_ONSTACK) != 0;

	/* Allocate space for the signal handler context. */
	if (onstack)
		fp = (struct netbsd32_sigframe *)((caddr_t)l->l_sigstk.ss_sp +
						l->l_sigstk.ss_size);
	else
		fp = (struct netbsd32_sigframe *)tf->fixreg[1];
	fp = (struct netbsd32_sigframe *)((long)(fp - 1) & ~0xf);

	netbsd32_si_to_si32(&frame.sf_si, (const siginfo_t *)&ksi->ksi_info);
	frame.sf_uc.uc_flags = _UC_SIGMASK;
#ifdef _UC_SETSTACK
	frame.sf_uc.uc_flags |= (l->l_sigstk.ss_flags & SS_ONSTACK)
	    ? _UC_SETSTACK : _UC_CLRSTACK;
#endif
	frame.sf_uc.uc_sigmask = *mask;
	frame.sf_uc.uc_link = 0;
	bzero(&frame.sf_uc.uc_stack, sizeof frame.sf_uc.uc_stack);
	cpu_getmcontext32(l, &frame.sf_uc.uc_mcontext, &frame.sf_uc.uc_flags);

	sendsig_reset(l, sig);

	mutex_exit(p->p_lock);
	error = copyout(&frame, fp, sizeof frame);
	mutex_enter(p->p_lock);

	if (error != 0) {
		/*
		 * Process has trashed its stack; give it an illegal
		 * instructoin to halt it in its tracks.
		 */
		sigexit(l, SIGILL);
		/* NOTREACHED */
	}

	/*
	 * Build context to run handler in.  Note the trampoline version
	 * numbers are coordinated with machine-dependent code in libc.
	 */
	switch (ps->sa_sigdesc[sig].sd_vers) {
	case 2:
		tf->fixreg[1] = (register_t)fp;
		tf->srr0 = (register_t)catcher;
		tf->fixreg[3] = (register_t)sig;
		tf->fixreg[4] = (register_t)&fp->sf_si;
		tf->fixreg[5] = (register_t)&fp->sf_uc;
		tf->fixreg[31] = tf->fixreg[5];
		tf->lr = (register_t)ps->sa_sigdesc[sig].sd_tramp;
		break;

	default:
		/* Don't know what trampoline version; kill it. */
		sigexit(l, SIGILL);
	}

	/* Remember that we're now on the signal stack. */
	if (onstack)
		l->l_sigstk.ss_flags |= SS_ONSTACK;
}

int
cpu_coredump32(l, iocookie, chdr)
	struct lwp *l;
	void *iocookie;
	struct core32 *chdr;
{
	printf("cpu_coredump32\n");
	return 0;
}

#if 0
int
netbsd32_md_ioctl(fp, cmd, data32, l)
	struct file *fp;
	netbsd32_u_long cmd;
	void *data32;
	struct lwp *l;
{
	printf("netbsd32_md_ioctl\n");

	return (*fp->f_ops->fo_ioctl)(fp, cmd, data32, l);
}
#endif

int
cpu_setmcontext32(l, mcp, flags)
	struct lwp *l;
	const mcontext32_t *mcp;
	u_int flags;
{
	struct trapframe *tf = trapframe(l);
	const __gregset32_t *gr = &mcp->__gregs;
	struct pcb *pcb = l->l_addr;
	int i;

	/* Restore GPR context, if any. */
	if (flags & _UC_CPU) {
		for (i = 0; i < 32; i++)
			tf->fixreg[i] = gr->__r[i];
		tf->cr   = gr->__r_cr;
		tf->lr   = gr->__r_lr;
		tf->srr0 = gr->__r_pc;
		tf->srr1 &= ~(PSL_FE0|PSL_FE1);
		tf->srr1 |= gr->__r_msr & (PSL_FE0|PSL_FE1);
		tf->ctr  = gr->__r_ctr;
		tf->xer  = gr->__r_xer;
		/* unused = gr->__r_mq; */
	}

	/* Restore FPR context, if any. */
	if ((flags & _UC_FPU) && mcp->__fpregs.__fpu_valid) {
		/* XXX we don't need to save the state, just to drop it */
		if (fpulwp)
			save_fpu(fpulwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_fpcpu)
			save_fpu_lwp(l);
#endif
		memcpy(pcb->pcb_fpu.fpreg, mcp->__fpregs.__fpu_regs,
		    sizeof pcb->pcb_fpu.fpreg);
		*(u_int64_t *)&pcb->pcb_fpu.fpscr = mcp->__fpregs.__fpu_fpscr;
		pcb->pcb_flags |= PCB_FPU;
		tf->srr1 &= ~PSL_FP;
	}

#ifdef ALTIVEC
	/* Restore VEC context, if any. */
	if ((flags & _UC_POWERPC_VEC) && mcp->__vrf.__vrsave) {
		/* XXX we don't need to save the state, just to drop it */
		if (veclwp)
			save_vec(veclwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_veccpu)
			save_vec_lwp(l);
#endif
		memcpy(pcb->pcb_vec.vreg, mcp->__vrf.__vr,
		    sizeof pcb->pcb_vec.vreg);
		pcb->pcb_vec.vscr[3] = mcp->__vrf.__vscr;
		pcb->pcb_vec.vrsave = mcp->__vrf.__vrsave;
		pcb->pcb_flags |= PCB_VEC;
		tf->srr1 &= ~PSL_VEC;
	}
#endif

#ifdef _UC_SETSTACK
	if (flags & _UC_SETSTACK)
		l->l_sigstk.ss_flags |= SS_ONSTACK;
	if (flags & _UC_CLRSTACK)
		l->l_sigstk.ss_flags &= ~SS_ONSTACK;
#endif
	return 0;
}

void
cpu_getmcontext32(l,mcp, flagp)
	struct lwp *l;
	mcontext32_t *mcp;
	u_int *flagp;
{
	const struct trapframe *tf = trapframe(l);
	__gregset32_t *gr = &mcp->__gregs;
	struct pcb *pcb = l->l_addr;
	int i;

	/* Save GPR context. */
	for (i = 0; i < 32; i++)			/* r0-31 */
		gr->__r[i] = tf->fixreg[i];
	gr->__r_cr  = tf->cr;
	gr->__r_lr  = tf->lr;
	gr->__r_pc  = tf->srr0;
	gr->__r_msr = tf->srr1;				/* XXX */
	gr->__r_ctr = tf->ctr;
	gr->__r_xer = tf->xer;
	gr->__r_mq  = 0;				/* For now. */
	*flagp |= _UC_CPU;

	/* Save FPR context, if any. */
	if (pcb->pcb_flags & PCB_FPU) {
		/* If the FPU has valid data, dump its context to the PCB first. */
		if (fpulwp)
			save_fpu(fpulwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_fpcpu)
			save_fpu_lwp(l);
#endif
		memcpy(mcp->__fpregs.__fpu_regs, pcb->pcb_fpu.fpreg,
		    sizeof mcp->__fpregs.__fpu_regs);
		mcp->__fpregs.__fpu_fpscr = *(u_int64_t *)&pcb->pcb_fpu.fpscr;
		mcp->__fpregs.__fpu_valid = 1;
		*flagp |= _UC_FPU;
	} else
		mcp->__fpregs.__fpu_valid = 0;

#ifdef ALTIVEC
	/* Save VEC context, if any. */
	if (pcb->pcb_flags & PCB_VEC) {
		/* If the vector unit has valid data, dump its context to the PCB first. */
		if (veclwp)
			save_vec(veclwp);
#ifdef MULTIPROCESSOR
		if (pcb->pcb_veccpu)
			save_vec_lwp(l);
#endif
		memcpy(mcp->__vrf.__vr, pcb->pcb_vec.vreg,
		    sizeof mcp->__vrf.__vr);
		mcp->__vrf.__vscr = pcb->pcb_vec.vscr[3];
		mcp->__vrf.__vrsave = pcb->pcb_vec.vrsave;
		*flagp |= _UC_POWERPC_VEC;
	} else
#endif
		mcp->__vrf.__vrsave = 0;
}

void
startlwp32(arg)
	void *arg;
{
	int err;
	ucontext32_t *uc = arg;
	struct lwp *l = curlwp;
	struct trapframe *frame = trapframe(l);

	err = cpu_setmcontext32(l, &uc->uc_mcontext, uc->uc_flags);
#ifdef DIAGNOSTIC
	if (err)
		printf("Error %d from cpu_setmcontext.", err);
#endif
	/* Note: we are freeing ucontext_t, not ucontext32_t. */
	kmem_free(uc, sizeof(ucontext_t));
	userret(l, frame);
}

vaddr_t
netbsd32_vm_default_addr(p, base, size)
	struct proc *p;
	vaddr_t base;
	vsize_t size;
{
	return round_page(base + MAXDSIZ);
}

int
netbsd32_sysarch(l, v, retval)
	struct lwp *l;
	const struct netbsd32_sysarch_args *v;
	register_t *retval;
{
	printf("netbsd32_sysarch\n");
	return EINVAL;
}

/* XXX FIXME */
void
netbsd32_cpu_upcall(l, type, nevents, ninterrupted, sas, ap, sp, upcall)
	struct lwp *l;
	int type, nevents, ninterrupted;
	void *sas, *ap, *sp;
	sa_upcall_t upcall;
{
	struct trapframe *tf;

	tf = trapframe(l);

	/*
	 * Build context to run handler in.
	 */
	tf->fixreg[1] = (register_t)((struct saframe *)sp - 1);
	tf->lr = 0;
	tf->fixreg[3] = type;
	tf->fixreg[4] = (register_t)sas;
	tf->fixreg[5] = nevents;
	tf->fixreg[6] = ninterrupted;
	tf->fixreg[7] = (register_t)ap;
	tf->srr0 = (register_t)upcall;
}
