/*	$Id: netbsd32_syscall.c,v 1.9 2011/01/18 03:17:55 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_altivec.h"
#include "opt_compat_linux.h"
#include "opt_ktrace.h"
#include "opt_multiprocessor.h"
#include "opt_sa.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/ktrace.h>
#include <sys/proc.h>
#include <sys/sa.h>
#include <sys/savar.h>
#include <sys/syscall.h>

#include <uvm/uvm_extern.h>

void netbsd32_syscall_intern(struct proc *);
void netbsd32_syscall_plain(struct trapframe *);
void netbsd32_syscall_fancy(struct trapframe *);

/* These definitions should probably be somewhere else			XXX */
#define FIRSTARG	3		/* first argument is in reg 3 */
#define NARGREG		8		/* 8 args are in registers */
#define MOREARGS(sp)	((caddr_t)(sp) + 8) /* more args go here */

void
netbsd32_syscall_intern(p)
	struct proc *p;
{

	if (trace_is_enabled(p)) {
		p->p_md.md_syscall = netbsd32_syscall_fancy;
		return;
	}
	p->p_md.md_syscall = netbsd32_syscall_plain;
}

void
netbsd32_syscall_plain(frame)
	struct trapframe *frame;
{
	struct lwp *l = curlwp;
	struct proc *p = l->l_proc;
	const struct sysent *callp;
	size_t argsize;
	register_t code;
	register_t rval[2];
	u_int *params;
	u_int args[12];
	int n, error;

	LWP_CACHE_CREDS(l, p);

	curcpu()->ci_data.cpu_nsyscall++;

	for (n = 0; n < 8; n++)
		args[n] = frame->fixreg[3 + n];

	code = frame->fixreg[0];
	callp = p->p_emul->e_sysent;
	params = args;
	n = NARGREG;

#ifdef KERN_SA
	if (__predict_false((l->l_savp)
	    && (l->l_savp->savp_pflags & SAVP_FLAG_DELIVERING)))
		l->l_savp->savp_pflags &= ~SAVP_FLAG_DELIVERING;
#endif

	switch (code) {
	case SYS_syscall:
		/*
		 * code is first argument,
		 * followed by actual args.
		 */
		code = *params++;
		n -= 1;
		break;
	case SYS___syscall:
		params++;
		code = *params++;
		n -= 2;
		break;
	}

	code &= (SYS_NSYSENT - 1);
	callp += code;
	argsize = callp->sy_argsize;

	if (argsize > n * sizeof(int)) {
		error = copyin(MOREARGS(frame->fixreg[1]), &args[8],
			       argsize - n * sizeof(int));
		if (error)
			goto syscall_bad;
	}

	rval[0] = 0;
	rval[1] = 0;

	error = (*callp->sy_call)(l, params, rval);
	switch (error) {
	case 0:
		frame->fixreg[FIRSTARG] = rval[0];
		frame->fixreg[FIRSTARG + 1] = rval[1];
		frame->cr &= ~0x10000000;
		break;
	case ERESTART:
		/*
		 * Set user's pc back to redo the system call.
		 */
		frame->srr0 -= 4;
		break;
	case EJUSTRETURN:
		/* nothing to do */
		break;
	default:
syscall_bad:
		if (p->p_emul->e_errno)
			error = p->p_emul->e_errno[error];
		frame->fixreg[FIRSTARG] = error;
		frame->cr |= 0x10000000;
		break;
	}

	userret(l, frame);
}

void
netbsd32_syscall_fancy(frame)
	struct trapframe *frame;
{
	struct lwp *l = curlwp;
	struct proc *p = l->l_proc;
	const struct sysent *callp;
	size_t argsize;
	register_t code;
	register_t rval[2];
	u_int *params;
	u_int args[12];
	int n, error;

	LWP_CACHE_CREDS(l, p);

	curcpu()->ci_data.cpu_nsyscall++;

	for (n = 0; n < 8; n++)
		args[n] = frame->fixreg[3 + n];

	code = frame->fixreg[0];
	callp = p->p_emul->e_sysent;
	params = args;
	n = NARGREG;

#ifdef KERN_SA
	if (__predict_false((l->l_savp)
	    && (l->l_savp->savp_pflags & SAVP_FLAG_DELIVERING)))
		l->l_savp->savp_pflags &= ~SAVP_FLAG_DELIVERING;
#endif

	switch (code) {
	case SYS_syscall:
		/*
		 * code is first argument,
		 * followed by actual args.
		 */
		code = *params++;
		n -= 1;
		break;
	case SYS___syscall:
		params++;
		code = *params++;
		n -= 2;
		break;
	}

	code &= (SYS_NSYSENT - 1);
	callp += code;
	argsize = callp->sy_argsize;

	if (argsize > n * sizeof(int)) {
		error = copyin(MOREARGS(frame->fixreg[1]), &args[8],
			       argsize - n * sizeof(int));
		if (error)
			goto syscall_bad;
	}

#if 0
	/* XXX ɤΤǤΤ */
	if ((error = trace_enter(code, params64, callp->sy_narg)) != 0)
		goto out;
#endif

	rval[0] = 0;
	rval[1] = 0;

	error = (*callp->sy_call)(l, params, rval);
/*out:*/
	switch (error) {
	case 0:
		frame->fixreg[FIRSTARG] = rval[0];
		frame->fixreg[FIRSTARG + 1] = rval[1];
		frame->cr &= ~0x10000000;
		break;
	case ERESTART:
		/*
		 * Set user's pc back to redo the system call.
		 */
		frame->srr0 -= 4;
		break;
	case EJUSTRETURN:
		/* nothing to do */
		break;
	default:
syscall_bad:
		if (p->p_emul->e_errno)
			error = p->p_emul->e_errno[error];
		frame->fixreg[FIRSTARG] = error;
		frame->cr |= 0x10000000;
		break;
	}

#if 0
	/* XXX ɤΤǤΤ */
	trace_exit(code, rval, error);
#endif

	userret(l, frame);
}
