/*	$Id: ehci_obio.c,v 1.11 2011/10/20 06:12:22 tsubai Exp $	*/

/*-
 * Copyright (c) 2006 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/device.h>
#include <sys/lock.h>

#define _PS3_BUS_DMA_PRIVATE
#include <machine/bus.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdivar.h>
#include <dev/usb/usb_mem.h>

#include <dev/usb/ehcireg.h>
#include <dev/usb/ehcivar.h>

struct ehci_obio_softc {
	ehci_softc_t sc;
};

int ehci_obio_match(struct device *, struct cfdata *, void *);
void ehci_obio_attach(struct device *, struct device *, void *);

extern struct device *ohci_companion;		/* XXX */

CFATTACH_DECL2_NEW(ehci_obio, sizeof(struct ehci_obio_softc), ehci_obio_match,
    ehci_obio_attach, NULL, ehci_activate, NULL, ehci_childdet);

struct ps3_bus_dma_tag ehci_bus_dma_tag = {
	0,			/* _bounce_thresh */
	0,			/* _dma_base */
	_bus_dmamap_create,
	_bus_dmamap_destroy,
	_bus_dmamap_load,
	_bus_dmamap_load_mbuf,
	_bus_dmamap_load_uio,
	_bus_dmamap_load_raw,
	_bus_dmamap_unload,
	NULL,			/* _dmamap_sync */
	_bus_dmamem_alloc,
	_bus_dmamem_free,
	_bus_dmamem_map,
	_bus_dmamem_unmap,
	_bus_dmamem_mmap
};

int
ehci_obio_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct confargs *ca = aux;

	if (ca->ca_type == 4 && ca->ca_func == 1)
		return 5;
	return 0;
}

void
ehci_obio_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct ehci_obio_softc *sc = device_private(self);
	struct confargs *ca = aux;
	const char *devname = self->dv_xname;
	u_long bus_id = ca->ca_bus;
	u_long dev_id = ca->ca_dev;
	u_long func = ca->ca_func;
	u_long bus_index = bus_id;	/* XXX */
	u_long device_index = dev_id;	/* XXX */
	u_long region_addr, region_size, region_type, addr, dma_base;
	int irq, err;

#ifdef UCONS
	if (ca->ca_bus == 1 && ca->ca_dev == 1) {
		printf(": used by console\n");
		return;
	}
#endif
	err = HV_open(bus_id, dev_id, 0);
	if (err == HV_EBUSY)
		err = 0;
	if (err) {
		printf(": couldn't open device\n");
		return;
	}

	if (ps3_read_region_info(bus_index, device_index, func, 
	    &region_addr, &region_size, &region_type)) {
		printf(": read_region_data failed\n");
		return;
	}

	err = HV_map_device_mmio_region(bus_id, dev_id, region_addr,
	    region_size, HV_PAGESIZE_4K, &addr);
	if (err) {
		printf(": map_device_mmio_region failed (%d)\n", err);
		return;
	}

	dma_base = ps3_allocate_dma_region(bus_id, dev_id, 1);
	if (dma_base == 0) {
		printf(": couldn't map dma region\n");
		return;
	}

	sc->sc.sc_size = region_size;
	sc->sc.iot = 0;
	bus_space_map(0, addr, region_size, 0, &sc->sc.ioh);

	/* Disable interrupts, so we don't get any spurious ones. */
	sc->sc.sc_offs = EREAD4(&sc->sc, EHCI_CAPLENGTH) & 0xff;
	EOWRITE2(&sc->sc, EHCI_USBINTR, 0);

	sc->sc.sc_bus.usbrev = USBREV_2_0;

	ehci_bus_dma_tag._dma_base = dma_base;
	sc->sc.sc_bus.dmatag = &ehci_bus_dma_tag;

	irq = ps3_connect_io_irq(ca->ca_intr);
	if (irq < 0) {
		printf(": connect_io_irq failed\n");
		return;
	}

	intr_establish(irq, IST_LEVEL, IPL_USB, ehci_intr, sc);
	printf(": irq %d\n", irq);

	/* Figure out vendor for root hub descriptor. */
	sc->sc.sc_id_vendor = 0;
	strcpy(sc->sc.sc_vendor, "PS3");
	sc->sc.sc_dev = self;
	sc->sc.sc_bus.hci_private = sc;

	if (ohci_companion) {
		sc->sc.sc_comps[0] = ohci_companion;
		sc->sc.sc_ncomp = 1;
		ohci_companion = NULL;
	}

	err = ehci_init(&sc->sc);
	if (err != USBD_NORMAL_COMPLETION) {
		printf("%s: ehci_init failed (%d)\n", devname, err);
		return;
	}

	sc->sc.sc_child = config_found(self, &sc->sc.sc_bus, usbctlprint);
}
