/*	$Id: flash.c,v 1.9 2011/10/20 07:36:22 tsubai Exp $	*/

/*-
 * Copyright (c) 2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

#include "ioconf.h"

void flash_attach __P((struct device *, struct device *, void *));
int flash_match __P((struct device *, struct cfdata *, void *));

struct flash_softc {
	struct device sc_dev;
	int sc_devid;
	int sc_region;
	u_int sc_size;		/* storage size */
	u_int sc_blk;		/* block size */
};

CFATTACH_DECL0(flash_obio, sizeof(struct flash_softc), flash_match, flash_attach);

dev_type_read(flashread);
dev_type_write(flashwrite);
dev_type_mmap(flashmmap);

const struct cdevsw flash_cdevsw = {
	nullopen, nullclose, flashread, flashwrite, noioctl,
	nostop, notty, nopoll, flashmmap, nokqfilter
};
struct bdevsw flash_bdevsw = {};

#define FLASH_BSIZE (256*1024)

static char flash_buf[FLASH_BSIZE];

int
flash_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (ca->ca_bus != 4)
		return 0;
	if (ca->ca_dev != 1)
		return 0;

	return 10;
}

void
flash_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct flash_softc *sc = (struct flash_softc *)self;
	struct confargs *ca = aux;
	int error;

	sc->sc_devid = ca->ca_dev;
	sc->sc_region = 5;

	error = HV_open(ca->ca_bus, ca->ca_dev, 0);
	if (error && error != HV_EBUSY) {
		printf(": open failed\n");
		return;
	}

	ps3_allocate_dma_region(ca->ca_bus, ca->ca_dev, 0);

	printf("\n");
	sc->sc_size = 4096*1024;	/* XXX 4MB */
	sc->sc_blk = 512;		/* XXX 512 bytes/block */
}

int
flashread(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct flash_softc *sc;
	struct iovec *iov;
	u_int64_t tag, status;
	u_int off, len;
	int rv, error;

	sc = (void *)device_lookup(&flash_cd, minor(dev));
	if (sc == NULL)
		return ENXIO;

	while (uio->uio_resid > 0) {
		iov = uio->uio_iov;
		if (iov->iov_len == 0) {
			uio->uio_iov++;
			uio->uio_iovcnt--;
			if (uio->uio_iovcnt < 0)
				panic("flashread");
			continue;
		}

		off = uio->uio_offset;
		len = min(iov->iov_len, MAXPHYS);

		/* EOF? */
		if (off == sc->sc_size)
			return uiomove(NULL, 0, uio);

		if (off > sc->sc_size)
			return EFAULT;
		if (off + len > sc->sc_size)
			return EFAULT;

		if (off % sc->sc_blk)
			return ENXIO;
		if (len % sc->sc_blk)
			return ENXIO;

		rv = HV_storage_read(sc->sc_devid, sc->sc_region,
		    off / sc->sc_blk, len / sc->sc_blk, 0,
		    (paddr_t)flash_buf, &tag);
		if (rv) {
			printf("read rv = %d\n", rv);
			return ENXIO;
		}

		for (;;) {
			rv = HV_storage_status(sc->sc_devid, &tag, &status);
			if (rv == 0)
				break;
		}
		if (status)
			printf("status %ld\n", status);

		error = uiomove(flash_buf, len, uio);
		if (error)
			return error;
	}
	return 0;
}

int
flashwrite(dev, uio, flag)
	dev_t dev;
	struct uio *uio;
	int flag;
{
	struct flash_softc *sc;
	struct iovec *iov;
	u_int64_t tag, status;
	u_int off, len, x;
	int rv, error;
	u_int start;

	sc = (void *)device_lookup(&flash_cd, minor(dev));
	if (sc == NULL)
		return ENXIO;

	start = uio->uio_offset;
	if (start % FLASH_BSIZE)
		return ENXIO;

	while (uio->uio_resid > 0) {
		iov = uio->uio_iov;
		if (iov->iov_len == 0) {
			uio->uio_iov++;
			uio->uio_iovcnt--;
			if (uio->uio_iovcnt < 0)
				panic("flashwrite");
			continue;
		}

		off = uio->uio_offset;
		len = min(iov->iov_len, MAXPHYS);

		if (off > sc->sc_size)
			return EFAULT;
		if (off + len > sc->sc_size)
			return EFAULT;

		x = off - start;
		if (x > FLASH_BSIZE)
			return ENXIO;
		if (x + len > FLASH_BSIZE)
			return ENXIO;

		error = uiomove(flash_buf + x, len, uio);
		if (error)
			return error;
	}

	rv = HV_storage_write(sc->sc_devid, sc->sc_region,
	    start / sc->sc_blk, FLASH_BSIZE / sc->sc_blk, 0,
	    (paddr_t)flash_buf, &tag);
	if (rv) {
		printf("write rv = %d\n", rv);
		return ENXIO;
	}

	for (;;) {
		rv = HV_storage_status(sc->sc_devid, &tag, &status);
		if (rv == 0)
			break;
	}
	if (status)
		printf("status %ld\n", status);

	return 0;
}

paddr_t
flashmmap(dev, off, prot)
        dev_t dev;
        off_t off;
	int prot;
{
	return -1;
}

#if 0

0x0000 "cell_ext_os_area"
0x0010 version
0x0014 offset1
0x0018 offset2
0x001c reserved
0x0020 boot loader format	(0: uncompressed, 1: gzipped)
0x0024 boot loader size
       reserved
0x0200 boot flag
       reserved
0x0210 num params
0x0220 RTC diff (8bytes)
0x0228 video mode (1byte)
0x0229 controller mode (1byte)
0x0230 IP address
0x0234 netmask
0x0238 default route
0x0240 nameserver 1
0x0244 nameserver 2
       ... (0x03ff)

0x0400 other OS area			 offset1 * 0x200
0x0400 RTC diff (8bytes)
       ...

0x0800 boot loader area			 offset2 * 0x200

#endif
