/*	$Id: if_gelic.c,v 1.10 2011/10/19 13:13:31 tsubai Exp $	*/

/*-
 * Copyright (c) 2006,2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/systm.h>
#include <sys/callout.h>

#include <uvm/uvm_extern.h>

#include <net/if.h>
#include <net/if_ether.h>
#include <net/if_media.h>

#include <machine/autoconf.h>
#include <machine/hcalls.h>

struct gelic_dmadesc {
	u_int32_t address;
	u_int32_t bufsize;
	u_int32_t next;
	u_int32_t cmdstat;
	u_int32_t result_size;		/* N/U */
	u_int32_t rxlen;		/* Receive length */
	u_int32_t data_status;
	u_int32_t rxerror;		/* Receive errors */
} __attribute__((__aligned__(32)));

#if 0
# define DPRINTF printf
#else
# define DPRINTF while (0) printf
#endif

#define NTXBUF 128
#define NRXBUF 128

#define sc_if sc_ethercom.ec_if

struct gelic_softc {
	struct device sc_dev;
	struct ethercom sc_ethercom;
	struct ifmedia sc_media;

	long sc_busid, sc_devid;
	long sc_dma;

	volatile struct gelic_dmadesc sc_txdesc[NTXBUF];
	volatile struct gelic_dmadesc sc_rxdesc[NRXBUF];

	u_long sc_txdesc_addr[NTXBUF];
	u_long sc_rxdesc_addr[NRXBUF];

	struct mbuf *sc_txbuf[NTXBUF];
	struct mbuf *sc_rxbuf[NRXBUF];

	int sc_txnext;
	int sc_txfree;
	int sc_rxnext;
	int sc_rxlast;
	int sc_vlanid;

	volatile u_long sc_intr_status;

	struct callout sc_tick_ch;
	u_char sc_laddr[6];
};

#define gelic_bus_addr(va) (sc->sc_dma + vtophys((vaddr_t)(va)))

#define GELIC_CMDSTAT_NOCSUM	0x00080000
#define GELIC_CMDSTAT_TCPCSUM	0x000a0000
#define GELIC_CMDSTAT_UDPCSUM	0x000b0000
#define GELIC_CMDSTAT_EOF	0x00040000

#define GELIC_DESCR_COMPLETE	0x00000000
#define GELIC_DESCR_OWNED	0xa0000000

#define GELIC_TXINT		(1L << 56)	/* GDTDCEINT */
#define GELIC_RXINT0		(1L << 60)	/* GRFANMINT */
#define GELIC_RXINT1		(1L << 14)	/* GDADCEINT */
#define GELIC_RXINT		(GELIC_RXINT0 | GELIC_RXINT1)

#define GELIC_RXERRORS		0x7def8000	/* ? */

/* net_control commands */
#define GELIC_GET_MAC_ADDRESS	0x0001
#define GELIC_GET_LINK_STATUS	0x0002
#define GELIC_GET_VLAN_ID	0x0004

/* GET_LINK_STATUS values */
#define GELIC_LINK_UP		0x0001
#define GELIC_FULL_DUPLEX	0x0002
#define GELIC_AUTO_NEG		0x0004
#define GELIC_SPEED_10		0x0010
#define GELIC_SPEED_100		0x0020
#define GELIC_SPEED_1000	0x0040

int gelic_match(struct device *, struct cfdata *, void *);
void gelic_attach(struct device *, struct device *, void *);

void gelic_start(struct ifnet *);
int gelic_intr(void *);
void gelic_rint(struct gelic_softc *);
void gelic_tint(struct gelic_softc *);

void gelic_reset(struct gelic_softc *);
void gelic_stop(struct gelic_softc *);
int gelic_init(struct ifnet *);
int gelic_ioctl(struct ifnet *, u_long, void *);
void gelic_watchdog(struct ifnet *);
void gelic_enable_intr(struct gelic_softc *, u_long);
struct mbuf *gelic_get(struct gelic_softc *);
int gelic_mediachange(struct ifnet *);
void gelic_mediastatus(struct ifnet *, struct ifmediareq *);
void gelic_setladrf(struct gelic_softc *);
void gelic_shutdown(void *);

CFATTACH_DECL0(gelic, sizeof(struct gelic_softc), gelic_match, gelic_attach);

int
gelic_match(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct confargs *ca = aux;

	if (ca->ca_type == 3)
		return 1;
	return 0;
}

void
gelic_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct gelic_softc *sc = (void *)self;
	struct ifnet *ifp = &sc->sc_if;
	struct confargs *ca = aux;
	volatile struct gelic_dmadesc *dp;
	long v1, v2;
	int i, rv;
	u_int irq;

	sc->sc_busid = ca->ca_bus;
	sc->sc_devid = ca->ca_dev;

	if ((long)sc->sc_txdesc & 31) {
		printf(": bad dma alignment\n");
		return;
	}

	rv = HV_open(sc->sc_busid, sc->sc_devid, 0);
	if (rv && rv != HV_EBUSY) {
		printf(": couldn't open device\n");
		return;
	}

	rv = HV_net_control(sc->sc_busid, sc->sc_devid,
	    GELIC_GET_MAC_ADDRESS, 0, 0, 0, &v1, &v2);
	if (rv || v1 == 0) {
		printf(": couldn't get MAC address\n");
		return;
	}
	bcopy((char *)&v1 + 2, sc->sc_laddr, 6);

	sc->sc_dma = ps3_allocate_dma_region(sc->sc_busid, sc->sc_devid, 0);
	if (sc->sc_dma == 0) {
		printf(": couldn't map dma region\n");
		return;
	}

	dp = sc->sc_txdesc;
	for (i = 0; i < NTXBUF; i++) {
		int prev = (i - 1) & (NTXBUF - 1);

		sc->sc_txdesc_addr[i] = gelic_bus_addr(&dp[i]);
		dp[prev].next = sc->sc_txdesc_addr[i];
	}

	dp = sc->sc_rxdesc;
	for (i = 0; i < NRXBUF; i++) {
		int prev = (i - 1) & (NTXBUF - 1);

		sc->sc_rxdesc_addr[i] = gelic_bus_addr(&dp[i]);
		dp[prev].next = sc->sc_rxdesc_addr[i];
	}

	for (i = 0; i < NRXBUF; i++) {
		struct mbuf *n;

		dp = &sc->sc_rxdesc[i];

		/* XXX page boundary */
		/* XXX RX buffers must be aligned to 128-byte boundary. */

		n = gelic_get(sc);
		sc->sc_rxbuf[i] = n;
		dp->address = gelic_bus_addr(mtod(n, vaddr_t));
	}

	sc->sc_vlanid = -1;
	for (i = 1; i < 2; i++) {
		rv = HV_net_control(sc->sc_busid, sc->sc_devid,
			GELIC_GET_VLAN_ID, i + 1, 0, 0, &v1, &v2);
		if (rv == 0)
			sc->sc_vlanid = v1;
	}

	rv = HV_net_set_intrstatus_indicator(sc->sc_busid, sc->sc_devid,
		vtophys((vaddr_t)&sc->sc_intr_status), 0);
	if (rv) {
		printf("net_set_intrstatus_indicator failed\n");
		return;
	}

	irq = ps3_connect_irq(sc->sc_busid, sc->sc_devid, ca->ca_intr);
	if (irq < 0) {
		printf(": couldn't connect irq\n");
		return;
	}

	intr_establish(irq, IST_LEVEL, IPL_NET, gelic_intr, sc);

	printf(": irq %d address %s\n", irq, ether_sprintf(sc->sc_laddr));

	bcopy(sc->sc_dev.dv_xname, ifp->if_xname, IFNAMSIZ);
	ifp->if_softc = sc;
	ifp->if_ioctl = gelic_ioctl;
	ifp->if_init = gelic_init;
	ifp->if_start = gelic_start;
	ifp->if_watchdog = gelic_watchdog;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	IFQ_SET_READY(&ifp->if_snd);

	callout_init(&sc->sc_tick_ch, 0);
	gelic_reset(sc);

	ifmedia_init(&sc->sc_media, 0, gelic_mediachange, gelic_mediastatus);

	/* Choose a default media. */
	ifmedia_add(&sc->sc_media, IFM_ETHER|IFM_AUTO, 0, NULL);
	ifmedia_set(&sc->sc_media, IFM_ETHER|IFM_AUTO);

	if_attach(ifp);
	ether_ifattach(ifp, sc->sc_laddr);

	shutdownhook_establish(gelic_shutdown, sc);
}

int
gelic_intr(v)
	void *v;
{
	struct gelic_softc *sc = v;
	u_long status;

	status = sc->sc_intr_status;

	if (status == 0)
		return 0;

	if (status & GELIC_RXINT)
		gelic_rint(sc);
	if (status & GELIC_TXINT)
		gelic_tint(sc);

	return 1;
}

void
gelic_rint(sc)
	struct gelic_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	volatile struct gelic_dmadesc *dp;
	struct mbuf *m, *n;
	int i, len;

	for (i = sc->sc_rxlast;; i = (i + 1) & (NRXBUF - 1)) {
		dp = &sc->sc_rxdesc[i];
		if ((dp->cmdstat & 0xf0000000) == GELIC_DESCR_OWNED)
			break;

		if (dp->rxerror & GELIC_RXERRORS) {
			ifp->if_ierrors++;
			goto next;
		}

		len = dp->rxlen;

#ifdef DIAGNOSTIC
		if (len > MCLBYTES) {
			printf("%s: too big packet (%d)\n", ifp->if_xname, len);
			ifp->if_ierrors++;
			goto next;
		}
#endif

		m = sc->sc_rxbuf[i];

		/* XXX Remove garbage. */
		m->m_data += 2;
		len -= 2;

		m->m_len = m->m_pkthdr.len = len;

		n = gelic_get(sc);
		if (n == NULL) {
			ifp->if_ierrors++;
			goto next;
		}
		sc->sc_rxbuf[i] = n;

#if NBPFILTER > 0
		/*
		 * Check if there's a BPF listener on this interface.
		 * If so, hand off the raw packet to BPF.
		 */
		if (ifp->if_bpf)
			bpf_mtap(ifp->if_bpf, m);
#endif
		(*ifp->if_input)(ifp, m);
		ifp->if_ipackets++;

		dp->address = gelic_bus_addr(mtod(n, vaddr_t));

next:
		dp->bufsize = 2000;
		dp->result_size = 0;
		dp->rxlen = 0;
		dp->data_status = 0;
		dp->rxerror = 0;
		asm volatile ("sync" ::: "memory");
		dp->cmdstat = GELIC_DESCR_OWNED;
	}
	sc->sc_rxlast = i;
}

void
gelic_tint(sc)
	struct gelic_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	volatile struct gelic_dmadesc *dp;
	int i;

	ifp->if_flags &= ~IFF_OACTIVE;
	ifp->if_timer = 0;

	i = sc->sc_txfree;

	for (;;) {
		dp = &sc->sc_txdesc[i];
		DPRINTF("[%d status 0x%x]", i, dp->cmdstat);

		/*
		 * Ǹdp롣Τϥơߤ
		 * 0x080c0202 / 0x80000
		 */

		if ((dp->cmdstat & 0xf0000000) == GELIC_DESCR_OWNED)
			break;
		if (dp->cmdstat == 0)
			break;

		/* XXX 顼Ȥ */

		if (sc->sc_txbuf[i])
			m_freem(sc->sc_txbuf[i]);
		sc->sc_txbuf[i] = NULL;
		dp->cmdstat = 0;
		i = (i + 1) & (NTXBUF - 1);
	}

	sc->sc_txfree = i;
	gelic_start(ifp);
}

void
gelic_start(ifp)
	struct ifnet *ifp;
{
	struct gelic_softc *sc = ifp->if_softc;
	volatile struct gelic_dmadesc *dp;
	struct mbuf *m;
	int i, n, done, len, rv;

	if ((ifp->if_flags & (IFF_RUNNING | IFF_OACTIVE)) != IFF_RUNNING)
		return;

	/* Calculate free segments. */
	i = sc->sc_txnext;
	done = sc->sc_txfree;
	n = ((done - i - 1) & (NTXBUF - 1)) + 1;

start:
	i = sc->sc_txnext;

	IFQ_DEQUEUE(&ifp->if_snd, m);
	if (m == 0)
		return;

#if NBPFILTER > 0
	/*
	 * If BPF is listening on this interface, let it see the packet
	 * before we commit it to the wire.
	 */
	if (ifp->if_bpf)
		bpf_mtap(ifp->if_bpf, m);
#endif

#ifdef DIAGNOSTIC
	if ((m->m_flags & M_PKTHDR) == 0) {
		printf("%s: mbuf w/o M_PKTHDR\n", ifp->if_xname);
		goto start;
	}
#endif

	/* XXX VLAN hack */
	if (sc->sc_vlanid != -1) {
		m->m_data -= 4;
		m->m_len += 4;
#if 0
		if (m->m_data < m->m_pktdat) {
			printf("%s: bad mbuf\n", ifp->if_xname);
			m_freem(m);
			goto start;
		}
#endif
		memmove(m->m_data, m->m_data + 4, 12);
		m->m_data[12] = 0x81;
		m->m_data[13] = 0x00;
		m->m_data[14] = sc->sc_vlanid >> 8;
		m->m_data[15] = sc->sc_vlanid;
	}

	DPRINTF("[gelic_start %d]", i);
	sc->sc_txbuf[i] = m;
	for (; m; m = m->m_next) {
		vaddr_t va = mtod(m, vaddr_t);
		int resid = m->m_len;

		while (resid > 0) {
			len = min(resid, NBPG - (va & PGOFSET));
			dp = &sc->sc_txdesc[i];
			dp->address = gelic_bus_addr(va);
			dp->bufsize = len;
			dp->cmdstat = GELIC_CMDSTAT_NOCSUM | GELIC_DESCR_OWNED;
			dp->result_size = 0;
			dp->rxlen = 0;
			dp->data_status = 0;
			dp->rxerror = 0;

			i = (i + 1) & (NTXBUF - 1);
			n--;

			dp->next = sc->sc_txdesc_addr[i];

			va += len;
			resid -= len;
			if (resid == 0 && m->m_next == NULL)
				goto done;

			if (n == 0) {
				ifp->if_flags |= IFF_OACTIVE;

				/* XXX */
				printf("%s: dropped\n", ifp->if_xname);
				m_freem(sc->sc_txbuf[sc->sc_txnext]);
				return;
			}
		}
	}
done:
	dp->cmdstat |= GELIC_CMDSTAT_EOF;
	dp->next = 0;

	asm volatile ("sync" ::: "memory");

	/* 5 seconds to watch for failing to transmit */
	ifp->if_timer = 5;
	ifp->if_opackets++;		/* # of pkts */

retry:
	rv = HV_net_start_txdma(sc->sc_busid, sc->sc_devid,
		sc->sc_txdesc_addr[sc->sc_txnext], 0);
	if (rv) {
		/* XXX for now */
		if (rv == HV_EBUSY)
			goto retry;
		printf("gelic_start: kick failed %d %d\n", sc->sc_txnext, rv);
		ifp->if_flags |= IFF_OACTIVE;
		m_freem(sc->sc_txbuf[sc->sc_txnext]);
		return;
	}

	sc->sc_txnext = i;

	/* ĤꥻȤʤʤä */
	if (n < 8) {
		ifp->if_flags |= IFF_OACTIVE;
		return;
	}

	goto start;
}

int
gelic_ioctl(ifp, cmd, data)
	struct ifnet *ifp;
	u_long cmd;
	void *data;
{
	struct gelic_softc *sc = ifp->if_softc;
	struct ifreq *ifr = data;
	int s, error = 0;

	s = splnet();

	switch (cmd) {

	case SIOCINITIFADDR:
	case SIOCSIFADDR:
		error = ether_ioctl(ifp, cmd, data);
		break;

	case SIOCSIFFLAGS:
		if ((error = ifioctl_common(ifp, cmd, data)) != 0)
			break;
		if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) == IFF_RUNNING) {
			/*
			 * If interface is marked down and it is running, then
			 * stop it.
			 */
			gelic_stop(sc);
			ifp->if_flags &= ~IFF_RUNNING;
		} else if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) == IFF_UP) {
			/*
			 * If interface is marked up and it is stopped, then
			 * start it.
			 */
			gelic_init(ifp);
		} else {
			/*
			 * Reset the interface to pick up changes in any other
			 * flags that affect hardware registers.
			 */
			gelic_reset(sc);
			gelic_init(ifp);
		}
		break;

	case SIOCADDMULTI:
	case SIOCDELMULTI:
		error = ether_ioctl(ifp, cmd, data);
		if (error == ENETRESET) {
			/*
			 * Multicast list has changed; set the hardware filter
			 * accordingly.
			 */
			if (ifp->if_flags & IFF_RUNNING) {
				gelic_init(ifp);
				/* gmac_setladrf(sc); */
			}
			error = 0;
		}
		break;

	case SIOCGIFMEDIA:
	case SIOCSIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->sc_media, cmd);
		break;

	default:
		error = ether_ioctl(ifp, cmd, data);
	}

	splx(s);
	return error;
}

int
gelic_init(ifp)
	struct ifnet *ifp;
{
	struct gelic_softc *sc = ifp->if_softc;
	volatile struct gelic_dmadesc *dp;
	int err, i;

	gelic_stop(sc);

	for (i = 0; i < NRXBUF; i++) {
		dp = &sc->sc_rxdesc[i];
		dp->address = gelic_bus_addr(mtod(sc->sc_rxbuf[i], vaddr_t));
		dp->bufsize = 2000;
		dp->result_size = 0;
		dp->rxlen = 0;
		dp->data_status = 0;
		dp->rxerror = 0;
		dp->cmdstat = GELIC_DESCR_OWNED;
	}
	asm volatile ("sync" ::: "memory");

	sc->sc_txnext = 0;
	sc->sc_txfree = 0;
	sc->sc_rxnext = 0;
	sc->sc_rxlast = 0;

	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;
	ifp->if_timer = 0;

	gelic_setladrf(sc);

	/* Enable TX and RX interrupts. */
	gelic_enable_intr(sc, GELIC_RXINT | GELIC_TXINT);

	err = HV_net_start_rxdma(sc->sc_busid, sc->sc_devid,
		sc->sc_rxdesc_addr[0], 0);
	if (err)
		printf("%s: starting RX DMA failed (%d)\n", ifp->if_xname, err);

	gelic_start(ifp);

	return 0;
}

void
gelic_stop(sc)
	struct gelic_softc *sc;
{
	int i;

	/* Disable all interrupts. */
	gelic_enable_intr(sc, 0);

	callout_stop(&sc->sc_tick_ch);

	/* Stop DMAs. */
	for (i = 0; i < 10; i++) {
		HV_net_stop_txdma(sc->sc_busid, sc->sc_devid, 0);
		HV_net_stop_rxdma(sc->sc_busid, sc->sc_devid, 0);
	}
}

void
gelic_reset(sc)
	struct gelic_softc *sc;
{
	gelic_stop(sc);
}

void
gelic_watchdog(ifp)
	struct ifnet *ifp;
{
	struct gelic_softc *sc = ifp->if_softc;

	printf("%s: device timeout\n", ifp->if_xname);
	ifp->if_oerrors++;

	gelic_reset(sc);
	gelic_init(ifp);
}

void
gelic_enable_intr(sc, intr)
	struct gelic_softc *sc;
	u_long intr;
{
	int rv;

	rv = HV_net_set_intrmask(sc->sc_busid, sc->sc_devid, intr, 0);
	if (rv)
		printf("net_set_intrmask failed (%d)\n", rv);
}

void
gelic_setladrf(sc)
	struct gelic_softc *sc;
{
	struct ifnet *ifp = &sc->sc_if;
	struct ether_multi *enm;
	struct ether_multistep step;
	struct ethercom *ec = &sc->sc_ethercom;
	u_long addr;

	if (HV_net_del_mcast(sc->sc_busid, sc->sc_devid, 0, 1))
		goto error;

#if 0
#define GELIC_BCAST_ADDR 0xffffffffffff

	if (HV_net_add_mcast(sc->sc_busid, sc->sc_devid, GELIC_BCAST_ADDR, 0))
		goto error;
#endif

	if (ifp->if_flags & IFF_PROMISC)
		goto allmulti;

	/* XXX max 32 multicast addresses. */
	ETHER_FIRST_MULTI(step, ec, enm);
	while (enm != NULL) {
		if (bcmp(enm->enm_addrlo, enm->enm_addrhi, 6)) {
			/*
			 * We must listen to a range of multicast addresses.
			 * For now, just accept all multicasts, rather than
			 * trying to set only those filter bits needed to match
			 * the range.  (At this time, the only use of address
			 * ranges is for IP multicast routing, for which the
			 * range is big enough to require all bits set.)
			 */
			goto allmulti;
		}
		addr = 0;
		bcopy(enm->enm_addrlo, (char *)&addr + 2, 6);
		if (HV_net_add_mcast(sc->sc_busid, sc->sc_devid, addr, 0))
			goto allmulti;

		ETHER_NEXT_MULTI(step, enm);
	}

	ifp->if_flags &= ~IFF_ALLMULTI;
	return;

allmulti:
	ifp->if_flags |= IFF_ALLMULTI;
	if (HV_net_add_mcast(sc->sc_busid, sc->sc_devid, 0, 1))
		goto error;
	return;

error:
	printf("%s: couldn't set multicast addresses\n", ifp->if_xname);
}

int
gelic_mediachange(ifp)
	struct ifnet *ifp;
{
	struct gelic_softc *sc = ifp->if_softc;
	struct ifmedia *ifm = &sc->sc_media;

	if (ifm->ifm_media == (IFM_ETHER|IFM_AUTO))
		return 0;
	return EINVAL;
}

#define GELIC_PORT 2

void
gelic_mediastatus(ifp, ifmr)
	struct ifnet *ifp;
	struct ifmediareq *ifmr;
{
	struct gelic_softc *sc = ifp->if_softc;
	u_long v1, v2;
	int err;

	ifmr->ifm_active = IFM_ETHER;

	err = HV_net_control(sc->sc_busid, sc->sc_devid,
		GELIC_GET_LINK_STATUS, GELIC_PORT, 0, 0, &v1, &v2);
	if (err) {
		ifmr->ifm_active = IFM_ETHER|IFM_NONE;
		return;
	}
	if ((v1 & GELIC_LINK_UP) == 0) {
		ifmr->ifm_active = IFM_ETHER|IFM_NONE;
		return;
	}

	if (v1 & GELIC_FULL_DUPLEX)
		ifmr->ifm_active |= IFM_FDX;

	switch (v1 & (GELIC_SPEED_10|GELIC_SPEED_100|GELIC_SPEED_1000)) {
	case GELIC_SPEED_10:
		ifmr->ifm_active |= IFM_10_T;
		break;
	case GELIC_SPEED_100:
		ifmr->ifm_active |= IFM_100_TX;
		break;
	case GELIC_SPEED_1000:
		ifmr->ifm_active |= IFM_1000_T;
		break;
	}
}

struct mbuf *
gelic_get(sc)
	struct gelic_softc *sc;
{
	struct mbuf *m;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
		return NULL;
	m->m_pkthdr.rcvif = &sc->sc_if;

	MCLGET(m, M_DONTWAIT);
	if ((m->m_flags & M_EXT) == 0) {
		m_free(m);
		return NULL;
	}

	return m;
}

void
gelic_shutdown(v)
	void *v;
{
	struct gelic_softc *sc = v;

	gelic_stop(sc);
}
