/*	$Id: ps3av.c,v 1.48 2011/11/09 09:01:48 tsubai Exp $	*/

/*-
 * Copyright (c) 2006,2007,2011 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <machine/ps3av.h>

static struct {
	int region;
	int vid;
	int n_hdmi;
	int n_avmulti;
	int n_spdif;
	u_int avport[PS3AV_AVPORT_MAX + PS3AV_SPDIF_MAX];
	u_int spdif[PS3AV_SPDIF_MAX];
} ps3av;

#define YUV444 PS3AV_CMD_CS_YUV444_8
#define RGB8   PS3AV_CMD_CS_RGB_8

struct vmode {
	u_int x:12;
	u_int y:12;
	u_int cs:4;
	u_int vid:4;
};
static struct vmode video_mode_table[] = {
	{  720,  480, YUV444, PS3AV_CMD_VID_480I },
	{  720,  576, YUV444, PS3AV_CMD_VID_576I },
	{  720,  480, YUV444, PS3AV_CMD_VID_480P },
	{  720,  576, YUV444, PS3AV_CMD_VID_576P },
	{ 1920, 1080, YUV444, PS3AV_CMD_VID_1080I_60HZ },
	{ 1920, 1080, YUV444, PS3AV_CMD_VID_1080I_50HZ },
	{ 1280,  720, YUV444, PS3AV_CMD_VID_720P_60HZ },
	{ 1280,  720, YUV444, PS3AV_CMD_VID_720P_50HZ },
	{ 1920, 1080, YUV444, PS3AV_CMD_VID_1080P_60HZ },
	{ 1920, 1080, YUV444, PS3AV_CMD_VID_1080P_50HZ },
	{ 1280,  768, RGB8,   PS3AV_CMD_VID_1280x768 },
	{ 1280, 1024, RGB8,   PS3AV_CMD_VID_1280x1024 },
	{ 1920, 1200, RGB8,   PS3AV_CMD_VID_1920x1200 },
	{ 0 }
};

static struct vmode *vidtovmode(int);
static int set_videomode(int, int, int);
static int bestvid(u_int, u_int);
static int hdmi_vid(struct ps3av_monitor_info *);
static int auto_videomode(int);

struct vmode *
vidtovmode(vid)
	int vid;
{
	struct vmode *mode;

	for (mode = video_mode_table; mode->vid; mode++)
		if (mode->vid == (vid & PS3AV_MODE_MASK))
			return mode;
	return NULL;
}

int
set_videomode(avport, head, vid)
	int avport, head, vid;
{
	struct vmode *mode;
	int cs, aspect, option = vid;

	vid &= PS3AV_MODE_MASK;
	mode = vidtovmode(vid);
	cs = mode->cs;
	if (option & (PS3AV_MODE_DVI | PS3AV_MODE_RGB))
		cs = RGB8;
#if 1
	switch (avport) {
	case PS3AV_CMD_AVPORT_HDMI_0:
	case PS3AV_CMD_AVPORT_HDMI_1:
		cs = RGB8;
	}
#endif
	if (mode->x * 100 / mode->y > 170)
		aspect = PS3AV_CMD_AV_ASPECT_16_9;
	else
		aspect = PS3AV_CMD_AV_ASPECT_4_3;

	ps3av_cmd_video_mode(head, vid);
	ps3av_cmd_av_video_cs(avport, vid, cs, aspect, 0);

#if 0	/* Not supported on retail PS3 products. */
	if (option & PS3AV_MODE_HDCP_OFF)
		ps3av_cmd_av_hdmi_mode(PS3AV_CMD_HDMI_HDCP_OFF);
	else
		ps3av_cmd_av_hdmi_mode(PS3AV_CMD_HDMI_NORMAL);
#endif

	return 0;
}

struct modebit {
	u_int bit:24;
	u_int vid:8;
};
static struct modebit table50[] = {
	{ PS3AV_RESBIT_1080P, PS3AV_CMD_VID_1080P_50HZ },
	{ PS3AV_RESBIT_1080I, PS3AV_CMD_VID_1080I_50HZ },
	{ PS3AV_RESBIT_720P,  PS3AV_CMD_VID_720P_50HZ },
	{ PS3AV_RESBIT_576P,  PS3AV_CMD_VID_576P },
	{ 0 }
};
static struct modebit table60[] = {
	{ PS3AV_RESBIT_1080P, PS3AV_CMD_VID_1080P_60HZ },
	{ PS3AV_RESBIT_1080I, PS3AV_CMD_VID_1080I_60HZ },
	{ PS3AV_RESBIT_720P,  PS3AV_CMD_VID_720P_60HZ },
	{ PS3AV_RESBIT_480P,  PS3AV_CMD_VID_480P },
	{ 0 }
};

int
bestvid(mode_50, mode_60)
	u_int mode_50, mode_60;
{
	struct modebit *t;
	u_int mode;

	mode_50 &= PS3AV_RESBIT_MASK_50;
	mode_60 &= PS3AV_RESBIT_MASK_60;

	if (mode_50 > mode_60) {
		t = table50;
		mode = mode_50;
	} else {
		t = table60;
		mode = mode_60;
	}
	for (; t->bit; t++) {
		if ((mode & t->bit) == t->bit)
			return t->vid;
	}
	return -1;
}

int
hdmi_vid(info)
	struct ps3av_monitor_info *info;
{
	int vid;

	if (info->monitor_type != PS3AV_MONITOR_HDMI)
		return -1;

	vid = bestvid(info->mode_50.native, info->mode_60.native);
	if (vid == -1)
		vid = bestvid(info->mode_50.support, info->mode_60.support);
	return vid;
}

int
auto_videomode(boot)
	int boot;
{
	u_char buf[256];
	struct ps3av_monitor_info *info = (void *)buf;
	u_int avport;
	int i, vid = -1;

	/* Probe HDMI. */
	for (i = 0; i < ps3av.n_hdmi; i++) {
		avport = PS3AV_CMD_AVPORT_HDMI_0 + i;
		if (ps3av_cmd_get_monitor_info(avport, buf, sizeof buf) < 0)
			break;
		if (info->monitor_type == PS3AV_MONITOR_DVI)
			return PS3AV_CMD_VID_480P | PS3AV_MODE_DVI;

		vid = hdmi_vid(info);
		if (vid != -1)
			goto found_HDMI;
	}

	/* Use AVMULTI. */
	vid = (ps3av.region & PS3AV_REGION_60) ?
	    PS3AV_CMD_VID_480I : PS3AV_CMD_VID_576I;
	if (ps3av.region & PS3AV_REGION_RGB)
		vid |= PS3AV_MODE_RGB;
	return vid;

found_HDMI:
	if (!boot)
		return vid;

	if (ps3av.region & PS3AV_REGION_60) {
		if (info->mode_60.support & PS3AV_RESBIT_480P)
			return PS3AV_CMD_VID_480P;
		if (info->mode_50.support & PS3AV_RESBIT_576P)
			return PS3AV_CMD_VID_576P;
		return PS3AV_CMD_VID_480P;
	} else {
		if (info->mode_50.support & PS3AV_RESBIT_576P)
			return PS3AV_CMD_VID_576P;
		if (info->mode_60.support & PS3AV_RESBIT_480P)
			return PS3AV_CMD_VID_480P;
		return PS3AV_CMD_VID_576P;
	}
}

int
ps3av_set_video_mode(vid)
	int vid;
{
	int i, port = 0;

	if ((vid & PS3AV_MODE_MASK) == 0) {
		u_int option = vid & ~PS3AV_MODE_MASK;

		vid = auto_videomode(0);
		vid |= option;
	}

	if (vidtovmode(vid) == NULL)
		return -1;

	ps3av.vid = vid;
	for (i = 0; i < ps3av.n_hdmi; i++) {
		set_videomode(ps3av.avport[port], 0, vid);
		set_videomode(ps3av.avport[port], 1, vid);
		port++;
	}
	for (i = 0; i < ps3av.n_avmulti; i++) {
		set_videomode(ps3av.avport[port], 0, vid);
		set_videomode(ps3av.avport[port], 1, vid);
		port++;
	}
	return 0;
}

int
ps3av_get_video_mode()
{
	return ps3av.vid;
}

int
ps3av_screen_size(vid, x, y)
	u_int vid, *x, *y;
{
	struct vmode *mode;

	mode = vidtovmode(vid);
	if (mode == NULL)
		return -1;
	*x = mode->x;
	*y = mode->y;
	return 0;
}

int
ps3av_video_signal_off()
{
	int i, n, rv = 0;

	n = ps3av.n_hdmi + ps3av.n_avmulti;
	for (i = 0; i < n; i++)
		rv |= ps3av_cmd_av_video_signal_off(ps3av.avport[i]);
	return rv;
}

int
ps3av_video_mute(on)
	int on;
{
	int i, n, rv = 0;

	n = ps3av.n_hdmi + ps3av.n_avmulti;
	for (i = 0; i < n; i++)
		rv |= ps3av_cmd_av_video_mute(ps3av.avport[i], on);
	return rv;
}

int
ps3av_audio_mute(on)
	int on;
{
	int i, nav, nopt, rv = 0;

	nav  = ps3av.n_hdmi + ps3av.n_avmulti;
	nopt = ps3av.n_spdif;
	for (i = 0; i < nav; i++)
		rv |= ps3av_cmd_av_audio_mute(ps3av.avport[i], on);
	for (i = 0; i < nopt; i++)
		rv |= ps3av_cmd_audio_mute(ps3av.spdif[i], on);
	return rv;
}

int
ps3av_init()
{
	int i, port = 0;

	bzero(&ps3av, sizeof ps3av);
	ps3av.region = PS3AV_REGION_60;		/* XXX boot data from flash */

	ps3av_vuart_init();

	if (ps3av_cmd_init() < 0)
		printf("ps3av_cmd_init() failed\n");

	if (ps3av_cmd_get_hwconf(&ps3av.n_hdmi, &ps3av.n_avmulti,
	    &ps3av.n_spdif) < 0)
		return -1;

	printf("ps3av: %d HDMI(s), %d AVMULTI, %d SPDIF found\n",
	    ps3av.n_hdmi, ps3av.n_avmulti, ps3av.n_spdif);

	ps3av.spdif[0] = PS3AV_CMD_AVPORT_SPDIF_0;
	for (i = 0; i < ps3av.n_hdmi; i++)
		ps3av.avport[port++] = PS3AV_CMD_AVPORT_HDMI_0 + i;
	for (i = 0; i < ps3av.n_avmulti; i++)
		ps3av.avport[port++] = PS3AV_CMD_AVPORT_AVMULTI_0 + i;
	for (i = 0; i < ps3av.n_spdif; i++)
		ps3av.avport[port++] = PS3AV_CMD_AVPORT_SPDIF_0 + i;

	ps3av.vid = auto_videomode(1);

	return 0;
}
