/*	$Id: ps3av_cmd.c,v 1.39 2011/11/07 00:22:19 tsubai Exp $	*/

/*-
 * Copyright (c) 2006,2007,2011 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <machine/hcalls.h>
#include <machine/ps3av.h>

int ps3av_debug;

#ifdef PS3AV_DEBUG
#define DPRINTF(...) do {		\
	if (ps3av_debug)		\
		printf(__VA_ARGS__);	\
} while (0)
#else
#define DPRINTF(...) do {} while (0)
#endif

#define CID_AV_INIT		0x00000001
#define CID_AV_FIN		0x00000002
#define CID_AV_GET_HWCONF	0x00000003
#define CID_AV_GET_MONITOR_INFO	0x00000004
#define CID_AV_ENABLE_EVENT	0x00000006
#define CID_AV_DISABLE_EVENT	0x00000007
#define CID_AV_TV_MUTE		0x0000000a

#define CID_AV_VIDEO_CS		0x00010001
#define CID_AV_VIDEO_MUTE	0x00010002
#define CID_AV_VIDEO_SIGNAL_OFF	0x00010003
#define CID_AV_AUDIO_PARAM	0x00020001
#define CID_AV_AUDIO_MUTE	0x00020002
#define CID_AV_HDMI_MODE	0x00040001

#define CID_VIDEO_INIT		0x01000001
#define CID_VIDEO_MODE		0x01000002
#define CID_VIDEO_FORMAT	0x01000004
#define CID_VIDEO_STRIDE	0x01000005

#define CID_AUDIO_INIT		0x02000001
#define CID_AUDIO_MODE		0x02000002
#define CID_AUDIO_MUTE		0x02000003
#define CID_AUDIO_ACTIVE	0x02000004
#define CID_AUDIO_INACTIVE	0x02000005
#define CID_AUDIO_SPDIF_BIT	0x02000006
#define CID_AUDIO_CTRL		0x02000007

#define CID_EVENT_UNPLUGGED	0x10000001
#define CID_EVENT_PLUGGED	0x10000002
#define CID_EVENT_HDCP_DONE	0x10000003
#define CID_EVENT_HDCP_FAIL	0x10000004
#define CID_EVENT_HDCP_AUTH	0x10000005
#define CID_EVENT_HDCP_ERROR	0x10000006

#define CID_AVB_PARAM		0x04000001

#define PS3AV_HDR_SIZE	4			/* version + size */
#define MINCMDLEN 	(PS3AV_HDR_SIZE + 4)	/* version, size, cid */

static void ps3av_set_hdr(u_int, u_int, void *);
static int ps3av_cmd(u_int, u_int, size_t, void *);

int
ps3av_cmd_init()
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int32_t status;
		};
	} __packed video_init;
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int32_t status;
		};
	} __packed audio_init;
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int32_t event_mask;
			u_int32_t status;
		};
	} __packed av_init;

	bzero(&video_init, sizeof video_init);
	ps3av_cmd(CID_VIDEO_INIT, MINCMDLEN, sizeof video_init, &video_init);
	if (video_init.status)
		return video_init.status;

	bzero(&audio_init, sizeof audio_init);
	ps3av_cmd(CID_AUDIO_INIT, MINCMDLEN, sizeof audio_init, &audio_init);
	if (audio_init.status)
		return audio_init.status;

	bzero(&av_init, sizeof av_init);
	ps3av_cmd(CID_AV_INIT, sizeof av_init, sizeof av_init, &av_init);
	if (av_init.status)
		return av_init.status;

	return 0;
}

int
ps3av_cmd_fin()
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	ps3av_cmd(CID_AV_FIN, MINCMDLEN, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_get_hwconf(hdmi, avmulti, spdif)
	int *hdmi, *avmulti, *spdif;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		u_int32_t status;		/* RECV */
		u_int16_t n_hdmi;
		u_int16_t n_avmulti;
		u_int16_t n_spdif;
		u_int16_t reserved;
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	ps3av_cmd(CID_AV_GET_HWCONF, MINCMDLEN, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;

	*hdmi    = cmd.n_hdmi;
	*avmulti = cmd.n_avmulti;
	*spdif   = cmd.n_spdif;
	return 0;
}

int
ps3av_cmd_get_monitor_info(avport, buf, len)
	u_int avport, len;
	void *buf;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int16_t avport;	/* SEND */
			struct {		/* RECV */
				u_int32_t status;
				u_int8_t info[256];
			};
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.avport = avport;
	ps3av_cmd(CID_AV_GET_MONITOR_INFO, MINCMDLEN + 4, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;

	if (cmd.size - 8 > len)
		printf("monitor_info truncated\n");
	bcopy(cmd.info, buf, len);

	/* ps3av_dump_monitor_info(buf); */
	return 0;
}

int
ps3av_cmd_enable_event(mask)
	u_int mask;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int32_t event_mask;
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.event_mask = mask;
	ps3av_cmd(CID_AV_ENABLE_EVENT, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_disable_event(mask)
	u_int mask;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int32_t event_mask;
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.event_mask = mask;
	ps3av_cmd(CID_AV_DISABLE_EVENT, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_av_tv_mute(avport, on)
	u_int avport, on;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {
				u_int16_t avport;	/* HDMI only */
				u_int16_t mute;
			};
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.avport = avport;
	cmd.mute = on;
	ps3av_cmd(CID_AV_TV_MUTE, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_av_video_mute(avport, on)
	u_int avport, on;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {
				u_int16_t avport;
				u_int16_t mute;
			};
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.avport = avport;
	cmd.mute = on;
	ps3av_cmd(CID_AV_VIDEO_MUTE, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_av_video_signal_off(port)
	u_int port;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int16_t avport;
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.avport = port;
	ps3av_cmd(CID_AV_VIDEO_SIGNAL_OFF, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_av_hdmi_mode(mode)
	u_int mode;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int8_t mode;
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.mode = mode;
	ps3av_cmd(CID_AV_HDMI_MODE, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

/* Convert vid to av_vid. */
static u_char av_vid[] = {
	PS3AV_CMD_AV_VID_480P,
	PS3AV_CMD_AV_VID_480I,
	PS3AV_CMD_AV_VID_480P,
	PS3AV_CMD_AV_VID_576I,
	PS3AV_CMD_AV_VID_480P,
	PS3AV_CMD_AV_VID_480P,
	PS3AV_CMD_AV_VID_576P,
	PS3AV_CMD_AV_VID_1080I_60HZ,
	PS3AV_CMD_AV_VID_1080I_50HZ,
	PS3AV_CMD_AV_VID_720P_60HZ,
	PS3AV_CMD_AV_VID_720P_50HZ,
	PS3AV_CMD_AV_VID_1080P_60HZ,
	PS3AV_CMD_AV_VID_1080P_50HZ,
	PS3AV_CMD_AV_VID_1280x768,
	PS3AV_CMD_AV_VID_1280x1024,
	PS3AV_CMD_AV_VID_1920x1200
};

/* Convert cs to av_cs and bitlen. */
static struct {
	u_char av_cs:4;
	u_char bitlen:4;
} av_cs[] = {
	{ PS3AV_CMD_AV_CS_RGB_8,    PS3AV_CMD_AV_CS_8 },
	{ PS3AV_CMD_AV_CS_RGB_8,    PS3AV_CMD_AV_CS_8 },
	{ PS3AV_CMD_AV_CS_YUV444_8, PS3AV_CMD_AV_CS_8 },
	{ PS3AV_CMD_AV_CS_YUV422_8, PS3AV_CMD_AV_CS_8 },
	{ PS3AV_CMD_AV_CS_XVYCC_8,  PS3AV_CMD_AV_CS_8 },
	{ PS3AV_CMD_AV_CS_RGB_8,    PS3AV_CMD_AV_CS_10 },
	{ PS3AV_CMD_AV_CS_YUV444_8, PS3AV_CMD_AV_CS_10 },
	{ PS3AV_CMD_AV_CS_YUV422_8, PS3AV_CMD_AV_CS_10 },
	{ PS3AV_CMD_AV_CS_XVYCC_8,  PS3AV_CMD_AV_CS_10 },
	{ PS3AV_CMD_AV_CS_RGB_8,    PS3AV_CMD_AV_CS_12 },
	{ PS3AV_CMD_AV_CS_YUV444_8, PS3AV_CMD_AV_CS_12 },
	{ PS3AV_CMD_AV_CS_YUV422_8, PS3AV_CMD_AV_CS_12 },
	{ PS3AV_CMD_AV_CS_XVYCC_8,  PS3AV_CMD_AV_CS_12 }
};

int
ps3av_cmd_av_video_cs(avport, vid, cs_out, aspect, dither)
	u_int avport, dither;
	int vid, cs_out, aspect;
{
	/* XXX AVB_PARAM Ȥʤưʤ */
	struct av_video_cs {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		u_int16_t avport;
		u_int16_t av_vid;
		u_int16_t av_cs_out;
		u_int16_t av_cs_in;
		u_int8_t dither;
		u_int8_t bitlen_out;
		u_int8_t super_white;
		u_int8_t aspect;
	};
	struct {				/* Use AVB_PARAM command. */
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {		/* SEND */
				u_int16_t n_video;
				u_int16_t n_audio;
				u_int16_t n_av_video;
				u_int16_t n_av_audio;
				struct av_video_cs cs;
			};
			u_int32_t status;	/* RECV */
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.cs.avport = avport;
	cmd.cs.aspect = aspect;
	cmd.cs.av_vid = av_vid[vid];
	cmd.cs.av_cs_in = PS3AV_CMD_AV_CS_RGB_8;
	cmd.cs.av_cs_out = av_cs[cs_out].av_cs;
	cmd.cs.bitlen_out = av_cs[cs_out].bitlen;
	if (dither & PS3AV_MODE_DITHER)
		cmd.cs.dither = PS3AV_CMD_AV_DITHER_ON | PS3AV_CMD_AV_DITHER_8BIT;
	else
		cmd.cs.dither = PS3AV_CMD_AV_DITHER_OFF;

	cmd.n_av_video = 1;
	cmd.n_video = cmd.n_audio = cmd.n_av_audio = 0;
	ps3av_set_hdr(CID_AV_VIDEO_CS, sizeof cmd.cs, &cmd.cs);

	ps3av_cmd(CID_AVB_PARAM, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_video_mode(head, vid)
	int head, vid;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {		/* SEND */
				u_int32_t head;
				u_int32_t reserved;
				u_int32_t vid;
				u_int16_t reserved1;
				u_int16_t width;
				u_int16_t reserved2;
				u_int16_t height;
				u_int32_t linebytes;
				u_int32_t video_out_format;
				u_int32_t fb_format;
				u_int16_t reserved3;
				u_int16_t rgb_order;
				u_int32_t reserved4;
			};
			u_int32_t status;	/* RECV */
		};
	} __packed cmd;
	u_int x, y;

	if (ps3av_screen_size(vid, &x, &y))
		return -1;
	if (vid == PS3AV_CMD_VID_480I && head == 1)
		vid = PS3AV_CMD_VID_480I_A;

	bzero(&cmd, sizeof cmd);
	cmd.width = x;
	cmd.height = y;
	cmd.linebytes = x * 4;
	cmd.head = head;
	cmd.vid = vid;
	cmd.video_out_format = PS3AV_CMD_VIDEO_OUT_FORMAT_RGB_12BIT;
	cmd.fb_format = PS3AV_CMD_VIDEO_FORMAT_ARGB_8BIT;
	cmd.rgb_order = PS3AV_CMD_VIDEO_ORDER_RGB;
	ps3av_cmd(CID_VIDEO_MODE, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_video_format(head, format, mute)
	u_int head, format, mute;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {
				u_int32_t head;
				u_int32_t fb_format;
				u_int16_t reserved;
				u_int16_t rgb_order;
			};
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.head = head;
	if (mute)	/* XXX ɬפ? */
		format = PS3AV_CMD_VIDEO_FORMAT_BLACK;
	cmd.fb_format = format;
	cmd.rgb_order = PS3AV_CMD_VIDEO_ORDER_RGB;
	ps3av_cmd(CID_VIDEO_FORMAT, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_video_stride(head, linebytes)
	u_int head, linebytes;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {
				u_int32_t head;
				u_int32_t linebytes;
			};
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.head = head;
	cmd.linebytes = linebytes;
	ps3av_cmd(CID_VIDEO_STRIDE, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_av_audio_mute(avport, on)
	u_int avport, on;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {
				u_int16_t avport;
				u_int16_t mute;
			};
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.avport = avport;
	cmd.mute = on ? PS3AV_CMD_MUTE_ON : PS3AV_CMD_MUTE_OFF;
	ps3av_cmd(CID_AV_AUDIO_MUTE, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_audio_mute(avport, on)
	u_int avport, on;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			struct {
				u_int8_t avport;	/* SPDIF port */
				u_int8_t reserved[3];
				u_int32_t mute;
			};
			u_int32_t status;
		};
	} __packed cmd;

	bzero(&cmd, sizeof cmd);
	cmd.avport = avport;
	cmd.mute = on ? PS3AV_CMD_MUTE_ON : PS3AV_CMD_MUTE_OFF;
	ps3av_cmd(CID_AUDIO_MUTE, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

int
ps3av_cmd_audio_active(active, port)
	int active;
	u_int port;
{
	struct {
		u_int16_t version;
		u_int16_t size;
		u_int32_t cid;
		union {
			u_int32_t audio_port;
			u_int32_t status;
		};
	} __packed cmd;
	u_int cid = active ? CID_AUDIO_ACTIVE : CID_AUDIO_INACTIVE;

	bzero(&cmd, sizeof cmd);
	cmd.audio_port = port;
	ps3av_cmd(cid, sizeof cmd, sizeof cmd, &cmd);
	if (cmd.status)
		return cmd.status;
	return 0;
}

#if 0
struct ps3av_pkt_avb_param {
	u_int16_t version;
	u_int16_t size;
	u_int32_t cid;
	union {
		struct {		/* SEND */
			u_int16_t n_video_pkt;
			u_int16_t n_audio_pkt;	/* always 0 (N/A) */
			u_int16_t n_av_video_pkt;
			u_int16_t n_av_audio_pkt;
			struct video_mode video[PS3AV_HEAD_MAX];
			struct av_video_cs av_video[PS3AV_AVPORT_MAX];
			struct av_audio_param av_audio[PS3AV_HDMI_MAX];
		};
		u_int32_t status;	/* RECV */
	};
} __packed;
#endif

static char *monitor_type(u_int);
static char *show_modes(u_int);

inline char *
monitor_type(monitor)
	u_int monitor;
{
	switch (monitor) {
	case PS3AV_MONITOR_HDMI:
		return "HDMI";
	case PS3AV_MONITOR_DVI:
		return "DVI";
	default:
		return "AVMULTI";
	}
}

char *
show_modes(mode)
	u_int mode;
{
	static char buf[32];	/* 1080P,1080I,720P,480P/576P */
	char *c = "";

	buf[0] = 0;
	if (mode & PS3AV_RESBIT_1080P) {
		strcat(buf, c);
		strcat(buf, "1080P");
		c = ",";
	}
	if (mode & PS3AV_RESBIT_1080I) {
		strcat(buf, c);
		strcat(buf, "1080I");
		c = ",";
	}
	if (mode & PS3AV_RESBIT_720P) {
		strcat(buf, c);
		strcat(buf, "720P");
		c = ",";
	}
	if (mode & PS3AV_RESBIT_480P) {
		strcat(buf, c);
		strcat(buf, "480P/576P");
	}
	return buf;
}

void
ps3av_dump_monitor_info(buf)
	void *buf;
{
	struct ps3av_monitor_info *mi = buf;
	/* struct ps3av_audio_info *ai = (void *)(mi + 1); */

	printf("avport: 0x%x\n", mi->avport);

#define EDID_VENDOR1(vendor) ('@' + ((vendor & 0x7c00) >> 10))
#define EDID_VENDOR2(vendor) ('@' + ((vendor & 0x03e0) >> 5))
#define EDID_VENDOR3(vendor) ('@' +  (vendor & 0x001f))

	printf("vendor: %c%c%c\n", EDID_VENDOR1(mi->vendor_id),
	    EDID_VENDOR2(mi->vendor_id), EDID_VENDOR3(mi->vendor_id));

	printf("product ID: 0x%x\n", le16toh(mi->product_id));
	printf("serial#: %d\n", le32toh(mi->serial));
	printf("week: %d\n", mi->week);
	printf("year: %d\n", mi->year + 1990);

	printf("monitor_type: %s (%d)\n", monitor_type(mi->monitor_type),
	    mi->monitor_type);
	printf("monitor_name: %.16s\n", mi->monitor_name);

	printf("supported modes:\n");
	printf("  60 native  0x%08x <%s>\n", mi->mode_60.native,
	    show_modes(mi->mode_60.native));
	printf("  60 support 0x%08x <%s>\n", mi->mode_60.support,
	    show_modes(mi->mode_60.support));

	printf("  50 native  0x%08x <%s>\n", mi->mode_50.native,
	    show_modes(mi->mode_50.native));
	printf("  50 support 0x%08x <%s>\n", mi->mode_50.support,
	    show_modes(mi->mode_50.support));

	printf("  other native 0x%08x support 0x%08x\n",
	    mi->mode_other.native, mi->mode_other.support);
	printf("   vesa native 0x%08x support 0x%08x\n",
	    mi->mode_vesa.native, mi->mode_vesa.support);

	if (0) {
	printf("color space:\n");
	printf("     rgb %d\n", mi->cs.rgb);
	printf("  yuv444 %d\n", mi->cs.yuv444);
	printf("  yuv422 %d\n", mi->cs.yuv422);

	printf("color info:\n");
	printf("    red X 0.%03d Y 0.%03d\n",
	    mi->color.red_x * 1000/1024, mi->color.red_y * 1000/1024);
	printf("  green X 0.%03d Y 0.%03d\n",
	    mi->color.green_x * 1000/1024, mi->color.green_y * 1000/1024);
	printf("   blue X 0.%03d Y 0.%03d\n",
	    mi->color.blue_x * 1000/1024, mi->color.blue_y * 1000/1024);
	printf("  white X 0.%03d Y 0.%03d\n",
	    mi->color.white_x * 1000/1024, mi->color.white_y * 1000/1024);
	printf("  gamma %d.%02d\n",
	    mi->color.gamma / 100, mi->color.gamma % 100);

	printf("supported_ai: 0x%02x\n", mi->supported_ai);
	printf("speaker_info: 0x%02x\n", mi->speaker_info);
	printf("n_audio: %d\n", mi->n_audio_block);
	}
	printf("\n");
}

struct vuart_send_hdr {
	u_int16_t version;
	u_int16_t size;
	u_int32_t cid;
} __packed;

struct vuart_reply_hdr {
	u_int16_t version;
	u_int16_t size;
	u_int32_t cid;
	u_int32_t status;
} __packed;

#define PS3AV_EVENT_CMD_MASK	0x10000000
#define PS3AV_EVENT_ID_MASK	0x0000ffff
#define PS3AV_REPLY_BIT		0x80000000

#define ps3av_event_get_port_id(cid) (((cid) >> 16) & 0xff)

static int vuart_write(void *, u_long);
static int vuart_read(void *, u_long);
static void xdelay(long);
static int send_cmd(void *, int, int);

static struct vuart {
	u_int port;
	u_int bufsize;
	paddr_t paddr;
	void *buf;
} ps3av_vuart;

void
ps3av_vuart_init()
{
	struct vuart *vuart = &ps3av_vuart;
	static char buf[4096];

	vuart->port = 0;
	vuart->bufsize = sizeof buf;
	vuart->paddr = (paddr_t)buf;
	vuart->buf = buf;
}

int
vuart_write(buf, size)
	void *buf;
	u_long size;
{
	struct vuart *vuart = &ps3av_vuart;
	u_long error, len;

	if (size > vuart->bufsize) {
		printf("vuart_write: size exceeds\n");
		size = vuart->bufsize;
	}

	bzero(vuart->buf, vuart->bufsize);	/* To be safe */
	bcopy(buf, vuart->buf, size);

	error = HV_write_vuart(vuart->port, vuart->paddr, size, &len);
	if (error) {
		printf("HV_write_vuart failed (error %ld)\n", error);
		return -1;
	}

	return len;
}

int
vuart_read(buf, size)
	void *buf;
	u_long size;
{
	struct vuart *vuart = &ps3av_vuart;
	u_long error, len = 0;
	long timo = 1000;	/* 1000 ms */

	if (size > vuart->bufsize) {
		printf("vuart_read: size exceeds\n");
		size = vuart->bufsize;
	}

	bzero(vuart->buf, vuart->bufsize);	/* To be safe. */

	while (timo > 0) {
		error = HV_read_vuart(vuart->port, vuart->paddr, size,
		    &len);
		if (error) {
			printf("HV_read_vuart failed (error %ld)\n", error);
			return -1;
		}
		if (len)
			goto done;
		xdelay(10);
		timo -= 10;
	}
	printf("vuart_read: timeout\n");
	return -1;
done:
	bcopy(vuart->buf, buf, len);
	return len;
}

/*
 * Wait about n ms.
 * This function is used for initializing console, when delay() is
 * not yet available.
 */
void
xdelay(n)
	long n;
{
	long x = 400000 * n;

	while (x-- > 0)
		asm volatile ("nop" :: "r"(x));
}

int
send_cmd(buf, sendlen, recvlen)
	void *buf;
	int sendlen, recvlen;
{
	struct vuart_send_hdr *send_hdr = buf;
	struct vuart_reply_hdr *reply_hdr = buf;
	u_int cmd = send_hdr->cid;

	if (vuart_write(buf, sendlen) < 0)
		return -1;
ignore:
	if (vuart_read(buf, recvlen) < 0)
		return -1;

	if (reply_hdr->size + PS3AV_HDR_SIZE > recvlen)
		printf("vuart: receive buffer overrun!\n");

	if (reply_hdr->cid & PS3AV_EVENT_CMD_MASK)
		goto ignore;

	if (reply_hdr->cid != (cmd | PS3AV_REPLY_BIT))
		return -1;

	if (reply_hdr->version != PS3AV_VERSION)
		return -1;

	return 0;
}

void
ps3av_set_hdr(cid, size, buf)
	u_int cid, size;
	void *buf;
{
	struct vuart_send_hdr *hdr = buf;

	hdr->version = PS3AV_VERSION;
	hdr->size = size - PS3AV_HDR_SIZE;
	hdr->cid = cid;
}

int
ps3av_cmd(cid, sendlen, recvlen, buf)
	u_int cid, sendlen;
	size_t recvlen;
	void *buf;
{
	ps3av_set_hdr(cid, sendlen, buf);
	send_cmd(buf, sendlen, recvlen);
	return 0;
}
