/*	$Id: ps3disk.c,v 1.55 2011/11/09 11:54:31 tsubai Exp $	*/

/*-
 * Copyright (c) 2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/buf.h>
#include <sys/callout.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

#include <uvm/uvm.h>

#include <dev/scsipi/scsi_all.h>
#include <dev/scsipi/scsi_disk.h>
#include <dev/scsipi/scsiconf.h>
#include <dev/scsipi/scsipi_all.h>
#include <dev/scsipi/scsipi_disk.h>

struct ps3d_scb {
	TAILQ_ENTRY(ps3d_scb) chain;
	struct scsipi_xfer *xs;
	u_int64_t tag;
	int need_copy;
};

struct ps3d_softc {
	struct device sc_dev;
	struct scsipi_channel sc_channel;
	struct scsipi_adapter sc_adapter;

	int sc_bustype;
	int sc_devid;
	int sc_region;
	int sc_blksize;
	u_long sc_off;
	u_long sc_nblks;
	struct scsipi_inquiry_data sc_inquiry;

	char *sc_devbuf;
	struct ps3d_scb *sc_nexus;	/* current command */
	TAILQ_HEAD(, ps3d_scb) free_scb;
	TAILQ_HEAD(, ps3d_scb) ready_scb;
#define PS3D_NSCB 16
	struct ps3d_scb sc_scb[PS3D_NSCB];
};

int ps3d_match(struct device *, struct cfdata *, void *);
void ps3d_attach(struct device *, struct device *, void *);
void ps3d_scsi_request(struct scsipi_channel *, scsipi_adapter_req_t, void *);
void ps3d_sched(struct ps3d_softc *);
int ps3d_intr(void *);
void ps3d_done(struct ps3d_softc *);
int ps3d_poll(struct ps3d_softc *, struct scsipi_xfer *);
void ps3d_timeout(void *);
struct ps3d_scb *ps3d_get_scb(struct ps3d_softc *);
void ps3d_free_scb(struct ps3d_softc *, struct ps3d_scb *);

/* Target emulations. */
static void inquiry(struct ps3d_softc *);
static void read_capacity(struct ps3d_softc *);
static void mode_sense(struct ps3d_softc *);
static void read(struct ps3d_softc *);
static void write(struct ps3d_softc *);
static void check_condition(struct ps3d_softc *, int, int, int);

/* Repository accesses. */
struct device_info {
	int port;
	int type;
	int regions;
	int blksize;
	u_long nblocks;
};
struct region_info {
	int id;
	u_long start;
	u_long size;
};
static int read_device_info(int, int, struct device_info *);
static int read_region_info(int, int, int, struct region_info *);

static int ps3d_atapi_cmd(struct ps3d_softc *);

CFATTACH_DECL0(ps3d, sizeof(struct ps3d_softc), ps3d_match, ps3d_attach);

int
ps3d_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;
	struct device_info device;

	if (ca->ca_bus != 4)
		return 0;

	if (read_device_info(ca->ca_bus, ca->ca_dev, &device) < 0)
		return 0;

	switch (device.type & SID_TYPE) {
	case T_DIRECT:
	case T_CDROM:
	case T_SIMPLE_DIRECT:
		return 1;
	}
	return 0;
}

void
ps3d_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct ps3d_softc *sc = (void *)self;
	struct confargs *ca = aux;
	int bus_id = ca->ca_bus;
	int dev_id = ca->ca_dev;
	int bus_index = bus_id;		/* XXX */
	int device_index = dev_id;	/* XXX */
	struct scsipi_inquiry_data *d = &sc->sc_inquiry;
	struct device_info device;
	struct region_info region;
	struct pglist mlist;
	int irq, err, i;

	sc->sc_adapter.adapt_dev = &sc->sc_dev;
	sc->sc_adapter.adapt_nchannels = 1;
	sc->sc_adapter.adapt_openings = 7;
	sc->sc_adapter.adapt_max_periph = 1;
	sc->sc_adapter.adapt_ioctl = NULL;
	sc->sc_adapter.adapt_minphys = minphys;
	sc->sc_adapter.adapt_request = ps3d_scsi_request;

	sc->sc_channel.chan_adapter = &sc->sc_adapter;
	sc->sc_channel.chan_bustype = &scsi_bustype;
	sc->sc_channel.chan_channel = 0;
	sc->sc_channel.chan_ntargets = 2;
	sc->sc_channel.chan_nluns = 1;
	sc->sc_channel.chan_id = 1;
	sc->sc_channel.chan_flags = SCSIPI_CHAN_NOSETTLE;

	TAILQ_INIT(&sc->free_scb);
	TAILQ_INIT(&sc->ready_scb);
	for (i = 0; i < PS3D_NSCB; i++)
		ps3d_free_scb(sc, &sc->sc_scb[i]);

	err = HV_open(ca->ca_bus, ca->ca_dev, 0);
	if (err && err != HV_EBUSY) {
		printf(": open failed (%d)\n", err);
		return;
	}

	if (read_device_info(bus_index, device_index, &device) < 0) {
		printf(": couldn't get device parameters\n");
		return;
	}
	device.type &= SID_TYPE;

	printf(": type %d, %d region(s), blksize %d, %ld blocks\n",
	    device.type, device.regions, device.blksize, device.nblocks);

	for (i = 0; i < device.regions; i++) {
		if (read_region_info(bus_index, device_index, i, &region) < 0)
			continue;
#if 0
		printf("  region[%d] id = %d, start = %ld, size = %ld\n",
		    i, region.id, region.start, region.size);
#endif
	}

	/* Search readable regions.  XXX How many regions? */
	for (i = 0; i < device.regions; i++) {
		u_int64_t tag;

		err = HV_storage_read(device_index, i, -1, 1, 0, 0, &tag);
		if (err == HV_EINVAL) {
			sc->sc_region = i;
			goto found;
		}
	}

	/* No readable region. */
	return;

found:
	d->device = device.type;	/* device type */
	d->version = 2;			/* ANSI version 2 */
	d->response_format = 2;		/* SCSI-2 */
	d->additional_length = SCSIPI_INQUIRY_LENGTH_SCSI2 - 4;

	memset(d->vendor, ' ',
	    sizeof d->vendor + sizeof d->product + sizeof d->revision);
	memcpy(d->vendor, "SCEI", 4);
	memcpy(d->revision, "4989", 4);

	sc->sc_devid = device_index;
	sc->sc_nblks = device.nblocks;
	sc->sc_blksize = device.blksize;

	/* XXX for now */
	switch (device.type) {
	case T_SIMPLE_DIRECT:
		sc->sc_bustype = SCSIPI_BUSTYPE_SCSI;
		d->device = T_DIRECT;
		memcpy(d->product, "PS3 FLASH", 9);
		break;
	case T_DIRECT:
		sc->sc_bustype = SCSIPI_BUSTYPE_SCSI;
		memcpy(d->product, "PS3 SATA", 8);
		break;
	case T_CDROM:
		sc->sc_bustype = SCSIPI_BUSTYPE_ATAPI;
		break;
	default:
		return;
	}

	ps3_allocate_dma_region(bus_id, dev_id, 0);

	irq = ps3_connect_irq(ca->ca_bus, ca->ca_dev, ca->ca_intr);
	if (irq < 0) {
		printf("%s: couldn't connect interrupt\n", self->dv_xname);
		return;
	}

	printf("%s: irq %d\n", self->dv_xname, irq);

	if (uvm_pglistalloc(MAXPHYS, 0, -1, 0, 0, &mlist, 1, 0)) {
		printf("%s: failed to allocate buffer\n", self->dv_xname);
		return;
	}
	sc->sc_devbuf = (void *)VM_PAGE_TO_PHYS(mlist.tqh_first);

	config_found(&sc->sc_dev, &sc->sc_channel, scsiprint);

	intr_establish(irq, IST_LEVEL, IPL_BIO, ps3d_intr, sc);
}

void
ps3d_scsi_request(chan, req, arg)
	struct scsipi_channel *chan;
	scsipi_adapter_req_t req;
	void *arg;
{
	struct scsipi_xfer *xs = arg;
	struct ps3d_softc *sc = (void *)chan->chan_adapter->adapt_dev;
	struct ps3d_scb *scb;
	int s;

	switch (req) {

	case ADAPTER_REQ_GROW_RESOURCES:
	case ADAPTER_REQ_SET_XFER_MODE:
	default:
		/* XXX Not supported. */
		xs->error = XS_DRIVER_STUFFUP;
		return;

	case ADAPTER_REQ_RUN_XFER:
		break;
	}

	if ((scb = ps3d_get_scb(sc)) == NULL) {
		xs->error = XS_RESOURCE_SHORTAGE;
		scsipi_done(xs);
		return;
	}
	scb->xs = xs;
	scb->need_copy = 0;

#if 0
{
	struct scsipi_generic *cmd = xs->cmd;
	int i;

	printf("ps3d_scsi_cmd: cdb = %02x", cmd->opcode);
	for (i = 0; i < 5; i++)
		printf(" %02x", cmd->bytes[i]);
	printf("\n");
}
#endif

	s = splbio();

	TAILQ_INSERT_TAIL(&sc->ready_scb, scb, chain);

	if (sc->sc_nexus == NULL)	/* IDLE */
		ps3d_sched(sc);

	splx(s);

	if ((xs->xs_control & XS_CTL_POLL) == 0)
		return;

	if (ps3d_poll(sc, xs)) {
		printf("%s: timeout\n", sc->sc_dev.dv_xname);
		if (ps3d_poll(sc, xs))
			printf("%s: timeout again\n", sc->sc_dev.dv_xname);
	}
}

void
ps3d_sched(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb;

next:
	scb = sc->ready_scb.tqh_first;
	if (scb == NULL)
		return;

	if (sc->sc_nexus == NULL) {
		struct scsipi_xfer *xs = scb->xs;
		struct scsipi_generic *cmd = xs->cmd;
		int timo = mstohz(xs->timeout) + 1;

		TAILQ_REMOVE(&sc->ready_scb, scb, chain);
		sc->sc_nexus = scb;

		callout_reset(&xs->xs_callout, timo, ps3d_timeout, scb);

#define TEST_UNIT_READY	0x00
#define READ_6		0x08
#define WRITE_6		0x0a
#define MODE_SENSE_6	0x1a
#define SYNCHRONIZE_CACHE_10 0x35
		if (sc->sc_bustype == SCSIPI_BUSTYPE_ATAPI) {
			/* Send all commands to drive except READ and WRITE. */
			switch (cmd->opcode) {
			case READ_6:
			case READ_10:
			case WRITE_6:
			case WRITE_10:
				break;
			default:
				ps3d_atapi_cmd(sc);
				goto next;
			}
		}

		switch (cmd->opcode) {
		case INQUIRY:
			inquiry(sc);
			break;

		case READ_CAPACITY_10:
			read_capacity(sc);
			break;

		case MODE_SENSE_6:
			mode_sense(sc);
			break;

		case READ_6:
		case READ_10:
			read(sc);
			break;

		case WRITE_6:
		case WRITE_10:
			write(sc);
			break;

		case SYNCHRONIZE_CACHE_10: /* 0x35 */
			check_condition(sc, 0, 0, 0);
			break;

		/* Ignore this for now. */
		case TEST_UNIT_READY:
			ps3d_done(sc);
			break;

		default:
			panic("unsupported scsi command (0x%x)", cmd->opcode);
		}
	}
	goto next;
}

int
ps3d_intr(arg)
	void *arg;
{
	struct ps3d_softc *sc = arg;
	struct ps3d_scb *scb = sc->sc_nexus;
	u_long tag, status;
	int rv;

	/* Clear interrupt. */
	rv = HV_storage_status(sc->sc_devid, &tag, &status);
	if (rv)
		return 0;
#ifdef DIAGNOSTIC
	if (scb->tag != tag)
		panic("%s: tag mismatch", sc->sc_dev.dv_xname);
#endif

	/* printf("done scb  %p\n", scb); */
	ps3d_done(sc);
	return 1;
}

void
inquiry(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;

	/* printf("INQUIRY\n"); */
	if (xs->datalen < SCSIPI_INQUIRY_LENGTH_SCSI2) {
		printf("short xs\n");
		xs->error = XS_DRIVER_STUFFUP;
		ps3d_done(sc);
		return;
	}
	bcopy(&sc->sc_inquiry, xs->data, SCSIPI_INQUIRY_LENGTH_SCSI2);

	ps3d_done(sc);
}

void
read_capacity(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;
	struct scsipi_read_capacity_10_data *reply = (void *)xs->data;

	/* printf("READ_CAPACITY\n"); */
	if (xs->datalen < sizeof *reply) {
		printf("short xs\n");
		xs->error = XS_DRIVER_STUFFUP;
		ps3d_done(sc);
		return;
	}
	_lto4b(sc->sc_nblks - 1, reply->addr);
	_lto4b(512, reply->length);

	ps3d_done(sc);
}

void
mode_sense(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;
	struct scsipi_generic *cmd = xs->cmd;

	/* printf("MODE SENSE\n"); */
	/* cmd: 1a 00 00 00 0c 00			*/
	/* cmd: 1a 08 04 00 1c 00	DBD=1 PAGE=4	*/
	if (cmd->bytes[1] == 0x00) {
		struct {
			struct scsi_mode_parameter_header_6 h;
			struct scsi_general_block_descriptor bd;
		} *reply = (void *)xs->data;

		/* printf("PAGE 0\n"); */
		if (xs->datalen < sizeof *reply) {
			printf("short xs\n");
			xs->error = XS_DRIVER_STUFFUP;
			ps3d_done(sc);
			return;
		}
		bzero(reply, sizeof *reply);
		reply->h.data_length = sizeof *reply - 1; /* param length */
		reply->h.medium_type = 0;
		reply->h.blk_desc_len = 8;		/* block desc len */
		_lto3b(sc->sc_nblks, reply->bd.nblocks);
		_lto3b(512, reply->bd.blklen);
		ps3d_done(sc);
		return;
	}
	if (cmd->bytes[1] == 0x04) {
		struct {
			struct scsi_mode_parameter_header_6 h;
			struct page_rigid_geometry p;
		} *reply = (void *)xs->data;
		int ncyl = (sc->sc_nblks + 1007) / 1008;

		/* printf("PAGE 4\n"); */
		if (xs->datalen < sizeof *reply) {
			printf("short xs\n");
			xs->error = XS_DRIVER_STUFFUP;
			ps3d_done(sc);
			return;
		}
		bzero(reply, sizeof *reply);
		reply->h.data_length = sizeof *reply - 1; /* mode param len */
		reply->h.blk_desc_len = 0;		/* no block desc */
		reply->p.pg_code = 4;			/* PAGE 4 */
		reply->p.pg_length = 0x16;
		_lto3b(ncyl, reply->p.ncyl);		/* cylinders */
		reply->p.nheads = 16;
		_lto2b(5400, reply->p.rpm);
		ps3d_done(sc);
		return;
	}
	check_condition(sc, 0, 0, 0);
}

void
read(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;
	u_long addr, len;
	u_int rv;

	addr = xs->bp->b_rawblkno;
	len = howmany(xs->datalen, sc->sc_blksize);

	/* printf("READ(6): addr 0x%lx, len 0x%lx\n", addr, len); */

	scb->need_copy = 1;
	/* dev, region, start, count, flags, addr, tag */
	rv = HV_storage_read(sc->sc_devid, sc->sc_region, sc->sc_off + addr,
	    len, 0, (paddr_t)sc->sc_devbuf, &scb->tag);
	if (rv)
		check_condition(sc, 0, 0x21, 0);
}

void
write(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;
	u_long addr, len;
	u_int rv;

	addr = xs->bp->b_rawblkno;
	len = howmany(xs->datalen, sc->sc_blksize);

	/* printf("WRITE(6): addr 0x%lx, len 0x%lx\n", addr, len); */

	bcopy(xs->data, sc->sc_devbuf, xs->datalen);
	/* dev, region, start, count, flags, addr, tag */
	rv = HV_storage_write(sc->sc_devid, sc->sc_region, sc->sc_off + addr,
	    len, 0, (paddr_t)sc->sc_devbuf, &scb->tag);
	if (rv)
		check_condition(sc, 0, 0x21, 0);
}

void
check_condition(sc, info, asc, ascq)
	struct ps3d_softc *sc;
	int info, asc, ascq;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;
#if 0
	struct scsi_sense_data *sense = &xs->sense.scsi_sense;

	bzero(sense, sizeof *sense);
	sense->response_code = SSD_RCODE_CURRENT;
	sense->flags = 5;		/* Illegal Request */
	_lto4b(info, sense->info);	/* sector number etc. */
	sense->extra_len = 10;
	sense->asc = asc;
	sense->ascq = ascq;

	xs->error = XS_SENSE;
#endif
	xs->status = SCSI_CHECK;
	ps3d_done(sc);
}

void
ps3d_done(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;

	callout_stop(&xs->xs_callout);
	sc->sc_nexus = NULL;

	if (scb->need_copy)
		bcopy(sc->sc_devbuf, xs->data, xs->datalen);

	if (xs->status == SCSI_OK)
		xs->resid = 0;

	if ((xs->xs_control & XS_CTL_POLL) == 0)
		ps3d_sched(sc);

	scsipi_done(xs);
	ps3d_free_scb(sc, scb);
}

int
ps3d_poll(sc, xs)
	struct ps3d_softc *sc;
	struct scsipi_xfer *xs;
{
	int count = xs->timeout;

	while (count) {
		ps3d_intr(sc);

		if (xs->xs_status & XS_STS_DONE)
			return 0;
		delay(1000);
		count--;
	};
	return 1;
}

void
ps3d_timeout(arg)
	void *arg;
{
	struct ps3d_scb *scb = arg;
	struct ps3d_softc *sc = (void *)
	    scb->xs->xs_periph->periph_channel->chan_adapter->adapt_dev;
	int s;

	printf("%s: timeout\n", sc->sc_dev.dv_xname);

	s = splbio();
	scb->xs->error = XS_TIMEOUT;
	ps3d_done(sc);
	splx(s);
}

struct ps3d_scb *
ps3d_get_scb(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb;
	int s;

	s = splbio();
	if ((scb = sc->free_scb.tqh_first) != NULL)
		TAILQ_REMOVE(&sc->free_scb, scb, chain);
	splx(s);

	return scb;
}

void
ps3d_free_scb(sc, scb)
	struct ps3d_softc *sc;
	struct ps3d_scb *scb;
{
	int s;

	s = splbio();
	TAILQ_INSERT_HEAD(&sc->free_scb, scb, chain);
	splx(s);
}

int
read_device_info(bus_index, device_index, info)
	int bus_index, device_index;
	struct device_info *info;
{
	u_long a1, a2, a3;
	struct repository_value v;

	/* device port */
	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("port");
	if (HV_read_repository(1, a1, a2, a3, 0, &v))
		return -1;
	info->port = v.value1;

	/* device type */
	a3 = key2("type");
	if (HV_read_repository(1, a1, a2, a3, 0, &v))
		return -1;
	info->type = v.value1;

	/* number of regions */
	a3 = key2("n_regs");
	if (HV_read_repository(1, a1, a2, a3, 0, &v))
		return -1;
	info->regions = v.value1;

	/* block size */
	a3 = key2("blk_size");
	if (HV_read_repository(1, a1, a2, a3, 0, &v))
		return -1;
	info->blksize = v.value1;

	/* number of logical blocks */
	a3 = key2("n_blocks");
	if (HV_read_repository(1, a1, a2, a3, 0, &v))
		return -1;
	info->nblocks = v.value1;

	return 0;
}

int
read_region_info(bus_index, device_index, region_index, info)
	int bus_index, device_index, region_index;
	struct region_info *info;
{
	u_long a1, a2, a3, a4;
	struct repository_value v;

	/* region id */
	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("region") | region_index;
	a4 = key2("id");
	if (HV_read_repository(1, a1, a2, a3, a4, &v))
		return -1;
	info->id = v.value1;

	/* start offset */
	a4 = key2("start");
	if (HV_read_repository(1, a1, a2, a3, a4, &v))
		return -1;
	info->start = v.value1;

	/* region size */
	a4 = key2("size");
	if (HV_read_repository(1, a1, a2, a3, a4, &v))
		return -1;
	info->size = v.value1;

	return 0;
}

struct HV_atapi_command {
	char cmd[32];		/* ATAPI command */
	u_int32_t cmdlen;	/* command length (should be 12) */
	u_int32_t blocks;
	u_int32_t block_size;
	u_int32_t mode;		/* transfer mode - 1:read, 2:write (?) */
	u_int32_t rw;		/* 0:write, 1:read */
	paddr_t data;
	u_int32_t datalen;	/* data length */
};

#define SEND_ATAPI_COMMAND 1

int
ps3d_atapi_cmd(sc)
	struct ps3d_softc *sc;
{
	struct ps3d_scb *scb = sc->sc_nexus;
	struct scsipi_xfer *xs = scb->xs;
	struct HV_atapi_command packet;
	int rv;

	bzero(&packet, sizeof packet);
	bcopy(xs->cmd, packet.cmd, 12);
	packet.cmdlen = 12;
	packet.mode = 1;
	packet.rw = 1;

	packet.blocks = 1;
	packet.block_size = packet.datalen = xs->datalen;
	packet.data = vtophys((vaddr_t)xs->data);

	rv = HV_storage_sendcmd(sc->sc_devid, SEND_ATAPI_COMMAND,
	    vtophys((vaddr_t)&packet), sizeof packet,
	    packet.data, packet.datalen, &scb->tag);
	if (rv)
		check_condition(sc, 0, 0, 0);

	return rv;
}

#if 0

port: 0 = primary master
      1 = primary slave
      2 = secondary master
      3 = secondary slave

/* NODEVICE */
ps3d0 at obio0: port 0, type 255, blksize 2048, 2147483647 blocks, 1 region
  region[0] id = 0, start = 0, size = 2147483647

/* FLASH? */
ps3d1 at obio0: port 0, type 14, blksize 512, 491008 blocks, 7 regions
  region[0] id = 0, start = 0, size = 491008
  region[1] id = 1, start = 1, size = 30207
  region[2] id = 2, start = 30720, size = 409088
  region[3] id = 3, start = 439808, size = 32768
  region[4] id = 4, start = 472576, size = 1024
  region[5] id = 5, start = 473600, size = 8192 (*)
  region[6] id = 6, start = 490496, size = 512

/* SATA */
ps3d2 at obio0: port 2, type 0, blksize 512, 117210240 blocks, 4 regions
  region[0] id = 0, start = 0, size = 117210240
  region[1] id = 1, start = 8, size = 20971512
  region[2] id = 2, start = 20971528, size = 4194296
  region[3] id = 3, start = 25165832, size = 92044408 (*)

/* CDROM */
ps3d3 at obio0: port 0, type 5, blksize 2048, 2147483647 blocks, 1 region
  region[0] id = 0, start = 0, size = 2147483647

sd0 at scsibus0 target 0 lun 0: <SCEI, PS3 FLASH, 4989> simplified direct fixed
sd0: 239 MB, 239 cyl, 64 head, 32 sec, 512 bytes/sect x 491008 sectors
sd1 at scsibus1 target 0 lun 0: <SCEI, PS3 SATA, 4989> disk fixed
sd1: 57231 MB, 116280 cyl, 16 head, 63 sec, 512 bytes/sect x 117210240 sectors
cd0 at scsibus2 target 0 lun 0: <SONY, PS-SYSTEM   302R, 4062> cdrom removable

#endif
