/*	$Id: rsxfb.c,v 1.87 2011/11/10 04:11:00 tsubai Exp $	*/

/*-
 * Copyright (c) 2006,2007,2011 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_rsxfb_mode.h"

#include <sys/param.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/systm.h>

#include <uvm/uvm_extern.h>

#include <dev/wscons/wsconsio.h>
#include <dev/wscons/wsdisplayvar.h>
#include <dev/wsfont/wsfont.h>

#include <machine/autoconf.h>
#include <machine/cpufunc.h>
#include <machine/hcalls.h>
#include <machine/ps3av.h>
#include <machine/rsxfbio.h>

/* gpu_context_attribute() values */
#define GPU_FIFO_SETUP		0x001
#define GPU_DISPLAY_MODE_SET	0x100
#define GPU_DISPLAY_FLIP	0x102
#define GPU_FB_SETUP		0x600
#define GPU_FB_BLIT		0x601

#define GPU_DMABASE	0x0d000000	/* iommu window */
#define DMA_VRAM	0xfeed0000	/* DMA from VRAM */
#define DMA_SYSTEM	0xfeed0001	/* DMA from system memory */
#define SURFACE_2D	0x313371c3	/* Graphic object */
#define IMAGE_FROM_CPU	0x31337808	/* Graphic object */
#define SCALED_IMAGE	0x3137af00	/* Graphic object */
#define CONTEXT_ROP	0x80000011	/* Graphic object */
#define IMAGE_PATTERN	0x80000012	/* Graphic object */
#define DMA_NOTIFIER_0	0x66604200	/* DMA notifier */

#define ROP_CLEAR	0x00	/* Always 0 */
#define ROP_INVCOPY	0x30	/* Copy inverted */
#define ROP_INVERT	0x50	/* Invert destination */
#define ROP_COPY	0xc0	/* Copy source */
#define ROP_SET		0xf0	/* Use PATTERN */

#define ATTR_REVERSE	1
#define ATTR_UNDERLINE	2

#define RSXFB_MMAP_VRAM	0x0
#define RSXFB_MMAP_FIFO	0x40000000
#define RSXFB_MMAP_REG	0x41000000

#define DDR_SIZE (128 * 0x100000)	/* Use 128 MB of VRAM. */
#define FIFO_SIZE 65536			/* 64KB FIFO */

struct rasops_info {
	struct wsdisplay_font *ri_font;
	int ri_width;
	int ri_height;
	int ri_stride;
	int ri_rows;
	int ri_cols;
	int ri_xscale;
	int ri_yscale;
	int ri_offset;
};

struct rsxfb_devconfig {		/* device configuration */
	struct rasops_info dc_ri;
	paddr_t dc_vram;		/* PA of video RAM */
	paddr_t dc_bits;
	paddr_t dc_fiforeg;
	paddr_t dc_fifo;
	u_int64_t dc_handle;		/* HV context handle */
	void *dc_reg;			/* FIFO register */
	u_int dc_width;
	u_int dc_height;
	u_int dc_bpp;
	u_int dc_dmaput;
	u_int dc_gpufifo;		/* GPU view of dc_fifo */
	u_int dc_gpubits;		/* GPU view of dc_bits */
};

struct rsxfb_softc {
	struct device sc_dev;
	struct rsxfb_devconfig *sc_dc;
	int sc_nscreens;
};

int rsxfbmatch(struct device *, struct cfdata *, void *);
void rsxfbattach(struct device *, struct device *, void *);

void rsxfb_init(struct rsxfb_devconfig *);
int rsxfb_set_mode(struct rsxfb_devconfig *, int);
void rsxfb_gpu_init(void);
void rsxfb_accel_init(struct rsxfb_devconfig *);
int rsxfb_cnattach(void);
int rsxfb_is_console(int);
int rsxfb_ioctl(void *, void *, u_long, void *, int, struct lwp *);
paddr_t rsxfb_mmap(void *, void *, off_t, int);
int rsxfb_alloc_screen(void *, const struct wsscreen_descr *, void **, int *,
    int *, long *);
void rsxfb_free_screen(void *, void *);
int rsxfb_show_screen(void *, void *, int, void (*)(void *, int, int), void *);

void rsxfb_cursor(void *, int, int, int);
int rsxfb_mapchar(void *, int, u_int *);
void rsxfb_putchar(void *, int, int, u_int, long);
void rsxfb_copycols(void *, int, int, int, int);
void rsxfb_erasecols(void *, int, int, int, long);
void rsxfb_copyrows(void *, int, int, int);
void rsxfb_eraserows(void *, int, int, long);
int rsxfb_allocattr(void *, int, int, int, long *);

static void drawchar(struct rasops_info *, int, u_int, long);
static void fullclear(struct rasops_info *);
static int blit_test(struct rsxfb_devconfig *, struct blit_param *);

#ifdef RSXFB_MODE
int rsxfb_default_mode = RSXFB_MODE;
#else
int rsxfb_default_mode = 0;
#endif

CFATTACH_DECL0(rsxfb, sizeof(struct rsxfb_softc), rsxfbmatch, rsxfbattach);

struct rsxfb_devconfig rsxfb_console_dc;

struct wsdisplay_emulops rsxfb_emulops = {
	rsxfb_cursor,
	rsxfb_mapchar,
	rsxfb_putchar,
	rsxfb_copycols,
	rsxfb_erasecols,
	rsxfb_copyrows,
	rsxfb_eraserows,
	rsxfb_allocattr,
	NULL,	/* replaceattr */
};

struct wsscreen_descr rsxfb_stdscreen = {
	"std", 80, 24,
	&rsxfb_emulops,
	0, 0,
	WSSCREEN_REVERSE /* | WSSCREEN_UNDERLINE */
};

struct wsdisplay_accessops rsxfb_accessops = {
	rsxfb_ioctl,
	rsxfb_mmap,
	rsxfb_alloc_screen,
	rsxfb_free_screen,
	rsxfb_show_screen,
	0 /* load_font */
};

const struct wsscreen_descr *rsxfb_scrlist[] = { &rsxfb_stdscreen };
struct wsscreen_list rsxfb_screenlist = { 1, rsxfb_scrlist };

#ifndef RSXFB_ACCEL
static char textvram[75][240];	/* 240x75 at 1920x1200 */
static char textattr[75][240];
#endif

int
rsxfbmatch(parent, match, aux)
	struct device *parent;
	struct cfdata *match;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "rsxfb") == 0)
		return 1;
	return 0;
}

void
rsxfbattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct rsxfb_softc *sc = (struct rsxfb_softc *)self;
	struct wsemuldisplaydev_attach_args a;
	struct rsxfb_devconfig *dc;
	int console;

	console = 1;

	if (console) {
		dc = &rsxfb_console_dc;
		sc->sc_nscreens = 1;
	}

	sc->sc_dc = dc;

	printf(": %d x %d, %dbpp\n", dc->dc_width, dc->dc_height, dc->dc_bpp);

#ifdef RSXFB_ACCEL
	dc->dc_reg = mapiodev(dc->dc_fiforeg, NBPG, 0);
#endif

	a.console = console;
	a.scrdata = &rsxfb_screenlist;
	a.accessops = &rsxfb_accessops;
	a.accesscookie = sc;

	config_found(self, &a, wsemuldisplaydevprint);
}

void
rsxfb_init(dc)
	struct rsxfb_devconfig *dc;
{
	struct rasops_info *ri = &dc->dc_ri;
	int font, mode;

	ps3av_init();
	rsxfb_gpu_init();
	wsfont_init();

	font = wsfont_find(NULL, 0, 0, 0, WSDISPLAY_FONTORDER_L2R,
	    WSDISPLAY_FONTORDER_L2R);
	if (font < 0) {
		printf("rsxfb: failed to find font\n");
		return;
	}
	wsfont_lock(font, &ri->ri_font);

	if (rsxfb_default_mode)
		mode = rsxfb_default_mode;
	else
		mode = ps3av_get_video_mode();
	rsxfb_set_mode(dc, mode);
#ifdef RSXFB_ACCEL
	rsxfb_accel_init(dc);
#endif
	fullclear(ri);
}

int
rsxfb_set_mode(dc, mode)
	struct rsxfb_devconfig *dc;
	int mode;
{
	struct rasops_info *ri = &dc->dc_ri;
	int width, height, depth, linebytes;
	int cols, rows;

	if (ps3av_set_video_mode(mode))
		return -1;
	ps3av_screen_size(mode, &width, &height);
	dc->dc_width = width;
	dc->dc_height = height;
	dc->dc_bpp = 32;

	depth = dc->dc_bpp / 8;
	linebytes = width * depth;

	cols = (width - 32) / ri->ri_font->fontwidth;
	rows = (height - 4) / ri->ri_font->fontheight;
	cols &= ~7;
	if (cols > 240)
		cols = 240;
	if (rows > 75)
		rows = 75;

	switch (mode) {
	case PS3AV_CMD_VID_480I:
	case PS3AV_CMD_VID_576I:
		ri->ri_offset = linebytes * 16 + 4 * 48;
		rows--;
		break;
	default:
		ri->ri_offset = linebytes * 2 + 64;	/* border */
	}

	ri->ri_width = width;
	ri->ri_height = height;
	ri->ri_stride = linebytes;
	ri->ri_xscale = ri->ri_font->fontwidth * depth;
	ri->ri_yscale = ri->ri_font->fontheight * linebytes;
	ri->ri_rows = rows;
	ri->ri_cols = cols;
	rsxfb_stdscreen.nrows = rows;
	rsxfb_stdscreen.ncols = cols;
	return 0;
}

int
rsxfb_cnattach()
{
	struct rsxfb_devconfig *dc = &rsxfb_console_dc;
	struct rasops_info *ri = &dc->dc_ri;
	long defattr;

	rsxfb_init(dc);

	rsxfb_allocattr(ri, 0, 0, 0, &defattr);
	wsdisplay_cnattach(&rsxfb_stdscreen, ri, 0, 0, defattr);

	return 0;
}

int
rsxfb_is_console(node)
	int node;
{
	return 1;
}

int
rsxfb_ioctl(v, vs, cmd, data, flag, l)
	void *v, *vs;
	u_long cmd;
	void *data;
	int flag;
	struct lwp *l;
{
	struct rsxfb_softc *sc = v;
	struct rsxfb_devconfig *dc = sc->sc_dc;
	struct rsxfbparam *rp;
	struct wsdisplay_fbinfo *wsi;
	int s;

	switch (cmd) {
	case WSDISPLAYIO_GTYPE:
		*(u_int *)data = 0;
		return 0;

	case WSDISPLAYIO_GINFO:
		wsi = data;
		wsi->height = dc->dc_height;
		wsi->width = dc->dc_width;
		wsi->depth = dc->dc_bpp;
		wsi->cmsize = 0;
		return 0;

	case WSDISPLAYIO_SVIDEO:
		if (*(int *)data)
			ps3av_set_video_mode(ps3av_get_video_mode());
		else
			ps3av_video_signal_off();
		/* XXX head? */
		return 0;

	case WSDISPLAYIO_SMODE:
		if (*(int *)data == WSDISPLAYIO_MODE_EMUL) {
			s = splhigh();
			HV_gpu_context_attribute(dc->dc_handle, GPU_FIFO_SETUP,
			    dc->dc_dmaput, dc->dc_dmaput, dc->dc_dmaput, 0);
			splx(s);
		}
		break;

	case RSXFBIOCGMODE:
		*(int *)data = ps3av_get_video_mode();
		return 0;

	case RSXFBIOCSMODE:
		if (rsxfb_set_mode(dc, *(int *)data))
			return EINVAL;
		return 0;

	case RSXFBIOCGPARAM:
		rp = data;
		rp->vram = RSXFB_MMAP_VRAM;
		rp->fifo = RSXFB_MMAP_FIFO;
		rp->fiforeg = RSXFB_MMAP_REG;
		rp->notifier = 0;
		rp->vram_size = DDR_SIZE >> PGSHIFT;
		rp->fifo_size = FIFO_SIZE >> PGSHIFT;
		rp->fiforeg_size = 1;
		rp->notifier_size = 0;
		rp->fifo_gpu = dc->dc_gpufifo;
		return 0;

	case RSXFBIOCBLIT:
		return blit_test(dc, data);
	}

	return EPASSTHROUGH;
}

paddr_t
rsxfb_mmap(v, vs, offset, prot)
	void *v, *vs;
	off_t offset;
	int prot;
{
	struct rsxfb_softc *sc = v;
	struct rsxfb_devconfig *dc = sc->sc_dc;

	if (offset < 0)
		return -1;

	if (offset == RSXFB_MMAP_REG)
		return dc->dc_fiforeg;

	if (offset >= RSXFB_MMAP_FIFO) {
		offset -= RSXFB_MMAP_FIFO;
		if (offset >= FIFO_SIZE)
			return -1;
		return dc->dc_fifo + offset;
	}

	if (offset >= RSXFB_MMAP_VRAM) {
		if (dc->dc_vram == 0)
			return -1;
		offset -= RSXFB_MMAP_VRAM;
		if (offset >= DDR_SIZE)
			return -1;
		return dc->dc_vram + offset;
	}
	return -1;
}

int
rsxfb_alloc_screen(v, type, cookiep, curxp, curyp, attrp)
	void *v;
	const struct wsscreen_descr *type;
	void **cookiep;
	int *curxp, *curyp;
	long *attrp;
{
	struct rsxfb_softc *sc = v;
	struct rasops_info *ri = &sc->sc_dc->dc_ri;

	if (sc->sc_nscreens > 0)
		return ENOMEM;

	*cookiep = ri;
	*curxp = 0;
	*curyp = 0;
	rsxfb_allocattr(ri, 0, 0, 0, attrp);
	sc->sc_nscreens++;
	return 0;
}

void
rsxfb_free_screen(v, cookie)
	void *v;
	void *cookie;
{
	struct rsxfb_softc *sc = v;

	if (sc->sc_dc == &rsxfb_console_dc)
		panic("rsxfb_free_screen: console");

	sc->sc_nscreens--;
}

int
rsxfb_show_screen(v, cookie, waitok, cb, cbarg)
	void *v;
	void *cookie;
	int waitok;
	void (*cb)(void *, int, int);
	void *cbarg;
{
	return 0;
}

void
rsxfb_gpu_init()
{
	struct rsxfb_devconfig *dc = &rsxfb_console_dc;
	u_int64_t mem_handle, dma_control, driver_info, reports, reports_size;
	paddr_t base;
#ifdef RSXFB_ACCEL
	static char fifo[8192 + FIFO_SIZE + 4096];
#else
	static char fifo[8192 + FIFO_SIZE + 1920*16*4];
#endif

	HV_gpu_open(0);

	/* HEAD 0 */
	HV_gpu_context_attribute(0, GPU_DISPLAY_MODE_SET, 0, 0, 0, 0);

	/* HEAD 1 */
	HV_gpu_context_attribute(0, GPU_DISPLAY_MODE_SET, 0, 0, 1, 0);

#ifdef RSXFB_ACCEL
	/* Allocate video RAM on RSX. */
	HV_gpu_alloc_memory(0, 0, 0, 0, 0, &mem_handle, &dc->dc_vram);
#else
	/* Allocate video RAM on RSX (128 MB for now). */
	HV_gpu_alloc_memory(DDR_SIZE, 0, 0, 0, 0, &mem_handle, &dc->dc_vram);
#endif

	/* Get GPU context handle. */
	HV_gpu_alloc_context(mem_handle, 0, &dc->dc_handle,
	    &dma_control, &driver_info, &reports, &reports_size);

	dc->dc_fiforeg = dma_control;
	dc->dc_fifo = roundup((vaddr_t)fifo, 8192);
	dc->dc_bits = dc->dc_fifo + FIFO_SIZE;

	/* iomap needs 1MB boundary. */
	base = dc->dc_fifo & ~0xfffff;
	dc->dc_gpufifo = GPU_DMABASE + dc->dc_fifo - base;
	dc->dc_gpubits = GPU_DMABASE + dc->dc_bits - base;
	dc->dc_dmaput = dc->dc_gpufifo;

	HV_gpu_context_iomap(dc->dc_handle, GPU_DMABASE, base, 2048*1024, 0);
	HV_gpu_context_attribute(dc->dc_handle, GPU_FB_SETUP,
	    dc->dc_fifo, FIFO_SIZE, dc->dc_gpufifo, 0);

	/* XXX base2MBRAMGPU饢ǤƤޤΤ! */

	/* HEAD 0 */
	HV_gpu_context_attribute(dc->dc_handle, GPU_DISPLAY_FLIP, 0, 0, 0, 0);

	/* HEAD 1 */
	HV_gpu_context_attribute(dc->dc_handle, GPU_DISPLAY_FLIP, 1, 0, 0, 0);

#if 0
	delay(100000);

	/* Unblank display. */
	ps3av_video_mute(0);
#endif
}

#define RECT(x, y, sync) ((u_int64_t)(sync) << 32 | (x) << 16 | (y))

void
drawchar(ri, col, c, attr)
	struct rasops_info *ri;
	int col;
	u_int c;
	long attr;
{
	struct rsxfb_devconfig *dc = (void *)ri;
	struct wsdisplay_font *f = ri->ri_font;
	u_char *addr, *fontp;
	u_int32_t *p;
	u_int w, h, pattern, fontstride, stride;
	int i;

	if (c < f->firstchar)
		return;
	if (c >= f->firstchar + f->numchars)
		return;

	h = f->fontheight;
	w = f->fontwidth;
	c -= f->firstchar;

#ifdef RSXFB_ACCEL
	stride = ri->ri_xscale;
#else
	stride = ri->ri_stride;
#endif
	fontstride = f->stride;
	addr = (char *)dc->dc_bits + col * ri->ri_xscale;
	fontp = (char *)f->data + fontstride * f->fontheight * c;

	while (h-- > 0) {
		p = (void *)addr;
		pattern = *(u_int *)fontp;
		if (attr & ATTR_REVERSE)
			pattern ^= 0xffffffff;
		for (i = 0; i < w; i++) {
			if (pattern & 0x80000000)
				p[i] = 0xffffff;
			else
				p[i] = 0;
			pattern <<= 1;
		}
		addr += stride;
		fontp += fontstride;
	}
}

#ifndef RSXFB_ACCEL
static void displine(struct rasops_info *, int);
static void dispall(struct rasops_info *);

void
fullclear(ri)
	struct rasops_info *ri;
{
	struct rsxfb_devconfig *dc = (void *)ri;
	int y;

	memset((void *)dc->dc_bits, 0xff, 1920*16*4);
	for (y = 0; y < ri->ri_height; y += 16) {
		HV_gpu_context_attribute(dc->dc_handle, GPU_FB_BLIT,
		    y * ri->ri_stride, dc->dc_gpubits,
		    RECT(ri->ri_width, 16, 1), ri->ri_stride);
	}
	memset(textvram, ' ', sizeof textvram);
	memset(textattr, 0,   sizeof textattr);
}

void
displine(ri, row)
	struct rasops_info *ri;
	int row;
{
	struct rsxfb_devconfig *dc = (void *)ri;
	u_int c, attr;
	int col;

	for (col = 0; col < ri->ri_cols; col++) {
		c    = textvram[row][col];
		attr = textattr[row][col] ^ ATTR_REVERSE;
		drawchar(ri, col, c, attr);
	}

	HV_gpu_context_attribute(dc->dc_handle, GPU_FB_BLIT,
	    ri->ri_offset + row * ri->ri_yscale, dc->dc_gpubits,
	    RECT(ri->ri_width, ri->ri_font->fontheight, 1), ri->ri_stride);
}

void
dispall(ri)
	struct rasops_info *ri;
{
	int row;

	for (row = 0; row < ri->ri_rows; row++)
		displine(ri, row);
}

void
rsxfb_cursor(cookie, on, row, col)
	void *cookie;
	int on, row, col;
{
	textattr[row][col] ^= ATTR_REVERSE;
	displine(cookie, row);
}
#endif

int
rsxfb_mapchar(cookie, c, cp)
	void *cookie;
	int c;
	u_int *cp;
{
	struct rasops_info *ri = cookie;

	if (c < ri->ri_font->firstchar) {
		*cp = ' ';
		return 0;
	}
	*cp = c;
	return 5;
}

#ifndef RSXFB_ACCEL
void
rsxfb_putchar(cookie, row, col, c, attr)
	void *cookie;
	int row, col;
	u_int c;
	long attr;
{
	textvram[row][col] = c;
	textattr[row][col] = attr;
	displine(cookie, row);
}

void
rsxfb_copycols(cookie, row, src, dst, num)
	void *cookie;
	int row, src, dst, num;
{
	bcopy(&textvram[row][src], &textvram[row][dst], num);
	bcopy(&textattr[row][src], &textattr[row][dst], num);
	displine(cookie, row);
}

void
rsxfb_erasecols(cookie, row, col, num, attr)
	void *cookie;
	int row, col, num;
	long attr;
{
	memset(&textvram[row][col], ' ',  num);
	memset(&textattr[row][col], attr, num);
	displine(cookie, row);
}

void
rsxfb_copyrows(cookie, src, dst, num)
	void *cookie;
	int src, dst, num;
{
	bcopy(textvram[src], textvram[dst], sizeof textvram[0] * num);
	bcopy(textattr[src], textattr[dst], sizeof textattr[0] * num);
	dispall(cookie);
}

void
rsxfb_eraserows(cookie, row, num, attr)
	void *cookie;
	int row, num;
	long attr;
{
	memset(textvram[row], ' ',  sizeof textvram[0] * num);
	memset(textattr[row], attr, sizeof textattr[0] * num);
	dispall(cookie);
}
#endif

int
rsxfb_allocattr(cookie, fg, bg, flag, attr)
	void *cookie;
	int fg, bg, flag;
	long *attr;
{
	u_int rev, ul;

	if (flag & (WSATTR_BLINK | WSATTR_HILIT | WSATTR_WSCOLORS))
		return EINVAL;
	fg = (flag & WSATTR_REVERSE) ? 0 : 1;
	bg = (flag & WSATTR_REVERSE) ? 1 : 0;
	rev = (flag & WSATTR_REVERSE) ? ATTR_REVERSE : 0;
	ul = (flag & WSATTR_UNDERLINE) ? ATTR_UNDERLINE : 0;
	*attr = fg << 24 | bg << 16 | ul | rev;
	return 0;
}

int
blit_test(dc, p)
	struct rsxfb_devconfig *dc;
	struct blit_param *p;
{
	int rv;

	printf("blit_test 0x%lx %ld 0x%lx 0x%lx %ld, ",
	    p->param1, p->param2, p->param3, p->param4, p->param5);

	rv = HV_gpu_context_attribute(dc->dc_handle,
	    p->param1, p->param2, p->param3, p->param4, p->param5);
	printf("rv = %d\n", rv);

	return 0;
}

#ifdef RSXFB_ACCEL
struct fiforeg {
	u_int32_t pad1[0x10];
	volatile u_int32_t put;
	volatile u_int32_t get;
	u_int32_t pad2[0x03];
	volatile u_int32_t top;
};

static void hash_insert(struct rsxfb_devconfig *, u_int, u_int, void *);
static void patch(struct rsxfb_devconfig *, u_int, void *, u_int);
static void docopy(struct rsxfb_devconfig *, u_int, u_int, u_int, u_int, u_int);

static char context_rop[32] = {		/* NV03_CONTEXT_ROP */
	0x43, 0x00, 0x08, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01
};
static char image_pattern[32] = {	/* NV04_IMAGE_PATTERN */
	0x44, 0x00, 0x08, 0x02, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01
};

#define FONTBASE 0x0fb00000	/* Start of font cache. */
#define RAMIN    0x0ff80000
#define RAMHT    0x0ff90000

#define OUT_RING(data) dmabase[dmacurrent++] = (data)
#define BEGIN_RING(chan, tag, size) \
    OUT_RING(((size) << 18) | ((chan) << 13) | (tag))

void
rsxfb_accel_init(dc)
	struct rsxfb_devconfig *dc;
{
	struct rasops_info *ri = &dc->dc_ri;
	struct wsdisplay_font *f = ri->ri_font;
	u_int c, fsize;
	u_int32_t *dmabase = (void *)dc->dc_fifo;
	int dmacurrent = 0;

	hash_insert(dc, CONTEXT_ROP,   0x14020, context_rop);
	hash_insert(dc, IMAGE_PATTERN, 0x14040, image_pattern);

	/* Default parameters.  Ǹľ褦! */

	BEGIN_RING(1, 0, 1);
	OUT_RING(CONTEXT_ROP);		/* NV03_CONTEXT_ROP */
	BEGIN_RING(1, 0x300, 1);
	OUT_RING(ROP_COPY);

	BEGIN_RING(1, 0, 1);
	OUT_RING(IMAGE_PATTERN);	/* NV04_IMAGE_PATTERN */
	BEGIN_RING(1, 0x300, 1);
	OUT_RING(3);			/* FORMAT = A8R8G8B8 */
	BEGIN_RING(1, 0x310, 4);
	OUT_RING(0xffffffff);		/* COLOR0 */
	OUT_RING(0xffffffff);		/* COLOR1 */
	OUT_RING(0xffffffff);		/* PATTERN0 */
	OUT_RING(0xffffffff);		/* PATTERN1 */

	BEGIN_RING(6, 0, 1);
	OUT_RING(SCALED_IMAGE);		/* NV40_SCALED_IMAGE_FROM_MEMORY */
	BEGIN_RING(6, 0x184, 3);
	OUT_RING(DMA_VRAM);
	OUT_RING(IMAGE_PATTERN);	/* PATTERN */
	OUT_RING(CONTEXT_ROP);		/* ROP */
	BEGIN_RING(6, 0x198, 1);
	OUT_RING(SURFACE_2D);		/* SURFACE */

	BEGIN_RING(6, 0x2fc, 3);
	OUT_RING(1);			/* COLOR_CONVERSION */
	OUT_RING(4);			/* FORMAT = X8R8G8B8 */
	OUT_RING(3);			/* OPERATION = SRCCOPY */

	/* XXX FORMAT = ARGB  ROP ưʤ褦 */

	/* Kick DMA. */
	HV_gpu_context_attribute(dc->dc_handle, GPU_FIFO_SETUP,
	    dc->dc_gpufifo, dc->dc_gpufifo, dc->dc_gpufifo, 0);
	dc->dc_dmaput = dc->dc_gpufifo + dmacurrent * 4;
	HV_gpu_context_attribute(dc->dc_handle, GPU_FIFO_SETUP,
	    dc->dc_dmaput, dc->dc_gpufifo, dc->dc_gpufifo, 0);

	fsize = ri->ri_xscale * f->fontheight;
	for (c = f->firstchar; c < f->numchars; c++) {
		drawchar(ri, 0, c, 0);
		patch(dc, FONTBASE + fsize * c, (void *)dc->dc_bits, fsize);
	}
}

void
hash_insert(dc, handle, addr, obj)
	struct rsxfb_devconfig *dc;
	u_int handle, addr;
	void *obj;
{
	u_int32_t htent[2];
	u_int hash, engine = 1, channel = 0;

	htent[0] = htole32(handle);
	htent[1] = htole32(engine << 20 | addr >> 4);

	hash = (handle ^ handle >> 11 ^ handle >> 22 ^ channel << 7) & 0x7ff;

	patch(dc, RAMHT + hash * 8, htent, 8);
	patch(dc, RAMIN + addr, obj, 32);
}

void
patch(dc, addr, data, len)
	struct rsxfb_devconfig *dc;
	u_int addr, len;
	void *data;
{
	u_int32_t *dmabase = (void *)dc->dc_fifo;
	int dmacurrent = 0;
	u_int dest, off, n, *p = data;

	if (len == 0 || len > 4096)
		return;

	/* Need 64-byte alignment. */
	dest = addr & ~63;
	off  = addr & 63;

	BEGIN_RING(3, 0, 1);
	OUT_RING(SURFACE_2D);		/* NV40_CONTEXT_SURFACES_2D */
	BEGIN_RING(3, 0x300, 4);
	OUT_RING(1);			/* FORMAT = Y8 */
	OUT_RING(8192 << 16 | 8192);	/* PITCH */
	OUT_RING(0);			/* SRC */
	OUT_RING(dest);			/* DEST offset */

	BEGIN_RING(5, 0, 1);
	OUT_RING(IMAGE_FROM_CPU);	/* NV40_IMAGE_FROM_CPU */
	BEGIN_RING(5, 0x19c, 1);
	OUT_RING(SURFACE_2D);		/* SURFACE */

	BEGIN_RING(5, 0x2fc, 5);
	OUT_RING(3);			/* OPERATION = SRCCOPY */
	OUT_RING(1);			/* FORMAT = Y8 */
	OUT_RING(0 << 16 | off);	/* OUT_POINT */
	OUT_RING(1 << 16 | len);	/* OUT_SIZE */
	OUT_RING(1 << 16 | len);	/* IN_SIZE */

	n = (len + 3) / 4;
	BEGIN_RING(5, 0x400, n + 1);
	while (n-- > 0)
		OUT_RING(*p++);
	OUT_RING(0);

	/* Kick DMA. */
	HV_gpu_context_attribute(dc->dc_handle, GPU_FIFO_SETUP,
	    dc->dc_gpufifo, dc->dc_gpufifo, dc->dc_gpufifo, 0);
	dc->dc_dmaput = dc->dc_gpufifo + dmacurrent * 4;
	HV_gpu_context_attribute(dc->dc_handle, GPU_FIFO_SETUP,
	    dc->dc_dmaput, dc->dc_gpufifo, dc->dc_gpufifo, 0);
	delay(10);
}

void
docopy(dc, addr, outp, size, format, rop)
	struct rsxfb_devconfig *dc;
	u_int addr, outp, size, format, rop;
{
	struct rasops_info *ri = &dc->dc_ri;
	struct fiforeg *reg;
	u_int stride = ri->ri_stride;
	u_int clip = ri->ri_height << 16 | ri->ri_width;
	u_int32_t *dmabase = (void *)dc->dc_fifo;
	int dmacurrent;
	int s;

	s = splhigh();
	dmacurrent = (dc->dc_dmaput - dc->dc_gpufifo) / 4;
	if (dmacurrent > FIFO_SIZE / 4 - 100) {
		OUT_RING(0x20000000 | dc->dc_gpufifo);
		dmacurrent = 0;
	}

	BEGIN_RING(1, 0, 1);
	OUT_RING(CONTEXT_ROP);
	BEGIN_RING(1, 0x300, 1);
	OUT_RING(rop);

	BEGIN_RING(1, 0, 1);
	OUT_RING(IMAGE_PATTERN);
	BEGIN_RING(1, 0x310, 4);
	OUT_RING(0xffffffff);			/* COLOR0 */
	OUT_RING(0xffffffff);			/* COLOR1 */
	OUT_RING(0xffffffff);			/* PATTERN0 */
	OUT_RING(0xffffffff);			/* PATTERN1 */

	BEGIN_RING(3, 0, 1);
	OUT_RING(SURFACE_2D);
	BEGIN_RING(3, 0x300, 4);
	OUT_RING(10);				/* FORMAT = A8R8G8B8 */
	OUT_RING(stride << 16 | stride);	/* PITCH */
	OUT_RING(0);				/* OFFSET_SRC */
	OUT_RING(ri->ri_offset);		/* OFFSET_DST */

	BEGIN_RING(6, 0, 1);
	OUT_RING(SCALED_IMAGE);
	BEGIN_RING(6, 0x184, 1);
	OUT_RING(DMA_VRAM);
	BEGIN_RING(6, 0x198, 1);
	OUT_RING(SURFACE_2D);			/* SURFACE */

	BEGIN_RING(6, 0x2fc, 9);
	OUT_RING(1);				/* COLOR_CONVERSION */
	OUT_RING(4);				/* FORMAT = X8R8G8B8 */
	OUT_RING(1);				/* OPERATION */
	OUT_RING(0);				/* CLIP_POINT */
	OUT_RING(clip);				/* CLIP_SIZE */
	OUT_RING(outp);				/* OUT_POINT */
	OUT_RING(size);				/* OUT_SIZE */
	OUT_RING(0x00100000);			/* DU_DX */
	OUT_RING(0x00100000);			/* DV_DY */

	BEGIN_RING(6, 0x400, 4);
	OUT_RING(size);				/* SIZE */
	OUT_RING(0 << 24 | 2 << 16 | format);	/* FORMAT */
	OUT_RING(addr);				/* ADDRESS */
	OUT_RING(0);				/* POINT */

	BEGIN_RING(6, 0x100, 1);
	OUT_RING(0);				/* NOP */

	/* Kick DMA. */
	reg = dc->dc_reg;
	if (reg) {
		int i;

		asm volatile ("sync" ::: "memory");
		dc->dc_dmaput = dc->dc_gpufifo + dmacurrent * 4;
		reg->put = dc->dc_dmaput;
		for (i = 0; i < 1000; i++) {
			asm volatile ("eieio");
			if (reg->get == reg->put)
				break;
		}
		/* XXX check DMA Notifier */
	} else {
		u_int get = dc->dc_dmaput;

		dc->dc_dmaput = dc->dc_gpufifo + dmacurrent * 4;
		HV_gpu_context_attribute(dc->dc_handle, GPU_FIFO_SETUP,
		    dc->dc_dmaput, get, get, 0);
		delay(1);
	}
	splx(s);
}

void
fullclear(ri)
	struct rasops_info *ri;
{
	struct rsxfb_devconfig *dc = (void *)ri;
	int offset = ri->ri_offset;
	u_int size = ri->ri_height << 16 | ri->ri_width;

	/* XXX Reset screen offset temporarily. */
	ri->ri_offset = 0;
	docopy(dc, 0, 0, size, ri->ri_stride, ROP_SET);
	ri->ri_offset = offset;
}

void
rsxfb_cursor(cookie, on, row, col)
	void *cookie;
	int on, row, col;
{
	struct rasops_info *ri = cookie;
	struct rsxfb_devconfig *dc = (void *)ri;
	u_int w, h, addr, outp, size;

	h = ri->ri_font->fontheight;
	w = ri->ri_font->fontwidth;

	addr = FONTBASE;
	outp = (h * row) << 16 | (w * col);
	size = h << 16 | w;
	docopy(dc, addr, outp, size, w * 4, ROP_INVERT);
}

void
rsxfb_putchar(cookie, row, col, c, attr)
	void *cookie;
	int row, col;
	u_int c;
	long attr;
{
	struct rasops_info *ri = cookie;
	struct rsxfb_devconfig *dc = (void *)ri;
	u_int w, h, addr, outp, size, rop;

	if (attr & ATTR_REVERSE)
		rop = ROP_COPY;
	else
		rop = ROP_INVCOPY;

	h = ri->ri_font->fontheight;
	w = ri->ri_font->fontwidth;

	addr = FONTBASE + c * ri->ri_xscale * h;
	outp = (h * row) << 16 | (w * col);
	size = h << 16 | w;
	docopy(dc, addr, outp, size, w * 4, rop);
}

void
rsxfb_copycols(cookie, row, src, dst, num)
	void *cookie;
	int row, src, dst, num;
{
	struct rasops_info *ri = cookie;
	struct rsxfb_devconfig *dc = (void *)ri;
	u_int w, h, addr, outp, size;

	h = ri->ri_font->fontheight;
	w = ri->ri_font->fontwidth;

	addr = ri->ri_offset + row * ri->ri_yscale + src * ri->ri_xscale;
	outp = (h * row) << 16 | (w * dst);
	size = h << 16 | (w * num);
	docopy(dc, addr, outp, size, ri->ri_stride, ROP_COPY);
}

void
rsxfb_erasecols(cookie, row, col, num, attr)
	void *cookie;
	int row, col, num;
	long attr;
{
	struct rasops_info *ri = cookie;
	struct rsxfb_devconfig *dc = (void *)ri;
	u_int w, h, outp, size, rop;

	if (attr & ATTR_REVERSE)
		rop = ROP_CLEAR;
	else
		rop = ROP_SET;

	h = ri->ri_font->fontheight;
	w = ri->ri_font->fontwidth;

	outp = (h * row) << 16 | (w * col);
	size = h << 16 | (w * num);
	docopy(dc, 0, outp, size, ri->ri_stride, rop);
}

void
rsxfb_copyrows(cookie, src, dst, num)
	void *cookie;
	int src, dst, num;
{
	struct rasops_info *ri = cookie;
	struct rsxfb_devconfig *dc = (void *)ri;
	u_int w, h, addr, outp, size;

	h = ri->ri_font->fontheight;
	w = ri->ri_font->fontwidth * ri->ri_cols;

	addr = ri->ri_offset + src * ri->ri_yscale;
	outp = (h * dst) << 16 | 0;
	size = (h * num) << 16 | w;
	docopy(dc, addr, outp, size, ri->ri_stride, ROP_COPY);
}

void
rsxfb_eraserows(cookie, row, num, attr)
	void *cookie;
	int row, num;
	long attr;
{
	struct rasops_info *ri = cookie;
	struct rsxfb_devconfig *dc = (void *)ri;
	u_int w, h, outp, size, rop;

	if (attr & ATTR_REVERSE)
		rop = ROP_CLEAR;
	else
		rop = ROP_SET;

	h = ri->ri_font->fontheight;
	w = ri->ri_font->fontwidth * ri->ri_cols;

	outp = (h * row) << 16 | 0;
	size = (h * num) << 16 | w;
	docopy(dc, 0, outp, size, ri->ri_stride, rop);
}
#endif
