/*	$Id: clock.c,v 1.12 2010/12/30 22:34:17 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_timebase.h"

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/timetc.h>

#include <dev/clock_subr.h>
#include <uvm/uvm_extern.h>

#include <machine/autoconf.h>
#include <machine/cpu.h>
#include <machine/cpufunc.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

/*
 * Initially we assume a processor with a timebase frequency of 80 MHz.
 */
static u_int ticks_per_sec = 80*1000*1000;
static u_int ns_per_tick = 13;
int ticks_per_intr;

#ifdef MULTIPROCESSOR
#define lasttb (curcpu()->ci_lasttb)
#define tickspending (curcpu()->ci_tickspending)
#else
static volatile u_int lasttb;
volatile int tickspending;
#endif

#ifdef TIMEBASE_FREQ
u_int timebase_freq = TIMEBASE_FREQ;
#else
u_int timebase_freq = 0;
#endif

#ifdef TIMEBASE_OFFSET
int timebase_offset = TIMEBASE_OFFSET;
#else
int timebase_offset = 0;
#endif

void decr_intr __P((struct clockframe *));
int ps3_gettime(struct todr_chip_handle *, struct timeval *);
int ps3_settime(struct todr_chip_handle *, struct timeval *);

#define DIFF_1970_2000 946684800

int
ps3_gettime(ch, t)
	struct todr_chip_handle *ch;
	struct timeval *t;
{
	u_int64_t rtc_time, tb;
	int rv;

	rv = HV_read_rtc(&rtc_time, &tb);
	if (rv)
		return rv;

	/* rtc_time += flash_rtc_offset; */
	t->tv_sec = rtc_time + DIFF_1970_2000;
	return 0;
}

int
ps3_settime(ch, t)
	struct todr_chip_handle *ch;
	struct timeval *t;
{
	/* XXX modify flash_rtc_offset */
	return 0;
}

void
decr_intr(frame)
	struct clockframe *frame;
{
	u_int tb;
	int tick;
	int nticks;
	int pri;

	/*
	 * Check whether we are initialized.
	 */
	if (!ticks_per_intr)
		return;

	/*
	 * Based on the actual time delay since the last decrementer reload,
	 * we arrange for earlier interrupt next time.
	 */
	tick = mfdec();
	for (nticks = 0; tick < 0; nticks++)
		tick += ticks_per_intr;
	mtdec(tick);

	curcpu()->ci_data.cpu_nintr++;
	intrcnt[CNT_CLOCK]++;

	pri = splclock();
	if (pri >= IPL_CLOCK)
		tickspending += nticks;
	else {
		nticks += tickspending;
		tickspending = 0;

		/*
		 * lasttb is used during microtime. Set it to the virtual
		 * start of this tick interval.
		 */
		asm ("mftb %0" : "=r"(tb));
		lasttb = tb + tick - ticks_per_intr;

		/*
		 * Reenable interrupts
		 */
		mtmsr(mfmsr() | PSL_EE);

		/*
		 * Do standard timer interrupt stuff.
		 */
		do {
			hardclock(frame);
		} while (--nticks > 0);
	}
	splx(pri);
}

#ifdef __HAVE_TIMECOUNTER
u_int get_timecount_mftb(struct timecounter *);

static struct timecounter tc_mftb = {
	get_timecount_mftb,	/* get_timecount */
	NULL,			/* poll_pps */
	~0,			/* counter_mask */
	0,			/* frequency */
	"timebase",		/* name */
	10			/* quality */
};
#endif

void
cpu_initclocks()
{
	static struct todr_chip_handle ps3_rtc;

	ticks_per_intr = ticks_per_sec / hz;
	mtdec(ticks_per_intr);

	ps3_rtc.todr_gettime = ps3_gettime;
	ps3_rtc.todr_settime = ps3_settime;
	todr_attach(&ps3_rtc);

#ifdef __HAVE_TIMECOUNTER
	tc_mftb.tc_frequency = ticks_per_sec;
	tc_init(&tc_mftb);
#endif
}

#ifdef __HAVE_TIMECOUNTER
u_int
get_timecount_mftb(tc)
	struct timecounter *tc;
{
	u_int tb;

	asm volatile ("mftb %0" : "=r"(tb));
	return tb;
}
#endif

void
calc_delay()
{
	u_long f, cpufreq, msr;

	if (timebase_freq != 0) {
		ticks_per_sec = timebase_freq;
		goto useit;
	}

	if (ps3_read_timebase_freq(&f))
		panic("read_timebase_freq");

	cpufreq = f * 40;
	cpufreq = (cpufreq + 50000000) / 100000000;
	printf("cpu %ld.%ld GHz, ", cpufreq / 10, cpufreq % 10);
	printf("timebase %ld MHz\n", (f + 500000) / 1000000);

	ticks_per_sec = f;

	/*
	 * Should check for correct CPU here?		XXX
	 */

	ticks_per_sec += timebase_offset;

useit:
	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	ns_per_tick = 1000000000 / ticks_per_sec;
	asm volatile ("mftb %0" : "=r"(lasttb));
	asm volatile ("mtdec %0" :: "r"(0x7fffffff));
	asm volatile ("mtmsr %0" :: "r"(msr));
}

#ifndef __HAVE_TIMECOUNTER
/*
 * Fill in *tvp with current time with microsecond resolution.
 */
void
microtime(tvp)
	struct timeval *tvp;
{
	u_int tb;
	u_int ticks;
	int msr;

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	asm ("mftb %0" : "=r"(tb));
	ticks = (tb - lasttb) * ns_per_tick;
	*tvp = time;
	mtmsr(msr);
	ticks /= 1000;
	tvp->tv_usec += ticks;
	while (tvp->tv_usec >= 1000000) {
		tvp->tv_usec -= 1000000;
		tvp->tv_sec++;
	}
}
#endif

/*
 * Wait for about n microseconds (at least!).
 */
void
delay(us)
	unsigned int us;
{
	u_int64_t n = us;
	u_int64_t tb;

	tb = mftb();
	tb += (n * 1000 + ns_per_tick - 1) / ns_per_tick;

	while ((int64_t)(mftb() - tb) < 0);
	asm volatile ("sync; isync");
}

/*
 * Nothing to do.
 */
void
setstatclockrate(arg)
	int arg;
{
	/* Do nothing */
}
