/*	$Id: sysmgr.c,v 1.14 2007/04/10 13:43:51 tsubai Exp $	*/

/*-
 * Copyright (c) 2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/reboot.h>
#include <machine/autoconf.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

#include <uvm/uvm_extern.h>

#define sysmgr_port 2

#define SYSMGR_REQUEST		1
#define  SYSMGR_REQUEST_SHUTDOWN	1
#define  SYSMGR_REQUEST_ID_MYSELF	0

#define SYSMGR_RESPONSE		2
#define  SYSMGR_RESPONSE_ACK		0
#define  SYSMGR_RESPONSE_NAK		255

#define SYSMGR_COMMAND		3
#define  SYSMGR_COMMAND_SHUTDOWN	1

#define SYSMGR_EXT_EVENT	4
#define  SYSMGR_EVENT_S			1
#define  SYSMGR_EVENT_L			2
#define  SYSMGR_EVENT_POWER_PRESSED	3
#define  SYSMGR_EVENT_RESET_PRESSED	5

#define SYSMGR_NEXTOP		5
#define  SYSMGR_NEXT_OPERATION_SHUTDOWN	1
#define  SYSMGR_NEXT_OPERATION_REBOOT	2
#define  SYSMGR_NEXT_OPERATION_FASTBOOT	0x82
#define  SYSMGR_WAKE_SOURCE_DEFAULT	0

#define SYSMGR_INTR_ENABLE	8
#define  SYSMGR_INTR_POWER		1
#define  SYSMGR_INTR_RESET		2
#define  SYSMGR_INTR_THERMAL		4
#define  SYSMGR_INTR_SIRCS		8

struct sysmgr_header {
	u_char version;
	u_char hlen;		/* header length (16) */
	u_char res1[2];
	u_int32_t plen;		/* payload length */
	u_int16_t service;
	u_char res2[2];
	u_int32_t tag;
};

struct sysmgr_command {
	u_char version;
	u_char type;
	u_char res[14];
};

struct sysmgr_external_event {
	u_char version;
	u_char type;
	u_char res1[2];
	u_int32_t param;
	u_char res2[8];
};

void sysmgr_init(void);

int sysmgr_recv(void *, u_int);
int sysmgr_intr(void);
void sysmgr_poll(void);

int sysmgr_request(u_int);
int sysmgr_response(u_int);
int sysmgr_next_operation(int);
int sysmgr_intr_enable(u_int);

static volatile u_char lastcmd;

void
sysmgr_init()
{
	sysmgr_intr_enable(0);
	vuart_set_recvsize(sysmgr_port, sizeof(struct sysmgr_header));
	vuart_enable_irq(sysmgr_port, VUART_INTR_RECV);
	vuart_intr_establish(sysmgr_port, sysmgr_intr);
	sysmgr_intr_enable(SYSMGR_INTR_POWER);
}

void
sysmgr_reboot()
{
	lastcmd = 0;
	sysmgr_next_operation(SYSMGR_NEXT_OPERATION_REBOOT);
	sysmgr_request(SYSMGR_REQUEST_SHUTDOWN);
	sysmgr_poll();

	for (;;);
}

void
sysmgr_shutdown()
{
	lastcmd = 0;
	sysmgr_next_operation(SYSMGR_NEXT_OPERATION_SHUTDOWN);
	sysmgr_request(SYSMGR_REQUEST_SHUTDOWN);
	sysmgr_poll();

	for (;;);
}

int
sysmgr_recv(buf, size)
	void *buf;
	u_int size;
{
	u_int64_t actlen;
	int err;

	for (;;) {
		HV_get_vuart_param(sysmgr_port, VUART_PARAM_RBYTES, &actlen);
		if (actlen >= size)
			break;
	}

	err = HV_read_vuart(sysmgr_port, vtophys((vaddr_t)buf), size, &actlen);
	if (err) {
		printf("sysmgr_recv: error %d, size %d, actlen %ld\n",
		    err, size, actlen);
		return 1;
	}
	if (actlen != size) {
		printf("sysmgr_recv: short data, size %d, actlen %ld\n",
		    size, actlen);
		return 1;
	}
	return 0;
}

int
sysmgr_intr()
{
	struct sysmgr_header header;
	struct sysmgr_command *cmd;
	struct sysmgr_external_event *event;
	char body[256];

	if (sysmgr_recv(&header, sizeof header))
		return 0;

	if (header.plen > sizeof body)
		panic("sysmgr: received broken packet");

	if (sysmgr_recv(body, header.plen))
		return 1;
	if (header.version != 1) {
		printf("sysmgr: ignored unknown version packet (%d)\n",
		    header.version);
		return 1;
	}

	switch (header.service) {

	case SYSMGR_COMMAND:
		cmd = (void *)body;
		/* printf("sysmgr: cmd %d\n", cmd->type); */
		if (cmd->type != SYSMGR_COMMAND_SHUTDOWN)
			panic("sysmgr: unknown command");
		lastcmd = cmd->type;
		return 1;

	case SYSMGR_EXT_EVENT:
		event = (void *)body;
		/* printf("sysmgr: event %d\n", event->type); */

		switch (event->type) {
		case SYSMGR_EVENT_POWER_PRESSED:
			sysmgr_shutdown();
		}
		return 1;

	default:
		printf("sysmgr: ignored command %d\n", header.service);
		return 1;
	}
}

void
sysmgr_poll()
{
	while (lastcmd != SYSMGR_COMMAND_SHUTDOWN)
		sysmgr_intr();
	sysmgr_response(SYSMGR_RESPONSE_ACK);
}

int
sysmgr_request(req)
	u_int req;
{
	static struct {
		struct sysmgr_header header;
		struct {
			u_char version;
			u_char request;
			u_char id;
			u_char res[13];
		} body;
	} p;
	u_int64_t actlen;

	bzero(&p, sizeof p);
	p.header.version = 1;
	p.header.hlen = sizeof p.header;
	p.header.plen = sizeof p.body;
	p.header.service = SYSMGR_REQUEST;
	p.body.version = 1;
	p.body.request = req;
	p.body.id = SYSMGR_REQUEST_ID_MYSELF;

	HV_write_vuart(sysmgr_port, vtophys((vaddr_t)&p), sizeof p, &actlen);
	return 0;
}

int
sysmgr_response(status)
	u_int status;
{
	static struct {
		struct sysmgr_header header;
		struct {
			u_char version;
			u_char res1[3];
			u_char status;
			u_char res2[11];
		} body;
	} p;
	u_int64_t actlen;

	bzero(&p, sizeof p);
	p.header.version = 1;
	p.header.hlen = sizeof p.header;
	p.header.plen = sizeof p.body;
	p.header.service = SYSMGR_RESPONSE;
	p.body.version = 1;
	p.body.status = status;

	HV_write_vuart(sysmgr_port, vtophys((vaddr_t)&p), sizeof p, &actlen);
	return 0;
}

int
sysmgr_next_operation(type)
	int type;
{
	static struct {
		struct sysmgr_header header;
		struct {
			u_char version;
			u_char type;
			u_char param;
			u_char res1;
			u_int32_t wakesrc;
			u_char res2[8];
		} body;
	} p;
	u_int64_t actlen;

	bzero(&p, sizeof p);
	p.header.version = 1;
	p.header.hlen = sizeof p.header;
	p.header.plen = sizeof p.body;
	p.header.service = SYSMGR_NEXTOP;
	p.body.version = 2;
	p.body.type = type;
	p.body.wakesrc = SYSMGR_WAKE_SOURCE_DEFAULT;

	HV_write_vuart(sysmgr_port, vtophys((vaddr_t)&p), sizeof p, &actlen);
	return 0;
}

int
sysmgr_intr_enable(intr)
	u_int intr;
{
	struct {
		struct sysmgr_header header;
		struct {
			u_char version;
			u_char res[3];
			u_int32_t intr;
		} body;
	} p;
	u_int64_t actlen;

	bzero(&p, sizeof p);
	p.header.version = 1;
	p.header.hlen = sizeof p.header;
	p.header.plen = sizeof p.body;
	p.header.service = SYSMGR_INTR_ENABLE;
	p.body.version = 1;
	p.body.intr = intr;

	HV_write_vuart(sysmgr_port, vtophys((vaddr_t)&p), sizeof p, &actlen);
	return 0;
}
