/*	$NetBSD: uhub.c,v 1.85 2007/02/26 13:33:09 drochner Exp $	*/
/*	$FreeBSD: src/sys/dev/usb/uhub.c,v 1.18 1999/11/17 22:33:43 n_hibma Exp $	*/

/*
 * Copyright (c) 1998, 2004 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Lennart Augustsson (lennart@augustsson.net) at
 * Carlstedt Research & Technology.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * USB spec: http://www.usb.org/developers/docs/usbspec.zip
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/proc.h>

#include <arch/ps3/ucons/usb.h>
#include <arch/ps3/ucons/usbdi.h>

MALLOC_DEFINE(M_USBDEV, "USB device", "USB device driver");

#undef USBDEVNAME
#define USBDEVNAME(x) "uhub0"

struct uhub_softc {
	usbd_device_handle	sc_hub;		/* USB device */
	usbd_pipe_handle	sc_ipipe;	/* interrupt pipe */
	u_int8_t		sc_status[4];
};
#define UHUB_PROTO(sc) ((sc)->sc_hub->ddesc.bDeviceProtocol)
#define UHUB_IS_HIGH_SPEED(sc) (UHUB_PROTO(sc) != UDPROTO_FSHUB)
#define UHUB_IS_SINGLE_TT(sc) (UHUB_PROTO(sc) == UDPROTO_HSHUBSTT)

Static int uhub_explore(usbd_device_handle hub);

static int usbd_reset_port(usbd_device_handle, int, usb_port_status_t *);
static int usbd_get_port_status(usbd_device_handle, int, usb_port_status_t *);
static int usbd_set_port_feature(usbd_device_handle dev, int, int);
static int usbd_clear_port_feature(usbd_device_handle, int, int);
static int usbd_set_interface(usbd_interface_handle, int);

/*
 * We need two attachment points:
 * hub to usb and hub to hub
 * Every other driver only connects to hubs
 */

void uhub_attach(usbd_device_handle);

void
uhub_attach(dev)
	usbd_device_handle dev;
{
	struct uhub_softc *sc;
	int err;
	struct usbd_hub *hub = NULL;
	usb_device_request_t req;
	usb_hub_descriptor_t hubdesc;
	int p, port, nports, pwrdly;
	usbd_interface_handle iface;
	usb_endpoint_descriptor_t *ed;
	size_t statuslen;
	char xname[8];
	static int n;

	sc = malloc(sizeof *sc, M_USB, M_NOWAIT|M_ZERO);
	sprintf(xname, "uhub%d", n++);

	sc->sc_hub = dev;

	err = usbd_set_config_index(dev, 0, 1);
	if (err)
		return;

	/* Get hub descriptor. */
	req.bmRequestType = UT_READ_CLASS_DEVICE;
	req.bRequest = UR_GET_DESCRIPTOR;
	USETW2(req.wValue, UDESC_HUB, 0);
	USETW(req.wIndex, 0);
	USETW(req.wLength, USB_HUB_DESCRIPTOR_SIZE);
	err = usbd_do_request(dev, &req, &hubdesc);
	nports = hubdesc.bNbrPorts;
	if (!err && nports > 7) {
		USETW(req.wLength, USB_HUB_DESCRIPTOR_SIZE + (nports+1) / 8);
		err = usbd_do_request(dev, &req, &hubdesc);
	}
	if (err)
		return;

	if (nports == 0) {
		printf("%s: no ports, hub ignored\n", xname);
		goto bad;
	}
	printf("%s: %d ports\n", xname, nports);

#if 0
printf("sizeof *hub = %ld\n", sizeof(*hub));	// 72
printf("sizeof usbd_port = %ld\n", sizeof(struct usbd_port));	// 16
#endif

	hub = malloc(sizeof(*hub) + (nports-1) * sizeof(struct usbd_port),
		     M_USBDEV, M_NOWAIT);
	if (hub == NULL)
		return;
	dev->hub = hub;
	hub->explore = uhub_explore;
	hub->hubdesc = hubdesc;

	/* Set up interrupt pipe. */
	iface = &dev->ifaces[0];

	if (UHUB_IS_HIGH_SPEED(sc) && !UHUB_IS_SINGLE_TT(sc)) {
		err = usbd_set_interface(iface, 1);
		if (err)
			printf("%s: can't enable multiple TTs\n",
			       USBDEVNAME(sc->sc_dev));
	}

	ed = iface->endpoints[0].edesc;
	if (ed == NULL) {
		printf("%s: no endpoint descriptor\n", xname);
		goto bad;
	}
	if ((ed->bmAttributes & UE_XFERTYPE) != UE_INTERRUPT) {
		printf("%s: bad interrupt endpoint\n", xname);
		goto bad;
	}

	statuslen = (nports + 1 + 7) / 8;

	err = usbd_open_pipe_intr(iface, ed->bEndpointAddress,
	    &sc->sc_ipipe, sc->sc_status, statuslen);
	if (err) {
		printf("%s: cannot open interrupt pipe\n", xname);
		goto bad;
	}

	/* Wait with power off for a while. */
	usb_delay_ms(USB_POWER_DOWN_TIME);

	/*
	 * To have the best chance of success we do things in the exact same
	 * order as Windoze98.  This should not be necessary, but some
	 * devices do not follow the USB specs to the letter.
	 *
	 * These are the events on the bus when a hub is attached:
	 *  Get device and config descriptors (see attach code)
	 *  Get hub descriptor (see above)
	 *  For all ports
	 *     turn on power
	 *     wait for power to become stable
	 * (all below happens in explore code)
	 *  For all ports
	 *     clear C_PORT_CONNECTION
	 *  For all ports
	 *     get port status
	 *     if device connected
	 *        wait 100 ms
	 *        turn on reset
	 *        wait
	 *        clear C_PORT_RESET
	 *        get port status
	 *        proceed with device attachment
	 */

#if 0
	if (UHUB_IS_HIGH_SPEED(sc) && nports > 0) {
		tts = malloc((UHUB_IS_SINGLE_TT(sc) ? 1 : nports) *
			     sizeof (struct usbd_tt), M_USBDEV, M_NOWAIT);
		if (!tts)
			goto bad;
	}
#endif
	/* Set up data structures */
	for (p = 0; p < nports; p++) {
		struct usbd_port *up = &hub->ports[p];
		up->device = NULL;
		up->portno = p+1;
		up->restartcnt = 0;
#if 0
		if (UHUB_IS_HIGH_SPEED(sc)) {
			up->tt = &tts[UHUB_IS_SINGLE_TT(sc) ? 0 : p];
			up->tt->hub = hub;
		} else {
			up->tt = NULL;
		}
#endif
	}

	/* XXX should check for none, individual, or ganged power? */

	pwrdly = dev->hub->hubdesc.bPwrOn2PwrGood * UHD_PWRON_FACTOR
	    + USB_EXTRA_POWER_UP_TIME;
	for (port = 1; port <= nports; port++) {
		/* Turn the power on. */
		usbd_set_port_feature(dev, port, UHF_PORT_POWER);

		/* Wait for stable power. */
		usb_delay_ms(pwrdly);
	}

	/* The usual exploration will finish the setup. */

	return;

 bad:
	if (hub)
		free(hub, M_USBDEV);
	dev->hub = NULL;
	return;
}

int
uhub_explore(dev)
	usbd_device_handle dev;
{
	usb_hub_descriptor_t *hd = &dev->hub->hubdesc;
	struct usbd_port *up;
	int err, speed, port;
	int change, status, reconnect;

	for (port = 1; port <= hd->bNbrPorts; port++) {
		up = &dev->hub->ports[port-1];
		err = usbd_get_port_status(dev, port, &up->status);
		if (err)
			continue;

		status = UGETW(up->status.wPortStatus);
		change = UGETW(up->status.wPortChange);
		reconnect = 0;

		if (change & UPS_C_PORT_ENABLED) {

			usbd_clear_port_feature(dev, port, UHF_C_PORT_ENABLE);
			if (change & UPS_C_CONNECT_STATUS) {
				/* Ignore the port error if the device
				   vanished. */
			} else if (status & UPS_PORT_ENABLED) {
				printf("%s: illegal enable change, port %d\n",
				       USBDEVNAME(sc->sc_dev), port);
			} else {
				/* Port error condition. */
				if (up->restartcnt++ < USBD_RESTART_MAX)
					goto disco;
				else
					printf("%s: port error, giving up "
					       "port %d\n",
					       USBDEVNAME(sc->sc_dev), port);
			}
		}
		if (!reconnect && !(change & UPS_C_CONNECT_STATUS)) {
			/* No status change, just do recursive explore. */
			if (up->device && up->device->hub)
				uhub_explore(up->device);
			continue;
		}

		/* We have a connect status change, handle it. */
		usbd_clear_port_feature(dev, port, UHF_C_PORT_CONNECTION);
		/*usbd_clear_port_feature(dev, port, UHF_C_PORT_ENABLE);*/

		/*
		 * If there is already a device on the port the change status
		 * must mean that is has disconnected.  Looking at the
		 * current connect status is not enough to figure this out
		 * since a new unit may have been connected before we handle
		 * the disconnect.
		 */
	disco:
		if (up->device != NULL) {
			/* Disconnected */
			panic("disconnected");
		}
		if (!(status & UPS_CURRENT_CONNECT_STATUS)) {
			/* Nothing connected, just ignore it. */
			continue;
		}

		/* Connected */

		if (!(status & UPS_PORT_POWER))
			printf("%s: strange, connected port %d has no power\n",
			       USBDEVNAME(sc->sc_dev), port);

		/* Wait for maximum device power up time. */
		usb_delay_ms(USB_PORT_POWERUP_DELAY);

		/* Reset port, which implies enabling it. */
		if (usbd_reset_port(dev, port, &up->status)) {
			printf("%s: port %d reset failed\n",
			       USBDEVNAME(sc->sc_dev), port);
			continue;
		}
		/* Get port status again, it might have changed during reset */
		err = usbd_get_port_status(dev, port, &up->status);
		if (err)
			continue;

		status = UGETW(up->status.wPortStatus);
		change = UGETW(up->status.wPortChange);
		if (!(status & UPS_CURRENT_CONNECT_STATUS)) {
			/* Nothing connected, just ignore it. */
			continue;
		}

		/* Figure out device speed */
		if (status & UPS_HIGH_SPEED)
			speed = USB_SPEED_HIGH;
		else if (status & UPS_LOW_SPEED)
			speed = USB_SPEED_LOW;
		else
			speed = USB_SPEED_FULL;

		/* Get device info and set its address. */
		up->device = malloc(sizeof *dev, M_USB, M_NOWAIT|M_ZERO);
		if (up->device == NULL)
			return USBD_NOMEM;
		err = usbd_new_device(dev->bus, speed, port, up->device);
		/* XXX retry a few times? */
		if (err) {
			/* Avoid addressing problems by disabling. */
			/* usbd_reset_port(dev, port, &up->status); */

			/*
			 * The unit refused to accept a new address, or had
			 * some other serious problem.  Since we cannot leave
			 * at 0 we have to disable the port instead.
			 */
			printf("%s: device problem, disabling port %d\n",
			       USBDEVNAME(sc->sc_dev), port);
			usbd_clear_port_feature(dev, port, UHF_PORT_ENABLE);
			continue;
		}

		/* The port set up succeeded, reset error count. */
		up->restartcnt = 0;

		usb_device_descriptor_t *dd = &up->device->ddesc;

		if (dd->bDeviceClass == UDCLASS_HUB) {
			printf("attaching uhub\n");
			uhub_attach(up->device);
			if (up->device->hub)
				uhub_explore(up->device);
		}

		int uplcom_match(void *);
		void uplcom_attach(void *);

		struct usb_attach_arg uaa;

		uaa.device = up->device;
		uaa.vendor = UGETW(dd->idVendor);
		uaa.product = UGETW(dd->idProduct);

		if (uplcom_match(&uaa)) {
			printf("attaching uplcom\n");
			uplcom_attach(&uaa);
		}
	}
	return 0;
}

////////

/* XXXX use altno */
int
usbd_set_interface(usbd_interface_handle iface, int altidx)
{
	usb_device_request_t req;
	void *endpoints;
	int err;

	endpoints = iface->endpoints;
	err = usbd_fill_iface_data(iface->device, iface->index, altidx);
	if (err)
		return err;

	/* new setting works, we can free old endpoints */
	if (endpoints)
		free(endpoints, M_USB);

	req.bmRequestType = UT_WRITE_INTERFACE;
	req.bRequest = UR_SET_INTERFACE;
	USETW(req.wValue, iface->idesc->bAlternateSetting);
	USETW(req.wIndex, iface->idesc->bInterfaceNumber);
	USETW(req.wLength, 0);
	return usbd_do_request(iface->device, &req, 0);
}

int
usbd_reset_port(usbd_device_handle dev, int port, usb_port_status_t *ps)
{
	usb_device_request_t req;
	int err;
	int n;

	req.bmRequestType = UT_WRITE_CLASS_OTHER;
	req.bRequest = UR_SET_FEATURE;
	USETW(req.wValue, UHF_PORT_RESET);
	USETW(req.wIndex, port);
	USETW(req.wLength, 0);
	err = usbd_do_request(dev, &req, 0);
	if (err)
		return err;
	n = 10;
	do {
		/* Wait for device to recover from reset. */
		usb_delay_ms(USB_PORT_RESET_DELAY);
		err = usbd_get_port_status(dev, port, ps);
		if (err)
			return err;

		/* If the device disappeared, just give up. */
		if (!(UGETW(ps->wPortStatus) & UPS_CURRENT_CONNECT_STATUS))
			return (USBD_NORMAL_COMPLETION);
	} while ((UGETW(ps->wPortChange) & UPS_C_PORT_RESET) == 0 && --n > 0);
	if (n == 0)
		return (USBD_TIMEOUT);
	err = usbd_clear_port_feature(dev, port, UHF_C_PORT_RESET);

	/* Wait for the device to recover from reset. */
	usb_delay_ms(USB_PORT_RESET_RECOVERY);
	return err;
}

int
usbd_get_port_status(usbd_device_handle dev, int port, usb_port_status_t *ps)
{
	usb_device_request_t req;

	req.bmRequestType = UT_READ_CLASS_OTHER;
	req.bRequest = UR_GET_STATUS;
	USETW(req.wValue, 0);
	USETW(req.wIndex, port);
	USETW(req.wLength, sizeof *ps);
	return usbd_do_request(dev, &req, ps);
}

int
usbd_clear_port_feature(usbd_device_handle dev, int port, int sel)
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_CLASS_OTHER;
	req.bRequest = UR_CLEAR_FEATURE;
	USETW(req.wValue, sel);
	USETW(req.wIndex, port);
	USETW(req.wLength, 0);
	return usbd_do_request(dev, &req, 0);
}

int
usbd_set_port_feature(usbd_device_handle dev, int port, int sel)
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_CLASS_OTHER;
	req.bRequest = UR_SET_FEATURE;
	USETW(req.wValue, sel);
	USETW(req.wIndex, port);
	USETW(req.wLength, 0);
	return usbd_do_request(dev, &req, 0);
}
