/*	$NetBSD: uplcom.c,v 1.52 2007/01/31 10:11:50 msaitoh Exp $	*/
/*
 * Copyright (c) 2001 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Ichiro FUKUHARA (ichiro@ichiro.org).
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * General information: http://www.prolific.com.tw/fr_pl2303.htm
 * http://www.hitachi-hitec.com/jyouhou/prolific/2303.pdf
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <arch/ps3/ucons/usb.h>
#include <arch/ps3/ucons/usbdi.h>
#include <dev/usb/usbcdc.h>
#include <dev/usb/usbdevs.h>

#define	UPLCOM_CONFIG_INDEX	0
#define	UPLCOM_IFACE_INDEX	0
#define	UPLCOM_SECOND_IFACE_INDEX	1

#define	UPLCOM_SET_REQUEST	0x01
#define	UPLCOM_SET_CRTSCTS_0	0x41
#define	UPLCOM_SET_CRTSCTS_HX	0x61

enum  pl2303_type {
	UPLCOM_TYPE_0,
	UPLCOM_TYPE_HX,
};

struct	uplcom_softc {
	usbd_device_handle	sc_udev;	/* USB device */
	usbd_interface_handle	sc_iface;	/* interface */
	int			sc_iface_number;	/* interface number */

	usbd_interface_handle	sc_intr_iface;	/* interrupt interface */
	int			sc_intr_number;	/* interrupt number */
	usbd_pipe_handle	sc_intr_pipe;	/* interrupt pipe */
	u_char			sc_intr_buf[32]; /* interrupt buffer */
	int			sc_isize;

	char			sc_dtr;		/* current DTR state */
	char			sc_rts;		/* current RTS state */

	char			sc_type;	/* PL2303 chip type */
};

/*
 * These are the maximum number of bytes transferred per frame.
 * The output buffer size cannot be increased due to the size encoding.
 */
#define UPLCOMIBUFSIZE 256
#define UPLCOMOBUFSIZE 256

static int uplcom_reset(struct uplcom_softc *);
static int uplcom_set_line_coding(struct uplcom_softc *, usb_cdc_line_state_t *);
/* static int uplcom_set_crtscts(struct uplcom_softc *); */

static void uplcom_set_line_state(struct uplcom_softc *);
static int uplcom_open(void *, int);
static int uplcom_vendor_control_write(usbd_device_handle, u_int16_t, u_int16_t);

static int bulkin, bulkout;

static const struct usb_devno uplcom_devs[] = {
	/* I/O DATA USB-RSAQ2 */
	{ USB_VENDOR_PROLIFIC, USB_PRODUCT_PROLIFIC_RSAQ2 },
	/* I/O DATA USB-RSAQ3 */
	{ USB_VENDOR_PROLIFIC, USB_PRODUCT_PROLIFIC_RSAQ3 },
	/* I/O DATA USB-RSAQ */
	{ USB_VENDOR_IODATA, USB_PRODUCT_IODATA_USBRSAQ },
	/* PLANEX USB-RS232 URS-03 */
	{ USB_VENDOR_ATEN, USB_PRODUCT_ATEN_UC232A },
	/* IOGEAR/ATEN UC-232A */
	{ USB_VENDOR_PROLIFIC, USB_PRODUCT_PROLIFIC_PL2303 },
	/* IOGEAR/ATENTRIPPLITE */
	{ USB_VENDOR_TRIPPLITE, USB_PRODUCT_TRIPPLITE_U209 },
	/* ELECOM UC-SGT */
	{ USB_VENDOR_ELECOM, USB_PRODUCT_ELECOM_UCSGT },
	/* ELECOM UC-SGT0 */
	{ USB_VENDOR_ELECOM, USB_PRODUCT_ELECOM_UCSGT0 },
	/* Panasonic 50" Touch Panel */
	{ USB_VENDOR_PANASONIC, USB_PRODUCT_PANASONIC_TYTP50P6S },
	/* RATOC REX-USB60 */
	{ USB_VENDOR_RATOC, USB_PRODUCT_RATOC_REXUSB60 },
	/* TDK USB-PHS Adapter UHA6400 */
	{ USB_VENDOR_TDK, USB_PRODUCT_TDK_UHA6400 },
	/* TDK USB-PDC Adapter UPA9664 */
	{ USB_VENDOR_TDK, USB_PRODUCT_TDK_UPA9664 },
	/* Sony Ericsson USB Cable */
	{ USB_VENDOR_SUSTEEN, USB_PRODUCT_SUSTEEN_DCU10 },
	/* SOURCENEXT KeikaiDenwa 8 */
	{ USB_VENDOR_SOURCENEXT, USB_PRODUCT_SOURCENEXT_KEIKAI8 },
	/* SOURCENEXT KeikaiDenwa 8 with charger */
	{ USB_VENDOR_SOURCENEXT, USB_PRODUCT_SOURCENEXT_KEIKAI8_CHG },
	/* HAL Corporation Crossam2+USB */
	{ USB_VENDOR_HAL, USB_PRODUCT_HAL_IMR001 },
	/* Sitecom USB to serial cable */
	{ USB_VENDOR_SITECOM, USB_PRODUCT_SITECOM_CN104 },
	/* Pharos USB GPS - Microsoft version */
	{ USB_VENDOR_PROLIFIC, USB_PRODUCT_PROLIFIC_PL2303X },
	/* Willcom WS002IN (DD) */
	{ USB_VENDOR_NETINDEX, USB_PRODUCT_NETINDEX_WS002IN },
};
#define uplcom_lookup(v, p) usb_lookup(uplcom_devs, v, p)

static const struct {
	uint16_t		vendor;
	uint16_t		product;
	enum pl2303_type	chiptype;
} uplcom_devs_ext[] = {
	/* I/O DATA USB-RSAQ3 */
	{ USB_VENDOR_PROLIFIC, USB_PRODUCT_PROLIFIC_RSAQ3, UPLCOM_TYPE_HX },
	{ USB_VENDOR_NETINDEX, USB_PRODUCT_NETINDEX_WS002IN, UPLCOM_TYPE_HX },
	/* I/O DATA USB-RSAQ5 */
	{ USB_VENDOR_IODATA, USB_PRODUCT_IODATA_USBRSAQ5, UPLCOM_TYPE_HX },
	{0, 0, 0}
};

int uplcom_match(void *);
void uplcom_attach(struct usb_attach_arg *);

int
uplcom_match(aux)
	void *aux;
{
	struct usb_attach_arg *uaa = aux;

	return uplcom_lookup(uaa->vendor, uaa->product) != NULL ? 1 : 0;
}

static struct uplcom_softc self;

void
uplcom_attach(uaa)
	struct usb_attach_arg *uaa;
{
	struct uplcom_softc *sc = (void *)&self;
	usbd_device_handle dev = uaa->device;
	usb_device_descriptor_t *ddesc;
	usb_config_descriptor_t *cdesc;
	usb_interface_descriptor_t *id;
	usb_endpoint_descriptor_t *ed;
	char *devname = "uplcom0";
	int err;
	int i;

        sc->sc_udev = dev;

	/* initialize endpoints */
	bulkin = bulkout = -1;
	sc->sc_intr_number = -1;
	sc->sc_intr_pipe = NULL;

	/* Move the device into the configured state. */
	err = usbd_set_config_index(dev, UPLCOM_CONFIG_INDEX, 1);
	if (err) {
		printf("%s: failed to set configuration\n",devname);
		return;
	}

	/* get the device descriptor */
	ddesc = &dev->ddesc;

	/* determine chip type */
	for (i = 0; uplcom_devs_ext[i].vendor != 0; i++) {
		if (uplcom_devs_ext[i].vendor == uaa->vendor &&
		    uplcom_devs_ext[i].product == uaa->product) {
			sc->sc_type = uplcom_devs_ext[i].chiptype;
			goto chiptype_determined;
		}
	}
	/*
	 * NOTE: The Linux driver distinguishes between UPLCOM_TYPE_0
	 * and UPLCOM_TYPE_1 type chips by testing other fields in the
	 * device descriptor.  As far as the uplcom driver is
	 * concerned, both types are identical.
	 * The bcdDevice field should also distinguish these versions,
	 * but who knows.
	 */
	if (UGETW(ddesc->bcdDevice) == 0x0300)
		sc->sc_type = UPLCOM_TYPE_HX;
	else
		sc->sc_type = UPLCOM_TYPE_0;
chiptype_determined:

	/* get the config descriptor */
	cdesc = dev->cdesc;
	if (cdesc == NULL) {
		printf("%s: failed to get configuration descriptor\n", devname);
		return;
	}

	/* get the (first/common) interface */
	sc->sc_iface = &dev->ifaces[UPLCOM_IFACE_INDEX];

	/* Find the interrupt endpoints */

	id = sc->sc_iface->idesc;
	sc->sc_iface_number = id->bInterfaceNumber;

	for (i = 0; i < id->bNumEndpoints; i++) {
		ed = sc->sc_iface->endpoints[i].edesc;
		if (ed == NULL) {
			printf("%s: no endpoint descriptor for %d\n",
			    devname, i);
			return;
		}

		if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN &&
		    UE_GET_XFERTYPE(ed->bmAttributes) == UE_INTERRUPT) {
			sc->sc_intr_number = ed->bEndpointAddress;
			sc->sc_isize = UGETW(ed->wMaxPacketSize);
		}
	}

	if (sc->sc_intr_number == -1) {
		printf("%s: could not find interrupt in\n", devname);
		return;
	}

	/* keep interface for interrupt */
	sc->sc_intr_iface = sc->sc_iface;

	/*
	 * USB-RSAQ1 has two interface
	 *
	 *  USB-RSAQ1       | USB-RSAQ2
 	 * -----------------+-----------------
	 * Interface 0      |Interface 0
	 *  Interrupt(0x81) | Interrupt(0x81)
	 * -----------------+ BulkIN(0x02)
	 * Interface 1	    | BulkOUT(0x83)
	 *   BulkIN(0x02)   |
	 *   BulkOUT(0x83)  |
	 */
	if (cdesc->bNumInterface == 2)
		sc->sc_iface = &dev->ifaces[UPLCOM_SECOND_IFACE_INDEX];

	/* Find the bulk{in,out} endpoints */

	id = sc->sc_iface->idesc;
	sc->sc_iface_number = id->bInterfaceNumber;

	for (i = 0; i < id->bNumEndpoints; i++) {
		ed = sc->sc_iface->endpoints[i].edesc;
		if (ed == NULL) {
			printf("%s: no endpoint descriptor for %d\n",
			    devname, i);
			return;
		}

		if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN &&
		    UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
			bulkin = ed->bEndpointAddress;
		} else if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_OUT &&
		    UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK) {
			bulkout = ed->bEndpointAddress;
		}
	}

	err = uplcom_reset(sc);
	if (err) {
		printf("%s: reset failed\n", devname);
		return;
	}
}

int
uplcom_reset(struct uplcom_softc *sc)
{
        usb_device_request_t req;

        req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
        req.bRequest = UPLCOM_SET_REQUEST;
        USETW(req.wValue, 0);
        USETW(req.wIndex, sc->sc_iface_number);
        USETW(req.wLength, 0);

	return usbd_do_request(sc->sc_udev, &req, 0);
}

void
uplcom_set_line_state(sc)
	struct uplcom_softc *sc;
{
	usb_device_request_t req;
	int ls = 0;

	if (sc->sc_dtr) ls |= UCDC_LINE_DTR;
	if (sc->sc_rts) ls |= UCDC_LINE_RTS;

	req.bmRequestType = UT_WRITE_CLASS_INTERFACE;
	req.bRequest = UCDC_SET_CONTROL_LINE_STATE;
	USETW(req.wValue, ls);
	USETW(req.wIndex, sc->sc_iface_number);
	USETW(req.wLength, 0);

	usbd_do_request(sc->sc_udev, &req, 0);
}

#if 0
int
uplcom_set_crtscts(sc)
	struct uplcom_softc *sc;
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	req.bRequest = UPLCOM_SET_REQUEST;
	USETW(req.wValue, 0);
	if (sc->sc_type == UPLCOM_TYPE_HX)
		USETW(req.wIndex, UPLCOM_SET_CRTSCTS_HX);
	else
		USETW(req.wIndex, UPLCOM_SET_CRTSCTS_0);
	USETW(req.wLength, 0);

	return usbd_do_request(sc->sc_udev, &req, 0);
}
#endif

int
uplcom_set_line_coding(sc, state)
	struct uplcom_softc *sc;
	usb_cdc_line_state_t *state;
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_CLASS_INTERFACE;
	req.bRequest = UCDC_SET_LINE_CODING;
	USETW(req.wValue, 0);
	USETW(req.wIndex, sc->sc_iface_number);
	USETW(req.wLength, UCDC_LINE_STATE_LENGTH);

	return usbd_do_request(sc->sc_udev, &req, state);
}

int
uplcom_vendor_control_write(dev, value, index)
	usbd_device_handle dev;
	u_int16_t value, index;
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_VENDOR_DEVICE;
	req.bRequest = UPLCOM_SET_REQUEST;
	USETW(req.wValue, value);
	USETW(req.wIndex, index);
	USETW(req.wLength, 0);

	return usbd_do_request(dev, &req, NULL);
}

int
uplcom_open(void *addr, int portno)
{
	struct uplcom_softc *sc = addr;
	int err;

	/* Some unknown device frobbing. */
	if (sc->sc_type == UPLCOM_TYPE_HX)
		uplcom_vendor_control_write(sc->sc_udev, 2, 0x44);
	else
		uplcom_vendor_control_write(sc->sc_udev, 2, 0x24);

	if (sc->sc_intr_pipe)
		return 0;

	if (sc->sc_isize > sizeof sc->sc_intr_buf)
		sc->sc_isize = sizeof sc->sc_intr_buf;

	err = usbd_open_pipe_intr(sc->sc_intr_iface, sc->sc_intr_number,
		&sc->sc_intr_pipe, sc->sc_intr_buf, sc->sc_isize);

	return err;
}

static usbd_pipe_handle bulkout_pipe;	/* bulk out pipe */
static usbd_pipe_handle bulkin_pipe;	/* bulk in pipe */
static struct usbd_xfer oxfer;		/* write request */
static struct usbd_xfer ixfer;		/* read request */

#include <dev/cons.h>

void uplcom_cnprobe(struct consdev *);
void uplcom_cninit(struct consdev *);
int uplcom_cngetc(dev_t);
void uplcom_cnputc(dev_t, int);
void uplcom_cnpollc(dev_t, int);

struct consdev consdev_uplcom = {
	uplcom_cnprobe,
	uplcom_cninit,
	uplcom_cngetc,
	uplcom_cnputc,
	uplcom_cnpollc
};

void
uplcom_cnprobe(cp)
	struct consdev *cp;
{
}

void
uplcom_cninit(cp)
	struct consdev *cp;
{
	struct uplcom_softc *sc = (void *)&self;
	usb_cdc_line_state_t ls;
	int err;

	err = uplcom_open(sc, 0);
	if (err)
		return;

	USETDW(ls.dwDTERate, 57600);
	ls.bCharFormat = UCDC_STOP_BIT_1;
	ls.bParityType = UCDC_PARITY_NONE;
	ls.bDataBits = 8;
	err = uplcom_set_line_coding(sc, &ls);
	if (err)
		return;

	/* uplcom_set_crtscts(sc); */

	sc->sc_dtr = sc->sc_rts = 1;
	uplcom_set_line_state(sc);

	/* XXX CLOCAL? */

	/* Open the bulk pipes */
	err = usbd_open_pipe(sc->sc_iface, bulkin, 0, &bulkin_pipe);
	if (err)
		return;

	err = usbd_open_pipe(sc->sc_iface, bulkout, USBD_EXCLUSIVE_USE,
	    &bulkout_pipe);
	if (err)
		return;

	bzero(&ixfer, sizeof ixfer);
	bzero(&oxfer, sizeof oxfer);
}

int
uplcom_cngetc(dev)
	dev_t dev;
{
	char ibuf[1];

	ixfer.dmabuf = ibuf;
	ixfer.rqflags |= URQ_DEV_DMABUF;

	ibuf[0] = 0;
	usbd_setup_xfer(&ixfer, bulkin_pipe, ibuf, 1,
		USBD_SHORT_XFER_OK | USBD_NO_COPY);
	usbd_transfer(&ixfer);

	return ibuf[0];
}

void
uplcom_cnputc(dev, c)
	dev_t dev;
	int c;
{
	char obuf[1];

	obuf[0] = c;

	oxfer.dmabuf = obuf;
	oxfer.rqflags |= URQ_DEV_DMABUF;

	usbd_setup_xfer(&oxfer, bulkout_pipe, obuf, 1, USBD_NO_COPY);
	usbd_transfer(&oxfer);
}

void
uplcom_cnpollc(dev, on)
	dev_t dev;
	int on;
{
}
