/*	$NetBSD: usbdi.c,v 1.119 2007/02/26 13:44:40 drochner Exp $	*/

/*
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Lennart Augustsson (lennart@augustsson.net) at
 * Carlstedt Research & Technology.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/device.h>
#include <sys/malloc.h>

#include <arch/ps3/ucons/usb.h>
#include <arch/ps3/ucons/usbdi.h>

static int usbd_close_pipe(usbd_pipe_handle);
static void usbd_free_buffer(usbd_xfer_handle);
static void usbd_free_xfer(usbd_xfer_handle);

static int usbd_set_config(usbd_device_handle, int);
static int usbd_getnewaddr(usbd_bus_handle bus);
static void usbd_free_iface_data(usbd_device_handle dev, int ifcno);

static usb_interface_descriptor_t *usbd_find_idesc(usb_config_descriptor_t *,
    int, int);

static int usbd_get_desc(usbd_device_handle dev, int, int, int, void *);
static int usbd_get_config_desc(usbd_device_handle, int,
    usb_config_descriptor_t *);
static int usbd_get_device_desc(usbd_device_handle, usb_device_descriptor_t *);
static int usbd_set_address(usbd_device_handle dev, int addr);

MALLOC_DEFINE(M_USB, "USB", "USB misc. memory");

static inline int
usbd_xfer_isread(usbd_xfer_handle xfer)
{
	if (xfer->rqflags & URQ_REQUEST)
		return xfer->request.bmRequestType & UT_READ;
	else
		return xfer->pipe->endpoint->edesc->bEndpointAddress &
			UE_DIR_IN;
}

int
usbd_open_pipe(iface, address, flags, pipe)
	usbd_interface_handle iface;
	u_int address, flags;
	usbd_pipe_handle *pipe;
{
	usbd_pipe_handle p;
	struct usbd_endpoint *ep;
	int i;

	for (i = 0; i < iface->idesc->bNumEndpoints; i++) {
		ep = &iface->endpoints[i];
		if (ep->edesc == NULL)
			return USBD_IOERROR;
		if (ep->edesc->bEndpointAddress == address)
			goto found;
	}
	return USBD_BAD_ADDRESS;
 found:
	if ((flags & USBD_EXCLUSIVE_USE) && ep->refcnt != 0)
		return USBD_IN_USE;
	p = usbd_setup_pipe(iface->device, iface, ep);
	if (p == NULL)
		return USBD_NOMEM;
	*pipe = p;
	return 0;
}

int
usbd_open_pipe_intr(iface, address, pipe, buffer, len)
	usbd_interface_handle iface;
	u_int address;
	usbd_pipe_handle *pipe;
	void *buffer;
	u_int32_t len;
{
	usbd_xfer_handle xfer;
	usbd_pipe_handle ipipe;
	int err;

	err = usbd_open_pipe(iface, address, USBD_EXCLUSIVE_USE, &ipipe);
	if (err)
		return err;
	xfer = malloc(sizeof(struct usbd_xfer), M_USB, M_NOWAIT|M_ZERO);
	if (xfer == NULL) {
		err = USBD_NOMEM;
		goto bad1;
	}
	usbd_setup_xfer(xfer, ipipe, buffer, len, USBD_SHORT_XFER_OK);
	ipipe->intrxfer = xfer;
	ipipe->repeat = 1;
	err = usbd_transfer(xfer);
	*pipe = ipipe;
	if (err != USBD_IN_PROGRESS)
		goto bad2;
	return 0;

 bad2:
	ipipe->intrxfer = NULL;
	ipipe->repeat = 0;
	usbd_free_xfer(xfer);
 bad1:
	usbd_close_pipe(ipipe);
	return err;
}

int
usbd_close_pipe(usbd_pipe_handle pipe)
{
	if (--pipe->refcnt != 0)
		return 0;
	pipe->endpoint->refcnt--;
	pipe->methods->close(pipe);
	if (pipe->intrxfer)
		usbd_free_xfer(pipe->intrxfer);
	free(pipe, M_USB);
	return 0;
}

int
usbd_transfer(usbd_xfer_handle xfer)
{
	usbd_pipe_handle pipe = xfer->pipe;
	int err;
	u_int size;

	xfer->done = 0;

	size = xfer->length;
	/* If there is no buffer, allocate one. */
	if (!(xfer->rqflags & URQ_DEV_DMABUF) && size != 0) {
#ifdef DIAGNOSTIC
		if (xfer->rqflags & URQ_AUTO_DMABUF)
			printf("usbd_transfer: has old buffer!\n");
#endif
		xfer->dmabuf = malloc(size, M_USB, M_NOWAIT);
		if (xfer->dmabuf == NULL)
			return USBD_NOMEM;
		xfer->rqflags |= URQ_AUTO_DMABUF;
	}

	/* Copy data if going out. */
	if (!(xfer->flags & USBD_NO_COPY) && size != 0 &&
	    !usbd_xfer_isread(xfer))
		memcpy(xfer->dmabuf, xfer->buffer, size);

	err = pipe->methods->transfer(xfer);

	if (err != USBD_IN_PROGRESS && err) {
		/* The transfer has not been queued, so free buffer. */
		if (xfer->rqflags & URQ_AUTO_DMABUF) {
			free(xfer->dmabuf, M_USB);
			xfer->rqflags &= ~URQ_AUTO_DMABUF;
		}
	}

	if (!(xfer->flags & USBD_SYNCHRONOUS))
		return err;

	/* Sync transfer, wait for completion. */
	if (err != USBD_IN_PROGRESS)
		return err;
	if (!xfer->done)
		panic("usbd_transfer: not done");

	return xfer->status;
}

void
usbd_free_buffer(usbd_xfer_handle xfer)
{
	xfer->rqflags &= ~URQ_AUTO_DMABUF;
	free(xfer->dmabuf, M_USB);
}

void
usbd_free_xfer(usbd_xfer_handle xfer)
{
	if (xfer->rqflags & URQ_AUTO_DMABUF)
		usbd_free_buffer(xfer);
	free(xfer, M_USB);
}

void
usbd_setup_xfer(xfer, pipe, buffer, length, flags)
	usbd_xfer_handle xfer;
	usbd_pipe_handle pipe;
	void *buffer;
	u_int32_t length;
	u_int16_t flags;
{
	xfer->pipe = pipe;
	xfer->buffer = buffer;
	xfer->length = length;
	xfer->actlen = 0;
	xfer->flags = flags;
	xfer->status = USBD_NOT_STARTED;
	xfer->rqflags &= ~URQ_REQUEST;
}

int
usbd_clear_endpoint_stall_async(usbd_pipe_handle pipe)
{
	panic("usbd_clear_endpoint_stall_async");
}

/*** Internal routines ***/

/* Called at splusb() */
void
usb_transfer_complete(usbd_xfer_handle xfer)
{
	usbd_pipe_handle pipe = xfer->pipe;
	int repeat;

	repeat = pipe->repeat;

	if (!(xfer->flags & USBD_NO_COPY) && xfer->actlen != 0 &&
	    usbd_xfer_isread(xfer)) {
#ifdef DIAGNOSTIC
		if (xfer->actlen > xfer->length) {
			printf("usb_transfer_complete: actlen > len %d > %d\n",
			       xfer->actlen, xfer->length);
			xfer->actlen = xfer->length;
		}
#endif
		memcpy(xfer->buffer, xfer->dmabuf, xfer->actlen);
	}

	/* if we allocated the buffer in usbd_transfer() we free it here. */
	if (xfer->rqflags & URQ_AUTO_DMABUF) {
		if (!repeat) {
			free(xfer->dmabuf, M_USB);
			xfer->rqflags &= ~URQ_AUTO_DMABUF;
		}
	}

	xfer->done = 1;
	if (!xfer->status && xfer->actlen < xfer->length &&
	    !(xfer->flags & USBD_SHORT_XFER_OK)) {
		xfer->status = USBD_SHORT_XFER;
	}

	pipe->methods->done(xfer);
}

int
usbd_do_request(dev, req, data)
	usbd_device_handle dev;
	usb_device_request_t *req;
	void *data;
{
	struct usbd_xfer xfer;
	int err;

	bzero(&xfer, sizeof xfer);
	xfer.pipe = dev->default_pipe;
	xfer.buffer = data;
	xfer.length = UGETW(req->wLength);
	xfer.actlen = 0;
	xfer.flags = USBD_SYNCHRONOUS;
	xfer.status = USBD_NOT_STARTED;
	xfer.request = *req;
	xfer.rqflags = URQ_REQUEST;

	err = usbd_transfer(&xfer);

	if (err == USBD_STALLED)
		panic("control endpoint stalled");

	return err;
}

/*
 * Search for a vendor/product pair in an array.  The item size is
 * given as an argument.
 */
const struct usb_devno *
usb_match_device(const struct usb_devno *tbl, u_int nentries, u_int sz,
		 u_int16_t vendor, u_int16_t product)
{
	while (nentries-- > 0) {
		u_int16_t tproduct = tbl->ud_product;
		if (tbl->ud_vendor == vendor &&
		    (tproduct == product || tproduct == USB_PRODUCT_ANY))
			return tbl;
		tbl = (const struct usb_devno *)((const char *)tbl + sz);
	}
	return NULL;
}

/* Delay for a certain number of ms */
void
usb_delay_ms(ms)
	u_int ms;
{
	/* Wait at least two clock ticks so we know the time has passed. */
	delay((ms+1) * 1000);
}

usb_interface_descriptor_t *
usbd_find_idesc(usb_config_descriptor_t *cd, int ifaceidx, int altidx)
{
	char *p = (char *)cd;
	char *end = p + UGETW(cd->wTotalLength);
	usb_interface_descriptor_t *d;
	int curidx, lastidx, curaidx = 0;

	for (curidx = lastidx = -1; p < end; ) {
		d = (usb_interface_descriptor_t *)p;
		if (d->bLength == 0) /* bad descriptor */
			break;
		p += d->bLength;
		if (p <= end && d->bDescriptorType == UDESC_INTERFACE) {
			if (d->bInterfaceNumber != lastidx) {
				lastidx = d->bInterfaceNumber;
				curidx++;
				curaidx = 0;
			} else
				curaidx++;
			if (ifaceidx == curidx && altidx == curaidx)
				return d;
		}
	}
	return NULL;
}

int
usbd_fill_iface_data(usbd_device_handle dev, int ifaceidx, int altidx)
{
	usbd_interface_handle ifc = &dev->ifaces[ifaceidx];
	usb_interface_descriptor_t *idesc;
	char *p, *end;
	int endpt, nendpt;

	idesc = usbd_find_idesc(dev->cdesc, ifaceidx, altidx);
	if (idesc == NULL)
		return (USBD_INVAL);
	ifc->device = dev;
	ifc->idesc = idesc;
	ifc->index = ifaceidx;
	nendpt = ifc->idesc->bNumEndpoints;

	if (nendpt != 0) {
		ifc->endpoints = malloc(nendpt * sizeof(struct usbd_endpoint),
					M_USB, M_NOWAIT);
		if (ifc->endpoints == NULL)
			return (USBD_NOMEM);
	} else
		ifc->endpoints = NULL;
	p = (char *)ifc->idesc + ifc->idesc->bLength;
	end = (char *)dev->cdesc + UGETW(dev->cdesc->wTotalLength);
#define ed ((usb_endpoint_descriptor_t *)p)
	for (endpt = 0; endpt < nendpt; endpt++) {
		for (; p < end; p += ed->bLength) {
			if (p + ed->bLength <= end && ed->bLength != 0 &&
			    ed->bDescriptorType == UDESC_ENDPOINT)
				goto found;
			if (ed->bLength == 0 ||
			    ed->bDescriptorType == UDESC_INTERFACE)
				break;
		}
		/* passed end, or bad desc */
		printf("usbd_fill_iface_data: bad descriptor(s): %s\n",
		       ed->bLength == 0 ? "0 length" :
		       ed->bDescriptorType == UDESC_INTERFACE ? "iface desc":
		       "out of data");
		goto bad;
	found:
		ifc->endpoints[endpt].edesc = ed;
		if (dev->speed == USB_SPEED_HIGH) {
			u_int mps;
			/* Control and bulk endpoints have max packet limits. */
			switch (UE_GET_XFERTYPE(ed->bmAttributes)) {
			case UE_CONTROL:
				mps = USB_2_MAX_CTRL_PACKET;
				goto check;
			case UE_BULK:
				mps = USB_2_MAX_BULK_PACKET;
			check:
				if (UGETW(ed->wMaxPacketSize) != mps) {
					USETW(ed->wMaxPacketSize, mps);
#ifdef DIAGNOSTIC
					printf("usbd_fill_iface_data: bad max "
					       "packet size\n");
#endif
				}
				break;
			default:
				break;
			}
		}
		ifc->endpoints[endpt].refcnt = 0;
		p += ed->bLength;
	}
#undef ed
	return 0;

 bad:
	if (ifc->endpoints != NULL) {
		free(ifc->endpoints, M_USB);
		ifc->endpoints = NULL;
	}
	return USBD_INVAL;
}

void
usbd_free_iface_data(usbd_device_handle dev, int ifcno)
{
	usbd_interface_handle ifc = &dev->ifaces[ifcno];
	if (ifc->endpoints)
		free(ifc->endpoints, M_USB);
}

Static int
usbd_set_config(usbd_device_handle dev, int conf)
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_DEVICE;
	req.bRequest = UR_SET_CONFIG;
	USETW(req.wValue, conf);
	USETW(req.wIndex, 0);
	USETW(req.wLength, 0);
	return usbd_do_request(dev, &req, 0);
}

int
usbd_set_config_index(usbd_device_handle dev, int index, int msg)
{
	usb_config_descriptor_t cd, *cdp;
	int err;
	int i, ifcidx, nifc, len;

	/* XXX check that all interfaces are idle */
	if (dev->config != USB_UNCONFIG_NO) {

		/* Free all configuration data structures. */
		nifc = dev->cdesc->bNumInterface;
		for (ifcidx = 0; ifcidx < nifc; ifcidx++)
			usbd_free_iface_data(dev, ifcidx);
		free(dev->ifaces, M_USB);
		free(dev->cdesc, M_USB);
		dev->ifaces = NULL;
		dev->cdesc = NULL;
		dev->config = USB_UNCONFIG_NO;
	}

	if (index == USB_UNCONFIG_INDEX) {
		/* We are unconfiguring the device, so leave unallocated. */
		err = usbd_set_config(dev, USB_UNCONFIG_NO);
		return err;
	}

	/* Get the short descriptor. */
	err = usbd_get_config_desc(dev, index, &cd);
	if (err)
		return err;
	len = UGETW(cd.wTotalLength);
	cdp = malloc(len, M_USB, M_NOWAIT);
	if (cdp == NULL)
		return USBD_NOMEM;

	/* Get the full descriptor.  Try a few times for slow devices. */
	for (i = 0; i < 3; i++) {
		err = usbd_get_desc(dev, UDESC_CONFIG, index, len, cdp);
		if (!err)
			break;
		usb_delay_ms(200);
	}
	if (err)
		goto bad;

	if (cdp->bDescriptorType != UDESC_CONFIG) {
		err = USBD_INVAL;
		goto bad;
	}

	/* Set the actual configuration value. */
	err = usbd_set_config(dev, cdp->bConfigurationValue);
	if (err)
		goto bad;

	/* Allocate and fill interface data. */
	nifc = cdp->bNumInterface;
	dev->ifaces = malloc(nifc * sizeof(struct usbd_interface),
			     M_USB, M_NOWAIT);
	if (dev->ifaces == NULL) {
		err = USBD_NOMEM;
		goto bad;
	}
	dev->cdesc = cdp;
	dev->config = cdp->bConfigurationValue;
	for (ifcidx = 0; ifcidx < nifc; ifcidx++) {
		err = usbd_fill_iface_data(dev, ifcidx, 0);
		if (err) {
			while (--ifcidx >= 0)
				usbd_free_iface_data(dev, ifcidx);
			goto bad;
		}
	}

	return 0;

 bad:
	free(cdp, M_USB);
	return err;
}

/* XXX add function for alternate settings */

usbd_pipe_handle
usbd_setup_pipe(dev, iface, ep)
	usbd_device_handle dev;
	usbd_interface_handle iface;
	struct usbd_endpoint *ep;
{
	usbd_pipe_handle p;
	int err;

	p = malloc(dev->bus->pipe_size, M_USB, M_NOWAIT);
	if (p == NULL)
		return NULL;
	p->device = dev;
	p->iface = iface;
	p->endpoint = ep;
	ep->refcnt++;
	p->refcnt = 1;
	p->intrxfer = 0;
	p->repeat = 0;
	err = dev->bus->methods->open_pipe(p);
	if (err) {
		free(p, M_USB);
		return NULL;
	}
	return p;
}

int
usbd_getnewaddr(usbd_bus_handle bus)
{
	int addr;

	for (addr = 1; addr < USB_MAX_DEVICES; addr++)
		if (bus->devices[addr] == 0)
			return addr;
	return -1;
}

/*
 * Called when a new device has been put in the powered state,
 * but not yet in the addressed state.
 * Get initial descriptor, set the address, get full descriptor,
 * and attach a driver.
 */
int
usbd_new_device(bus, speed, port, dev)
	usbd_bus_handle bus;
	int speed, port;
	struct usbd_device *dev;
{
	usb_device_descriptor_t *dd;
	int err, addr, i;

	addr = usbd_getnewaddr(bus);
	if (addr < 0) {
		/* No free USB addresses. */
		return USBD_NO_ADDR;
	}

	dev->bus = bus;

	/* Set up default endpoint handle. */
	dev->def_ep.edesc = &dev->def_ep_desc;

	/* Set up default endpoint descriptor. */
	dev->def_ep_desc.bLength = USB_ENDPOINT_DESCRIPTOR_SIZE;
	dev->def_ep_desc.bDescriptorType = UDESC_ENDPOINT;
	dev->def_ep_desc.bEndpointAddress = USB_CONTROL_ENDPOINT;
	dev->def_ep_desc.bmAttributes = UE_CONTROL;
	USETW(dev->def_ep_desc.wMaxPacketSize, USB_MAX_IPACKET);
	dev->def_ep_desc.bInterval = 0;

	dev->address = USB_START_ADDR;
	dev->ddesc.bMaxPacketSize = 0;
	dev->speed = speed;

	/* Establish the default pipe. */
	dev->default_pipe = usbd_setup_pipe(dev, 0, &dev->def_ep);
	if (dev->default_pipe == NULL)
		return USBD_NOMEM;

	/* Set the address.  Do this early; some devices need that. */
	err = usbd_set_address(dev, addr);
	if (err)
		return USBD_SET_ADDR_FAILED;

	/* Allow device time to set new address */
	usb_delay_ms(USB_SET_ADDRESS_SETTLE);
	dev->address = addr;	/* New device address now */
	bus->devices[addr] = dev;

	dd = &dev->ddesc;
	/* Try a few times in case the device is slow (i.e. outside specs.) */
	for (i = 0; i < 10; i++) {
		/* Get the first 8 bytes of the device descriptor. */
		err = usbd_get_desc(dev, UDESC_DEVICE, 0, USB_MAX_IPACKET, dd);
		if (!err)
			break;
		usb_delay_ms(200);
	}
	if (err)
		return err;

	if (speed == USB_SPEED_HIGH) {
		/* Max packet size must be 64 (sec 5.5.3). */
		dd->bMaxPacketSize = USB_2_MAX_CTRL_PACKET;
	}

	if (dd->bDescriptorType != UDESC_DEVICE) {
		/* Illegal device descriptor */
		return USBD_INVAL;
	}

	if (dd->bLength < USB_DEVICE_DESCRIPTOR_SIZE)
		return USBD_INVAL;

	USETW(dev->def_ep_desc.wMaxPacketSize, dd->bMaxPacketSize);

	return usbd_get_device_desc(dev, dd);
}

int
usbd_get_desc(usbd_device_handle dev, int type, int index, int len, void *desc)
{
	usb_device_request_t req;

	req.bmRequestType = UT_READ_DEVICE;
	req.bRequest = UR_GET_DESCRIPTOR;
	USETW2(req.wValue, type, index);
	USETW(req.wIndex, 0);
	USETW(req.wLength, len);
	return usbd_do_request(dev, &req, desc);
}

int
usbd_get_config_desc(usbd_device_handle dev, int confidx,
		     usb_config_descriptor_t *d)
{
	int err;

	err = usbd_get_desc(dev, UDESC_CONFIG, confidx,
			    USB_CONFIG_DESCRIPTOR_SIZE, d);
	if (err)
		return err;
	if (d->bDescriptorType != UDESC_CONFIG)
		return USBD_INVAL;

	return 0;
}

int
usbd_get_device_desc(usbd_device_handle dev, usb_device_descriptor_t *d)
{
	return usbd_get_desc(dev, UDESC_DEVICE,
			     0, USB_DEVICE_DESCRIPTOR_SIZE, d);
}

int
usbd_set_address(usbd_device_handle dev, int addr)
{
	usb_device_request_t req;

	req.bmRequestType = UT_WRITE_DEVICE;
	req.bRequest = UR_SET_ADDRESS;
	USETW(req.wValue, addr);
	USETW(req.wIndex, 0);
	USETW(req.wLength, 0);
	return usbd_do_request(dev, &req, 0);
}
